#!/bin/bash
#*****************************************************************************#
# build-chromium - Build Chromium RPMs for your current Yocto image           #
# Author: Marc-Oliver Westerburg <westerburg@garz-fricke.com>                 #
# Copyright 2015 Garz & Fricke GmbH                                           #
#*****************************************************************************#

#*****************************************************************************#
# Helper Functions                                                            #
#*****************************************************************************#
do_bitbake ()
{
	echo "Building $1"
	export CHROMIUM_VERSION="$1%"
	bitbake chromium -c release
}


#*****************************************************************************#
# Script                                                                      #
#*****************************************************************************#

# Check if a Yocto build directory can be found
if [ -z $BUILDDIR ]
then
	echo "No Yocto build directory found. Please call \"setplatform\" first"
	exit 1
fi

# Check if currently used Yocto build environment supports building specific Chromium versions
echo $BB_ENV_EXTRAWHITE|grep -q CHROMIUM_VERSION
if [ $? -ne 0 ]
then
	echo "Yocto build environment out-dated. Please uppdate your Yocto-BSP and re-run \"setplatform\"."
	exit 1
fi

# Change to current Yocto build directory
cd $BUILDDIR 
if [ $? -ne 0 ]
then
	echo "Yocto build directory ($BUILDDIR) does not exist."
	exit 1
fi	

# Search for available Chromium recipes
available_recipes=`find ../sources/meta-guf/meta*/recipes-browser/chromium/* \
					    ../sources/meta-browser/recipes-browser/chromium/* \
					    -maxdepth 0 -type f -name "chromium_*.bb" \
					    -print | sort -V 2>/dev/null`

# Generate versions numbers from recipes and determine their support status
for i in $available_recipes
do
	pref=`(grep "DEFAULT_PREFERENCE" $i || echo "0") | sed -e "s/.*\"\(.*\)\"/\1/"`
	if [ $pref == "-1" ]
	then
		status="(beta)"
	elif [ $pref == "1" ]
	then
		status="(stable-default)"
	else
		status="(stable)"
	fi
	p=`basename $i|sed -e "s/chromium_\(.*\).bb/\1/"`
	available_versions="$available_versions $p:$status"
done



# Print usage information if called without parameters
if [ $# -eq 0 ]
then
	echo "Please specify a Chromium version"
	echo ""
	echo "  Syntax: $0 <Chromium version> [<Chromium version> ...]"
	echo ""
	echo "Available versions:"
	echo "--------------------"
	echo "   default      (version may depend on current Yocto image configuration)"
	echo "   all          (builds all available versions)"
	echo "   all-stable   (builds all available stable versions)"
	for p in $available_versions
	do
		echo "   `echo $p|sed -e "s/\(.*\):\(.*\)/\1  \t\2/"`"
	done
	echo "  Multiple version numbers may be specified."
	echo "  Version numbers may be abbreviated."
	exit 1
fi


# Build desired Chromium versions
while [ ! -z $1 ]
do
	if [ $1 == "all-stable" ]
	then
		for p in $available_versions
		do
			version=`echo $p|sed -e "s/\(.*\):\(.*\)/\1/"`
			status=`echo $p|sed -e "s/\(.*\):\(.*\)/\2/"`
			if [ $status == "(stable)" -o $status == "(stable-default)" ]
			then
				do_bitbake $version
			fi
		done
		exit 0
	elif [ $1 == "all" ]
	then
		for p in $available_versions
		do
			version=`echo $p|sed -e "s/\(.*\):\(.*\)/\1/"`
			do_bitbake $version
		done
		exit 0
	else	
		do_bitbake $1
		if [ $1 == "default" ]
		then
			exit 0
		fi
	fi		
	shift
done	
