#!/usr/bin/python3
#
# Call this script from your build dir, e.g. Yocto-Dora/build-vincell-0,
# with the according image as a parameter.
#
# $ ../parse_packages.py guf-image
#

# imports --------------------------------------------------------------------#
import subprocess
import re
import argparse
import textwrap
import os
import fnmatch


# helper functions -----------------------------------------------------------#

def check_package_file(historydir):
    historydir_exists = os.path.isdir(historydir)
    if historydir_exists is False:
        print("Directory '" + historydir + "' does not exist. The buildhistory could not be found")
    else:
        package_file_exists = os.path.isfile(historydir + '/installed-packages.txt')
        package_names_file_exists = os.path.isfile(historydir + '/installed-package-names.txt')
        return package_file_exists and package_names_file_exists
    return False

# parse packages and versions
def get_image_packages_with_version(historydir):
    packages_file = open(historydir + '/installed-packages.txt', 'r')
    packages_name_file = open(historydir + '/installed-package-names.txt', 'r')
    package_list = list()
    for package_name in packages_name_file:
        packages_line = packages_file.readline()

        version = packages_line[(len(package_name)):]
        #print( package_name + '  ' + version)
        sep = re.search('\.[^.]*\.rpm', version).group(0)
        version = version[:version.rfind(sep)]

        package_name = package_name.strip("\n")

        # create list of tuples with name and version
        package_list.append((package_name, version))

    return package_list

# print packages in <name>-<version> format
def std_print(package_list, recipesguf):
    print('-------------------------------------------------------')
    for item in package_list:
        if not item[0] in recipesguf:
            print(item[0] + "-" + item[1])
    print('-------------------------------------------------------')
    print('Guf recepies')
    for item in package_list:
        if item[0] in recipesguf:
            print(item[0] + "-" + item[1])
    print('-------------------------------------------------------')

def html_print_header(with_description):
    print('<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"')
    print('		"http://www.w3.org/TR/html4/loose.dtd">')
    print('<html>')
    print('<head>')
    print('	<link rel="stylesheet" type="text/css" href="../../common/style.css" />')
    print('</head>')
    print('<body>')
    print('	<div id="content">')
    print('		<table class="sort">')
    if with_description:
        print('			<tr class="tableheader"><td style="width:36%">Package name</td><td>Version</td><td>Description</td></tr>')
    else:
        print('			<tr class="tableheader"><td style="width:36%">Package name</td><td>Version</td></tr>')

def html_print_footer():
    print('		</table>')
    print('	</div>')
    print('</body>')
    print('</html>')


def html_print(package_list, recipesguf, out, with_description):
    print ( out)
    # print standard packages
    if( (out == 0) or (out == 1)):
        html_print_header(with_description)

        for item in package_list:
            if not item[0] in recipesguf:
                print('\t<tr><td class="tableleft">'+ item[0] +'</td><td>'+ item[1] + '</td></tr>')
        
        html_print_footer()

    if( out == 0 ):
        print('-------------------------------------------------------')

    if( (out == 0) or (out == 2)):
        # print guf packages
        html_print_header(with_description)
        for item in package_list:
            if item[0] in recipesguf:
                if with_description:
                    print('\t<tr><td class="tableleft">'+ item[0] +'</td><td>'+ item[1] + '</td><td>' + recipesguf[item[0]] + '</td></tr>')
                else:
                    print('\t<tr><td class="tableleft">'+ item[0] +'</td><td>'+ item[1] + '</td></tr>')

        html_print_footer()

def recursive_glob(treeroot, pattern):
    results = []
    for base, dirs, files in os.walk(treeroot):
      goodfiles = fnmatch.filter(files, pattern)
      results.extend(os.path.join(base, f) for f in goodfiles)
    return results

def get_description_from_receipe(receipe):
    f = open(receipe, 'r')
    content = f.read()
    f.close()
    description = re.search('(?<=DESCRIPTION)\s*=\s*"[^"]*"', content, flags=re.MULTILINE)
    if description and len(description.group(0)) > 0:
        return description.group(0).strip(" \"=")
    summary = re.search('(?<=SUMMARY)\s*=\s*"[^"]*"', content, flags=re.MULTILINE)
    if summary and len(summary.group(0)) > 0:
        return summary.group(0).strip(" \"=")
    return ""


def read_guf_packagenames(guflayer):
    recipes = recursive_glob(guflayer, '*.bb')
    recipesguf = {}
    for f in recipes:
        if 'recipes-guf' in f:
            description = get_description_from_receipe (f)
            name = (f.rsplit('/',1)[1]).split('.bb')[0].split('_')[0]
            recipesguf[name]= description
    return recipesguf

# main functionality-----------------------------------------------------------#

# Parse arguments
parser = argparse.ArgumentParser(
    formatter_class=argparse.RawDescriptionHelpFormatter,
    description=textwrap.dedent('''\
    Call this script from anywhere,
    with the according image as parameter. With the '--html' option
    enabled the output will be formatted in html-syntax.

    $ parse_packages.py -d buildhistory/images/santaro/eglibc/guf-image -g source/meta-guf '''))

parser.add_argument('-d', '--histdir', 
    help='Folder with build history files', default='.')
parser.add_argument('-g', '--guflayer', 
    help='Use this directory as guf layer to sort out guf packages', default='')
parser.add_argument('-o', '--outputtype', 
    help='Outputtype 0 = all, 1 = standard only, 2 = guf only', default=0)
args = parser.parse_args()

# initialize environment
historydir = args.histdir

if args.guflayer != '':
    recipesguf = read_guf_packagenames(args.guflayer)
    # print(recipesguf)
else:
    recipesguf = []

# parse packages and print the result
if not check_package_file(historydir):
    print('An error occoured. Exiting with non-zero status 1')
    exit(1)

packages = get_image_packages_with_version(historydir)

# print package list
html_print(packages, recipesguf,int( args.outputtype), (int( args.outputtype) == 2) )
exit(0)

# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
