/*
 * Copyright (C) 2013 Carsten Behling, Garz & Fricke GmbH
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <limits.h>
#include <linux/version.h>

#define AC97_PROCFS_ENTRY "/proc/asound/card0/codec97#0/"
/*#define AC97_PROCFS_ENTRY "/test.txt" */
#define AC97_VENDOR_ID1 0x7c
#define AC97_VENDOR_ID2 0x7e

struct ac97 {
	int fd;
	char procfs_entry[256];
};

typedef struct
{
	unsigned short id1;
	unsigned short id2;
	char		   *name;
	unsigned short def_reg1;
	unsigned short def_msk1;
	unsigned short def_reg2;
	unsigned short def_msk2;
} codec_t;

static codec_t codec_list[] =
{
	{ 0x5053, 0x4304, "NXP UCB1400", 0x1a, 0x0707, 0x1c, 0x0f0f },
	{ 0x574D, 0x4C05, "Wolfson WM9705", 0x1a, 0x0707, 0x1c, 0x0f0f },
	{ 0xffff, 0xffff, "unknown"    , 0xff, 0xffff, 0xff, 0xffff }
};

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,10,17)

#define CODEC_FILE "ac97#0-0+regs"

static int read_ac97(struct ac97 *codec, short reg, unsigned short *value)
{
#define BUFFER_LENGTH 13
	char path[256];
	char buffer[BUFFER_LENGTH];
	int readpos = 0;	int ret = 0, read_reg, read_val;
	
	snprintf(path, sizeof(path), "%s%s", codec->procfs_entry, CODEC_FILE);	
	codec->fd = open(path, O_RDWR | O_SYNC | O_RSYNC);
	if(codec->fd < 0) {
		fprintf(stderr,"Failed to open %s\n", path);
		return 1;
	}
	do{
		ret = read( codec->fd, buffer, BUFFER_LENGTH - 1);
		if( ret < 0)
			goto on_error;
		buffer[ret] = 0;
		ret = sscanf( buffer , "0:%02x = %04x\n", &read_reg, &read_val);
	}while( ret == 2 && read_reg < reg );

	if( ret == 2 && read_reg == reg)
	{
		*value = read_val;
		goto no_error;
	}
	
on_error:
	fprintf(stderr,"Failure reading on %s: %d\n", path, ret);
	
no_error:
	close(codec->fd);
	return 0;	
}

static int write_ac97(struct ac97 *codec, short reg, unsigned short *value)
{
	char path[256];
	char buffer[256];
	int readpos = 0;	int ret = 0, read_reg, read_val;

	snprintf(path, sizeof(path), "%s%s", codec->procfs_entry, CODEC_FILE);	
	codec->fd = open(path, O_RDWR | O_SYNC | O_RSYNC);
	if(codec->fd < 0) {
		fprintf(stderr,"Failed to open %s\n", path);
		return 1;
	}

	ret = snprintf(buffer, sizeof(buffer), "%02x %04x", reg, *value);	
	ret = write( codec->fd, buffer, ret);
	if( ret > 0)
		goto no_error;
	
on_error:
	fprintf(stderr,"Failure writing to %s: %d\n", path, ret);
	
no_error:
	close(codec->fd);
	return 0;	
}

#else
static int read_ac97(struct ac97 *codec, short reg, unsigned short *value)
{
	char path[256];
	char value_string[5];
	
	snprintf(path, sizeof(path), "%s%02x", codec->procfs_entry, reg);	
	codec->fd = open(path, O_RDWR | O_SYNC | O_RSYNC);
	if(codec->fd < 0) {
		fprintf(stderr,"Failed to open %s\n", path);
		return 1;
	}
	if(read(codec->fd, (void*)value_string, sizeof(value_string))
		!= sizeof(value_string)) {
		close(codec->fd);
		fprintf(stderr,"Failed to read from %s\n", path);
		return 1;	
	}
	sscanf(value_string, "%hx", value);
	close(codec->fd);
	return 0;	
}

static int write_ac97(struct ac97 *codec, short reg, unsigned short *value)
{
	char path[256];
	char value_string[5];

	snprintf(path, sizeof(path), "%s%02x", codec->procfs_entry, reg);	
	codec->fd = open(path, O_RDWR | O_SYNC | O_RSYNC);
	if(codec->fd < 0) {
		fprintf(stderr,"Failed to open %s\n", path);
		return 1;
	}
	snprintf(value_string, sizeof(value_string), "%04x", *value);
	
	if(write(codec->fd, (const void*)value_string, sizeof(value_string))
		!= sizeof(value_string)) {
		close(codec->fd);
		fprintf(stderr,"Failed to write to %s\n", path);
		return 1;
	}
	close(codec->fd);
	return 0;

}
#endif

static void usage(void)
{
	fprintf(stderr, "Usage: ac97regtest [options]\n");
	fprintf(stderr, "   -h            Print this help.\n");
	fprintf(stderr, "   -l <length>   number of read/write accesses to perform.\n");
	fprintf(stderr, "   -1 <reg,mask> 1. register to use for test.\n");
	fprintf(stderr, "   -2 <reg,mask> 2. register to use for test.\n");
	fprintf(stderr, "\n");
	fprintf(stderr, "All parameters are optional. Known codecs have compiled-in\n");
	fprintf(stderr, "default register/mask values.\n");
	return;
}

int main(int argc, char *argv[])
{
	int opt;	
	long iterations = 1000; /* number of registers read/write accesses */
	char *tmp;
	struct ac97 codec;
	unsigned long i = 0;
	unsigned long n_errors = 0;
	unsigned short vid1 = 0xffff;
	unsigned short vid2 = 0xffff;
	unsigned short reg1 = 0xff;
	unsigned short reg2 = 0xff;
	unsigned short msk1 = 0xffff;
	unsigned short msk2 = 0xffff;
	unsigned short prv1 = 0;
	unsigned short prv2 = 0;
	
	while ((opt = getopt(argc, argv, "hvdl:1:2:")) != -1) {
		switch (opt) {
			case 'l':
				iterations = strtol(optarg, NULL, 0);
				if(iterations <= 0) {
					fprintf(stderr, "Parameter error:\n");
					fprintf(stderr, "  l must be greater than 1.\n");
					fprintf(stderr, "\n");
					usage();
					exit(EXIT_FAILURE);
				}
				break;
			case '1':
				tmp = strtok(optarg, ",");
				sscanf(tmp, "0x%hx", &reg1);
				if(reg1 > 0x7e) {
					fprintf(stderr, "Parameter error:\n");
					fprintf(stderr, "  reg must be lower than 0x80.\n");
					fprintf(stderr, "\n");
					usage();
					exit(EXIT_FAILURE);
				}
				tmp = strtok(NULL, ",");
				sscanf(tmp, "0x%hx", &msk1);
				break;
			case '2':
				tmp = strtok(optarg, ",");
				sscanf(tmp, "0x%hx", &reg2);
				if(reg2 > 0x7e) {
					fprintf(stderr, "Parameter error:\n");
					fprintf(stderr, "  reg must be lower than 0x80.\n");
					fprintf(stderr, "\n");
					usage();
					exit(EXIT_FAILURE);
				}
				tmp = strtok(NULL, ",");
				sscanf(tmp, "0x%hx", &msk2);
				break;
			case 'h':
			default: /* '?' */
				usage();
				exit(EXIT_FAILURE);
				break;
		}
	}
	
	strncpy(codec.procfs_entry, AC97_PROCFS_ENTRY, sizeof(codec.procfs_entry));

	if (read_ac97(&codec, AC97_VENDOR_ID1, &vid1) != 0 ||
		read_ac97(&codec, AC97_VENDOR_ID2, &vid2) != 0) {
		printf("Failed reading AC97 codec ID\n");
		goto out;
	}

	for (i = 0; codec_list[i].id1 != 0xffff; i++) {
		if ((codec_list[i].id1 == vid1) &&
			(codec_list[i].id2 == vid2))
			break;
	}
	
	if (reg1 == 0xff) {
		reg1 = codec_list[i].def_reg1;
		msk1 = codec_list[i].def_msk1;
	}

	if (reg2 == 0xff) {
		reg2 = codec_list[i].def_reg2;
		msk2 = codec_list[i].def_msk2;
	}
	
	printf("Found AC97 codec: 0x%04x 0x%04x (%s)\n", codec_list[i].id1, codec_list[i].id2, codec_list[i].name);
	printf("Executing %lu reads/writes\n", iterations);
	printf("on registers: 0x%04x (Mask: 0x%04x), 0x%04x (Mask: 0x%04x)\n", reg1, msk1, reg2, msk2);
		

	for (i = 0; i < iterations; i++) {
		unsigned short  val1 = (unsigned short)(random() >> 16) & msk1;
		unsigned short  val2 = (unsigned short)(random() >> 16) & msk2;
		unsigned short ret1 = 0;
		unsigned short ret2 = 0;

		if (write_ac97(&codec, reg1, &val1) != 0 ||
			write_ac97(&codec, reg2, &val2) != 0) {
			printf("Write error on round %lu\n", i);
			n_errors++;
			continue;
		}
		usleep(10000);
		if (read_ac97(&codec, reg1, &ret1) != 0 ||
			read_ac97(&codec, reg2, &ret2) != 0) {
			printf("Read error on round %lu\n", i);
			n_errors++;
			continue;
		}
		usleep(10000);		
		if (((ret1 & msk1) != val1) ||
			((ret2 & msk2) != val2)) {
			unsigned short tmp1 = 0;
			unsigned short tmp2 = 0;
			
			read_ac97(&codec, reg1, &tmp1);
			read_ac97(&codec, reg2, &tmp2);
			
			printf("Compare error on round %lu:\n", i);
			printf("  0x%02x:0x%04x should be 0x%04x (prv: 0x%04x re:0x%04x)", reg1, ret1 & msk1, val1, prv1, tmp1 & msk1);
			
			if ((ret1 & msk1) != val1) {
				if ((ret1 & msk1) == (prv1 & msk1))
					printf(" #WE\n");
				else if ((ret1 & msk1) != (tmp1 & msk1))
					printf(" #RE\n");
				else
					printf(" #??\n");
			}
			else
				printf("\n");
			printf("  0x%02x:0x%04x should be 0x%04x (prv: 0x%04x re:0x%04x)", reg2, ret2 & msk2, val2, prv2, tmp2 & msk2);

			if ((ret2 & msk2) != val2)
			{
				if ((ret2 & msk2) == (prv2 & msk2))
					printf(" #WE\n");
				else if ((ret2 & msk2) != (tmp2 & msk2))
					printf(" #RE\n");
				else
					printf(" #??\n");
			}
			else
				printf("\n");
			n_errors++;
			continue;
		}
		prv1 = val1;
		prv2 = val2;
	}
	
	printf("finished with %lu errors\n", n_errors);

out:
	if (!n_errors)
		exit(EXIT_SUCCESS);
	else
		exit(EXIT_FAILURE);
}
