#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>

#include <errno.h>

#include <sys/ioctl.h>

#ifdef ANDROID
#include <linux/i2c.h>
#else
#include <linux/i2c-dev.h>
#endif

#define DEVICE_NAME "pcf8563"
#define DEVICE_ADDRESS 0x51
#define TIMER_CTRL_REG 0x0E
#define TIMER_CTRL_REG_TD (0x3 << 0)
#define TIMER_CTRL_REG_TE (0x1 << 7)


#define DEFAULT_BUS "/dev/i2c-0"

#define FNG_INIT 0x3
#define FNG_REG  0x2
#define FNG_ALT  0x1
#define FNG_INV  0x0

#define MODE_INIT "initialized"
#define MODE_REG  "regular"
#define MODE_ALT  "alternative"
#define MODE_INV  "invalid"

#define V_INFO	3
#define V_DEBUG	2
#define V_WARN	1
#define V_ERROR	0

char *mode[] = {
	MODE_INV,
	MODE_ALT,
	MODE_REG,
	MODE_INIT,
};

int verbosity = V_WARN;

void print_usage() {
	if(verbosity <= V_ERROR)
		return;

	fprintf(stderr, "Usage: bootselect [-qv] [<bootmode>]\n");
	fprintf(stderr, "\n");
	fprintf(stderr, "When called without parameters, bootselect reads the\n");
	fprintf(stderr, "current boot-mode setting from the RTC-register\n");
	fprintf(stderr, "\n");
	fprintf(stderr, "Options:\n");
	fprintf(stderr, "    <bootmode>    Writes the given boot-mode into the RTC-register\n");
	fprintf(stderr, "                  May be one of: %s %s %s %s\n", MODE_INIT,MODE_REG,MODE_ALT,MODE_INV);
	fprintf(stderr, "    -v            Turns on debug messages on stderr\n");
	fprintf(stderr, "    -vv           Turns on info messages on stderr\n");
	fprintf(stderr, "    -q            Turns off warning messages on stderr\n");
	fprintf(stderr, "    -qq           Turns off all messages on stderr\n");
}
#define SEARCH_PATCH_CNT 2
static char * searchpath[SEARCH_PATCH_CNT] = 
{
	[0] = "/sys/class/i2c-adapter/i2c-%d/0-0051/name",
	[1] = "/sys/class/i2c-dev/i2c-%d/device/0-0051/name",
};

int i2c_find() {
	int fd;
	char str[255];
	int j = 0;
	for(j= 0; j < SEARCH_PATCH_CNT; j++)
	{
		for(int i=0;i<255;i++) {
			snprintf(str, 255, searchpath[j], i);
			fd = open(str, O_RDONLY);
			if(fd<0) {
				if(verbosity >= V_INFO)
					fprintf(stderr, "Info: No RTC-Device on i2c-%d\n", i);
				continue;
			}
			else {
				read(fd, str, strlen(DEVICE_NAME));
				str[strlen(DEVICE_NAME)]='\0';
				close(fd);
				if(strncmp(str,DEVICE_NAME,strlen(DEVICE_NAME)+1)==0) {
					if(verbosity >= V_DEBUG)
						fprintf(stderr, "Debug: Found RTC-Device on i2c-%d\n", i);
					return i;
				}
			}
		}
	}
	return -1;
}

int i2c_open(char *path) {
	if(verbosity >= V_DEBUG)
		fprintf(stderr, "Debug: opening i2c-bus on %s\n", path);

	int fd = open(path, O_RDWR);
	if(fd<0) return fd;

	//NOTE: We need to use I2C_SLAVE_FORCE since the kernel's
	//      RTC-Driver has claimed the i2c-device. This is save
	//      because the driver does not use the register we are
	//	interested in.
	int ret = ioctl(fd, I2C_SLAVE_FORCE, DEVICE_ADDRESS);
	if(ret<0) return -1;
	else      return fd;
}

int i2c_read_reg(int fd, char reg) {
	char value = 0;
	if(write(fd, &reg, 1)!=1) {
		return -1;
	}

	if(read(fd, &value, 1)!=1) {
		return -1;
	}

	if(verbosity >= V_DEBUG)
		fprintf(stderr, "Debug: read 0x%x from 0x%x\n", value, reg);

	return value;
}

int i2c_write_reg(int fd, char reg, char value) {
	char buf[2];
	buf[0] = reg;
	buf[1] = value;

	if(write(fd, buf, 2) != 2) {
		return -1;
	}

	if(verbosity >= V_DEBUG)
		fprintf(stderr, "Debug: wrote 0x%x to 0x%x\n", value, reg);

	return 0;
}

int parse_mode(char *str) {
	int ret = -1;
	for(int j=0;j<=FNG_INIT;j++) 
	{
		unsigned int len = strlen(mode[j]);
		if( len > strlen(str))
			len = strlen(str);
		if( strncmp(str, mode[j], len) == 0) {
			ret = j;
			break;
		}
	}
	return ret;
}

int main(int argc, char** argv) {
	int write_value = -1;
	int current_value = 0;

	for(int i=1;i<argc;i++) {
		char *arg = argv[i];
		write_value = parse_mode(arg);

		if(write_value>=0) {
			continue;
		}	
		else if(arg[0]=='-'){
			while(*(++arg)) {
				if(*arg == 'q')
					verbosity--;
				else if(*arg == 'v')
					verbosity++;
				else {
					print_usage();
					return -1;
				}
			}
		}
		else {
			print_usage();
			return -1;
		}
	}

	int bus_id = i2c_find();
	if(bus_id<0) {
		if(verbosity >= V_ERROR)
			fprintf(stderr, "Error: Could not find i2c-bus with the rtc-device\n");
		return -2;
	}

	char dev_path[16];
	snprintf(dev_path,16,"/dev/i2c-%d", bus_id);
	int fd = i2c_open(dev_path);

	if(fd<0) {
		if(verbosity >= V_ERROR)
			fprintf(stderr, "Error: could not open i2c device (%d: %s)\n", 
					errno, strerror(errno));
		return -3;
	}

	if(write_value>=0) {
		if(i2c_write_reg(fd, TIMER_CTRL_REG, write_value)) {
			if(verbosity >= V_ERROR)
				fprintf(stderr, "Error: could not write new value! (%d: %s)\n", 
					errno, strerror(errno));
			return -4;
		}
	}

	current_value = i2c_read_reg(fd, TIMER_CTRL_REG);

	if(current_value<0) {
		if(verbosity >= V_ERROR)
			fprintf(stderr, "Error: could not read current value! (%d: %s)\n",
					errno,strerror(errno));
		return -5;
	}
	
	if(current_value & TIMER_CTRL_REG_TE) {
		if(verbosity >= V_WARN)
			fprintf(stderr, "Warning: Timer is enabled!\n");
	}

	fprintf(stdout, "%s\n", mode[current_value & TIMER_CTRL_REG_TD]);
	
	if(write_value>=0 && (current_value & TIMER_CTRL_REG_TD) != write_value) {
		if(verbosity >= V_ERROR)
			fprintf(stderr, "Error: Value could not be saved!\n");
		return -6;
	}

	return 0;
}

