#!/bin/sh

# Yocto installation script for Flash-N-Go System - (c) 2015 Garz & Fricke GmbH
# 
# This script is designed to run from the Flash-N-Go System v4.0 or higher.
# It partitions the device for Linux Yocto and writes Linux Yocto images
# to the internal eMMC memory.
# 
# The name of the root filesystem image can optionally be specified on the
# command line. If this parameter is omitted, the default image is used.
# The script can load the image files either from a local directory (e.g. USB
# drive, SD card or /tmp directory) or over the network via TFTP.
# 
# To execute it locally, boot Flash-N-Go System, place the binary files and
# this script into a common directory and execute the script from the
# Flash-N-Go System shell (replace path with e.g. /mnt/mstick1):
# 
# /path/to/images/fng-install.sh [<Options>]
# 
# To run it using TFTP, configure your TFTP server so that it points to the BSP
# folder containing this script and the binary images.
# Connect your device to the Ethernet, boot Flash-N-Go System and make sure
# it can access your TFTP server (try with 'ping', configure Ethernet with
# 'sconfig' if necessary). Then run the following command from the Flash-N-Go
# System shell (replace the IP address by your TFTP server's IP address and
# remove the '#' sign):
# 
# export TFTP=192.168.1.100; curl tftp://$TFTP/fng-install.sh | sh [-s -- <Options>]
#
# Options:
# -a|--AdditionalFile  Additional OS file (like device tree files)
# -b|--BS              Boot script
# -f|--FS              Root file system archive
# -i|--Image           Boot logo
# -o|--OS              Kernel image
# -p|--ParamFile       Parameter file (each parameter on a separate line)
# -r|--RPM             Additional RPM packages
# -t|--TFTP            Additional file (Format: <filename>:<target>)
# Not yet implemented:
# -e|--EraseRPM        RPMs to erase

BOOTLOADER_MIN_VERSION="@BOOTLOADER_MINIMAL_VERSION@"
ROOTFS="@ROOTFS@"
KERNEL="@KERNEL@"
BOOTSCRIPT="@BOOTSCRIPT@"
DEFAULT_DEVICETREE="@DEVICETREE@"
DEFAULT_DEVICETREE_PREFIX="@DEVICETREE_PREFIX@"
ROOTFS_TYPE="@ROOTFS_TYPE@"

usage()
{
    # Extract help text from header above
	awk '/^[^#]/{p=0}p;/#!\/bin\/sh/{p=1}' $0 | cut -b 3-
}

handleAdditionalFile()
{
    
    for addFile in `echo "$@" | tr -d '"'` ; do
        extension="${addFile##*.}"
        echo "$extension"
        case "$extension" in
         "dtb")
           DEVICETREE="${DEVICETREE} ${addFile}"
           ;;
         "rpm")
           RPM="${RPM} ${addFile}"
           ;;
         *)
            # is string delimited by ':'
            extension="${addFile##*:}"
            if [ "$extension" != "" ] ; then 
                TFTPFILES="${TFTPFILES} ${addFile}"
            fi
           ;;
        esac
    done
}

parse_args()
{
while :; do
    case $1 in
		-h|-\?|--help)   # Call a "usage" function to display a synopsis, then exit.
			usage
			exit
			;;
        -a|--AdditionalFile)
			if [ -n "$2" ]; then
                handleAdditionalFile "$2"
                shift 2
                continue
            else
                printf 'ERROR: "--AdditionalFile" requires a non-empty option argument.\n' >&2
                exit 1
            fi
            ;;
		-a=*|--AdditionalFile=*)
			handleAdditionalFile "${1#*=}"
			shift # past argument=value
			;;
		-a=|--AdditionalFile=)         # Handle the case of an empty --AdditionalFile=
            printf 'ERROR: "--AdditionalFile" requires a non-empty option argument.\n' >&2
            exit 1
			;;            
        -b|--BS)
			if [ -n "$2" ]; then
                BOOTSCRIPT="$2"
                shift 2
                continue
            else
                printf 'ERROR: "--BS" requires a non-empty option argument.\n' >&2
                exit 1
            fi
            ;;
		-b=*|--BS=*)
			BOOTSCRIPT="${1#*=}"
			shift # past argument=value
			;;
		-b=|--BS=)         # Handle the case of an empty --BS=
            printf 'ERROR: "--BS" requires a non-empty option argument.\n' >&2
            exit 1
			;;
        -i|--Image)
			if [ -n "$2" ]; then
                IMAGE="$2"
                shift 2
                continue
            else
                printf 'ERROR: "--Image" requires a non-empty option argument.\n' >&2
                exit 1
            fi
            ;;
		-i=*|--Image=*)
			IMAGE="${1#*=}"
			shift # past argument=value
			;;
		-i=|--Image=)         # Handle the case of an empty --Image=
            printf 'ERROR: "--Image" requires a non-empty option argument.\n' >&2
            exit 1
			;;
		-o|--OS)
			if [ -n "$2" ]; then
                KERNEL="$2"
                shift 2
                continue
            else
                printf 'ERROR: "--OS" requires a non-empty option argument.\n' >&2
                exit 1
            fi
            ;;
		-o=*|--OS=*)
			KERNEL="${1#*=}"
			shift # past argument=value
			;;
		-o=|--OS=)         # Handle the case of an empty --OS=
            printf 'ERROR: "--OS" requires a non-empty option argument.\n' >&2
            exit 1
			;;
		-p|--ParamFile)
			if [ -n "$2" ]; then
                TMPFILE1=`mktemp`
                TMPFILE2=`mktemp`
                curl ${CURL_PREFIX}/${2} -o "${TMPFILE1}"
                tr -d '\r' < ${TMPFILE1} > ${TMPFILE2} && mv "${TMPFILE2}" "${TMPFILE1}"
                while read i ; do parse_args "$i" ; done < "${TMPFILE1}"
                rm "$TMPFILE1"
                shift 2
                continue
            else
                printf 'ERROR: "--Image" requires a non-empty option argument.\n' >&2
                exit 1
            fi
            ;;
		-p=*|--ParamFile=*)
            TMPFILE1=`mktemp`
            TMPFILE2=`mktemp`
            curl ${CURL_PREFIX}/${1#*=} -o "${TMPFILE1}"
            tr -d '\r' < ${TMPFILE1} > ${TMPFILE2} && mv "${TMPFILE2}" "${TMPFILE1}"
            while read i ; do parse_args "$i" ; done < "${TMPFILE1}"
            rm "$TMPFILE1"
			shift # past argument=value
			;;
		-p=|--ParamFile=)         # Handle the case of an empty --ParamFile=
            printf 'ERROR: "--ParamFile" requires a non-empty option argument.\n' >&2
            exit 1
			;;
		-r|--RPM)
			if [ -n "$2" ]; then
                RPM="$RPM $2"
                shift 2
                continue
            else
                printf 'ERROR: "--RPM" requires a non-empty option argument.\n' >&2
                exit 1
            fi
            ;;
		-r=*|--RPM=*)
			RPM="$RPM ${1#*=}"
			shift # past argument=value
			;;
		-r=|--RPM=)         # Handle the case of an empty --RPM=
            printf 'ERROR: "--RPM" requires a non-empty option argument.\n' >&2
            exit 1
			;;
		-t|--TFTP)
			if [ -n "$2" ]; then
                TFTPFILES="$TFTPFILES $2"
                shift 2
                continue
            else
                printf 'ERROR: "--TFTP" requires a non-empty option argument.\n' >&2
                exit 1
            fi
            ;;
		-t=*|--TFTP=*)
			TFTPFILES="$TFTPFILES ${1#*=}"
			shift # past argument=value
			;;
		-t=|--TFTP=)         # Handle the case of an empty --TFTP=
            printf 'ERROR: "--TFTP" requires a non-empty option argument.\n' >&2
            exit 1
			;;
		-f|--FS)
			if [ -n "$2" ]; then
                ROOTFS="$2"
                shift 2
                continue
            else
                printf 'ERROR: "--FS" requires a non-empty option argument.\n' >&2
                exit 1
            fi
            ;;
		-f=*|--FS=*)
			ROOTFS="${1#*=}"
			shift # past argument=value
			;;
		-f=|--FS=)         # Handle the case of an empty --FS=
            printf 'ERROR: "--FS" requires a non-empty option argument.\n' >&2
            exit 1
			;;
		--)              # End of all options.
			shift
			break
			;;
		*)
			break
			;;
	esac
done
}

unmountall()
{
	#umount /tmp/emmc/* if exists and mounted ( might happen if prior update was interupted )
	for mnt in $(mount | egrep '/mnt|mmcblk0p' | sed 's/\([^ ]*\) on \([^ ]*\).*/\2/' | sort -r)
	do 
		echo Unmounting $mnt
		umount $mnt || true
	done
	rmdir /mnt/* 2>/dev/null|| true
}

cleanup()
{
	echo Cleanup workspace...
	unmountall

	echo Update failed. 
	exit 1
}

check_fng_boot_version()
{
	fngboot_major=`cat /proc/cpuinfo | grep FNGBoot | sed 's/FNGBoot\s*:\s*\([0-9]\+\)\.\([0-9]\+\)r\([0-9]\+\)/\1/'`
	fngboot_minor=`cat /proc/cpuinfo | grep FNGBoot | sed 's/FNGBoot\s*:\s*\([0-9]\+\)\.\([0-9]\+\)r\([0-9]\+\)/\2/'`
	fngboot_revision=`cat /proc/cpuinfo | grep FNGBoot | sed 's/FNGBoot\s*:\s*\([0-9]\+\)\.\([0-9]\+\)r\([0-9]\+\)/\3/'`
	min_fngboot_major=`echo "$BOOTLOADER_MIN_VERSION" | sed 's/\([0-9]\+\)\.\([0-9]\+\)r\([0-9]\+\)/\1/'`
	min_fngboot_minor=`echo "$BOOTLOADER_MIN_VERSION" | sed 's/\([0-9]\+\)\.\([0-9]\+\)r\([0-9]\+\)/\2/'`
	min_fngboot_revision=`echo "$BOOTLOADER_MIN_VERSION" | sed 's/\([0-9]\+\)\.\([0-9]\+\)r\([0-9]\+\)/\3/'`

	echo "This release expects at least FNGBoot ${min_fngboot_major}.${min_fngboot_minor}r${min_fngboot_revision}"
	echo "your system has fngboot ${fngboot_major}.${fngboot_minor}r$fngboot_revision"

	if [ "$fngboot_major" -gt "$min_fngboot_major" ]; then return 0; fi
	if [ "$fngboot_minor" -gt "$min_fngboot_minor" ]; then return 0; fi
	if [ "$fngboot_revision" -ge "$min_fngboot_revision" ]; then return 0; fi

	echo "Your FNGBoot version is outdated."
	echo "Please update FNGBoot to version  ${min_fngboot_major}.${min_fngboot_minor}r${min_fngboot_revision} or newer."
	echo "Update failed."
	exit 1

}

mounttarget()
{
    mkdir ${TARGETROOT}
    mount /dev/mmcblk0p3 ${TARGETROOT}
	mount -t tmpfs none $ABSTARGETTMP
    mount -t proc proc ${TARGETROOT}/proc/
    mount -t sysfs sys ${TARGETROOT}/sys/
    mount -o bind /dev ${TARGETROOT}/dev/
}

unmounttarget()
{
    if [ -n $ABSTARGETTMP -a -d $ABSTARGETTMP ] ; then
        rm -rf $ABSTARGETTMP/* 
        sync
        umount $ABSTARGETTMP || true
    fi
    if [ -n ${TARGETROOT} -a -d ${TARGETROOT} ] ; then
        umount ${TARGETROOT}/proc/ || true
        umount ${TARGETROOT}/sys/ || true
        umount ${TARGETROOT}/dev/ || true
        umount ${TARGETROOT} || true
        rmdir ${TARGETROOT} || true
    fi
}

rpminstall()
{
    TARGETROOT="/mnt/target"
    TARGETTMP="/var/volatile/tmp"
    ABSTARGETTMP="${TARGETROOT}${TARGETTMP}"

    mounttarget

    cd $ABSTARGETTMP
    for f in $RPM; do
    sync
    curl "${CURL_PREFIX}/$f" -O 2>&1
    RPMBASES="$RPMBASES `basename $f`"
    done

    echo "#!/bin/bash" > $ABSTARGETTMP/rpminstall
    echo "cd ${TARGETTMP}" >> $ABSTARGETTMP/rpminstall
    echo "rpm -i $RPMBASES" >> $ABSTARGETTMP/rpminstall
    chmod a+x $ABSTARGETTMP/rpminstall

    chroot ${TARGETROOT} bash ${TARGETTMP}/rpminstall

    cd -
    unmounttarget
    unmountall
}


# Stop on errors
set -e

# Exit if not in FLASH-N-GO system
cat /etc/hostname | grep FLASH-N-GO > /dev/null || {
	echo "This script can be run from FLASH-N-GO only"
	exit 1
}

# Exit if not on an eMMC-based system
test -e /dev/mmcblk0boot1 || {
	echo "This script works for eMMC only"
	exit 1
}

# Exit if the FNGBoot version does not match
check_fng_boot_version

# Choose cURL prefix dependent on whether TFTP variable is set
if [ -z "$TFTP" ]
then
	FILEPATH=`cd $(dirname $0) && pwd`
	CURL_PREFIX="file://$FILEPATH"
else
	CURL_PREFIX="tftp://$TFTP"
fi

parse_args "$@"

set +e
# Test if the download works, attempt to load the beginning of the rootfs file
curl "${CURL_PREFIX}/${ROOTFS}" > /dev/null 2>&1 &
PID=$!
usleep 300000
kill $PID 2> /dev/null
wait $PID
ret=$?

# Successful return value is either 0 (file was downloaded completely) or 143
# (download was in progress and cancelled by "kill").
if [[ "x$ret" != "x0" ]] && [[ "x$ret" != "x143" ]]
then
	echo
	echo "ERROR: Download from \"${CURL_PREFIX}/${ROOTFS}\" failed."
	usage
	if [ -z "$TFTP" ]
	then
		echo "\"${ROOTFS}\" was not found in script directory."
		echo "Make sure that all files belonging to the release are contained in the script"
		echo "directory, or export TFTP variable to get files from TFTP server."
	else
		echo "Network connection does not seem to work."
		echo "Make sure to set the correct address and path with export TFTP=..., or unset"
		echo "TFTP to use local files."
	fi
	exit 1
fi
set -e

unmountall

trap cleanup EXIT
trap cleanup INT


# Partition sizes
KERNEL_SIZE=16

# Calculate offsets and sizes
EMMC_SIZE=$(parted /dev/mmcblk0 unit mib print | grep Disk |
			awk '{print $3}' | awk -F "M" '{print $1}')
FLASH_N_GO_SIZE=$(parted /dev/mmcblk0 unit mib print | grep " 1 " |
					awk '{print $4}' | awk -F "." '{print $1}')
ROOTFS_SIZE=$(expr $EMMC_SIZE - $FLASH_N_GO_SIZE - $KERNEL_SIZE)

# Create partitions
sfdisk --force -uM /dev/mmcblk0 << EOF
,${FLASH_N_GO_SIZE},b
,${KERNEL_SIZE},b
,${ROOTFS_SIZE},83
EOF

# Wait until device nodes are there
sleep 2

# Format all Yocto partitions
mkfs.vfat -n LINUX -F 16       /dev/mmcblk0p2
mkfs.${ROOTFS_TYPE} -L ROOTFS  /dev/mmcblk0p3

# Write rootfs partition
mkdir -p /tmp/emmc
mount /dev/mmcblk0p3 -t ${ROOTFS_TYPE} /tmp/emmc
echo "Loading ${ROOTFS} ..."
curl ${CURL_PREFIX}/${ROOTFS} | tar -C /tmp/emmc -xz
#unmount everything 
unmountall

# Get kernel via TFTP and write it to flash
echo "Loading ${KERNEL} ..."
mount /dev/mmcblk0p2 /tmp/emmc
curl ${CURL_PREFIX}/${KERNEL} > /tmp/emmc/${KERNEL}

# Write boot script to flash
echo "Loading ${BOOTSCRIPT} ..."
curl ${CURL_PREFIX}/${BOOTSCRIPT} > /tmp/emmc/boot.cfg

# Write device tree script to flash
if [ "${DEVICETREE}" == "" ];
then
    for DT in ${DEFAULT_DEVICETREE};
	do
		DEVICETREE="${DEVICETREE} ${DEFAULT_DEVICETREE_PREFIX}${DT}"
	done
fi

for DT in ${DEVICETREE};
do
    echo "Loading ${DT} ..."
    curl ${CURL_PREFIX}/${DT} > /tmp/emmc/${DT}
done

umount /tmp/emmc
rm -Rf /tmp/emmc
unmountall 
sync

# Install RPM files
if [ -n "$RPM" ] ; then
    echo "Try to install following RPM files: ${RPM} ..."
    rpminstall
fi

# Install arbitrary files (Format "<file>:<location>" where location can be a filename or an existing dir.
if [ -n "$TFTPFILES" ]; then
    emmc_mount
    for t in $TFTPFILES ;
    do
        #try to split by ':'
        FILE="${t%:*}"
        LOCATION="${t#*:}"
        if [ -n "$FILE" -a -n "$LOCATION" ]; then
            if [ -d "$LOCATION" ]; then
                cd  ${LOCATION}
                curl ${CURL_PREFIX}/${FILE} -O
                cd  -
            else
                curl ${CURL_PREFIX}/${FILE} -o ${LOCATION}
            fi
        fi
    done
    unmountall
    sync
fi

trap unmountall EXIT
echo "Update successful"
exit 0
