from install_script import Installer

from errors import FNGToolError
from errors import FNGPackageError

import time,os
import ntpath

class installer(Installer):
	"""
	This installs a Garz&Fricke Yocto release on an i.MX6 platform, using the
	Flash-N-Go Update application.
	
	These files are needed:
		GUF-Yocto-<VERSION>-<PLATFORM>-linuximage.bin
		GUF-Yocto-<VERSION>-<PLATFORM>-root.tar.gz
		GUF-Yocto-<VERSION>-<PLATFORM>-boot.cfg
		md5sums.txt

	The files can be found in the 'prebuilt_images' directory of G&F Linux releases
	"""
	MB=2**20
	def __init__(self, pkg, sys):
		super(installer, self).__init__(pkg, sys)

		self.current_platform = sys.information.cpuinfo['Hardware'].split()[3]
		self.files = {}

		found = False
		for fname in pkg:
			if 'linuximage' in fname and self.current_platform in fname:
				fstr = ntpath.basename(fname).rsplit('-',2) # naming scheme: GUF-Yocto-7.0-r3850-0-SANTARO-linuximage.bin
				self.pkg_version = fstr[0]
				self.pkg_platform = fstr[1]
				found = True
				break

		# is the package suitable for this system?
		if not found:
			raise FNGPackageError("Package %s is not compatible with platform %s"%(self.name, self.current_platform))
		if not sys.partitions('mmcblk0'):
			raise FNGPackageError("Package %s is meant for platforms with eMMC-Flash storage"%self.name)

		self.files['kernel'] = pkg.uri('%s-%s-linuximage.bin'%(
							self.pkg_version,
							self.pkg_platform))
		self.files['rootfs'] = pkg.uri('%s-%s-root.tar.gz'%(
							self.pkg_version,
							self.pkg_platform))
		self.files['bootcfg'] = pkg.uri('%s-%s-boot.cfg'%(
							self.pkg_version,
							self.pkg_platform))
		self.files['devicetree1'] = pkg.uri('%s-%s-imx6-santaro-dl.dtb'%(
							self.pkg_version,
							self.pkg_platform))
		self.files['devicetree2'] = pkg.uri('%s-%s-imx6-santaro-q.dtb'%(
							self.pkg_version,
							self.pkg_platform))
		self.files['devicetree3'] = pkg.uri('%s-%s-imx6-santoka-dl.dtb'%(
							self.pkg_version,
							self.pkg_platform))
		self.files['checksum'] = pkg.uri('md5sums.txt')

		# exclude files that are not needed for this package from md5 verification
		self.verify_whitelist = []
		infile = open(self.files['checksum'], 'rU')
		for line in infile:
			fname = line.split()[1]
			if not any(fname in s for s in self.files.values()):
				self.verify_whitelist.append(fname)

		self.name = self.pkg_version
		self.version = {}

		self.properties = {}
		self.properties['Platform'] = self.pkg_platform


	def prepare(self, ui, pkg, system):
		pass

	def install(self, ui, pkg, system):
		ui.task().set_min(0)
		ui.task().set_max(1)
		
		fng_part = None

		tsk = ui.subtask('verifying package...', 0.04)

		# raises FNGVerifyError when files are corrupted
		system.verify(self.files['checksum'], self.verify_whitelist, task=tsk)

		self.milestone('verified package', 0.04)

		#delete all partitions except the first (that is the fng-system)
		for p in system.partitions('mmcblk0'):
			if p.number == 1:
				fng_part = p
			else:
				p.remove()
				self.milestone('deleted partition %s'%p.number, 0.04)
		self.milestone('deleted partitions', 0.08)

		kernel_size = 16*self.MB
		if fng_part == None:
			# No FnG-system partition found, assuming 32MB at start of disk
			kernel_start = 32*self.MB
		else:
			kernel_start = fng_part.end+1
		
		kernel_part = system.partitions('mmcblk0').create(
			start = kernel_start,
			end   = kernel_start + kernel_size
		)
		self.milestone('created new partition "kernel" with %d MB'%(kernel_part.size/self.MB), 0.12)

		rootfs_part = system.partitions('mmcblk0').create(
			start = kernel_part.end+1,
			end   = system.partitions('mmcblk0').props['size']-1,
		)
		self.milestone('created new partition "rootfs" with %d MB'%(rootfs_part.size/self.MB), 0.16)

		# wait for new partition-nodes to become available:
		time.sleep(2)

		self.milestone('creating vfat filesystem for kernel...', 0.16)
		kernel_part.format('vfat', 'LINUX')
		self.milestone('filesystem created', 0.2)
		self.milestone('creating @ROOTFS_TYPE@ filesystem for rootfs... (this might take some time)', 0.2)
		rootfs_part.format('@ROOTFS_TYPE@', 'ROOTFS')
		self.milestone('filesystem created', 0.24)

		tsk = ui.subtask('installing rootfs to eMMC-flash', 0.9)
		ret = system.tar2part(self.files['rootfs'], rootfs_part, compression='gzip', task=tsk, estim_files=10000)
		
		tsk = ui.subtask('installing kernel to eMMC-flash', 1.0)
		kernel_mnt = kernel_part.mount()
		ret = system.mmc_copy(self.files['kernel'], kernel_mnt+'/linuximage', task=tsk)
		ret = system.mmc_copy(self.files['bootcfg'], kernel_mnt+'/boot.cfg', task=tsk)
		ret = system.mmc_copy(self.files['devicetree1'], kernel_mnt+'/imx6-santaro-dl.dtb', task=tsk)
		ret = system.mmc_copy(self.files['devicetree2'], kernel_mnt+'/imx6-santaro-q.dtb', task=tsk)
		ret = system.mmc_copy(self.files['devicetree3'], kernel_mnt+'/imx6-santoka-dl.dtb', task=tsk)
		self.milestone('created boot-configuration file', 1.0)
		kernel_part.umount()

		self.milestone('installation complete', 1)

		ui.task().done()
