// gfeeprom.c : Defines the entry point for the application.
//
/*

Garz und Fricke gfeeprom tool
=======================================

The gfeeprom tool is a commandline tool to access the eeprom on guf platforms.
The eeprom insn't formated with a real file system, but so called areas.
An area is a plain data region of a fixed size address by a number.

The area 0 is reserved for a special data structure containing the following data:
'SDC-CSPU' 
article number(s): '900-2842'
serial number(s): '01569671'
major: '1'
minor: '0'
comment: 'No comment'

This data is written during the test sequence after production.

The area 1 is reserved.
The areas above 50 are ment for customer usage, while the areas below 50 are meant for system usage.

The operating hours counter uses the areas 2 and 3.


The gfeeprom command line interface:

gfeeprom version 1.3 [$Revision: 4590 $] (Feb 11 2015) <grimm@garz-fricke.com>
Options:
  --clean [-d <device file>] [-fc]
  --show [-d <device file>] [-fc] [-bd | -hs ] [-df <filen name>] [-ea <area number>]
  --program [-d <device file>] [-ea <area number>] [-sf <source file>] [-ss <string>] [ -hs]
  --version

General suboptions:
	-d Use -d for not using the standard file '/dev/gfeeprom/bus0eepromA0'. Normally not needed.
	-q Use -q for quiet mode, this suppress all output except for the data, 
				errors can only be detected using the return value.


Option --clean: 
	Resets an EEPROM content to empty.
		Use -fc for cleaning the full eeprom. If not set only all areas above area id 50 will be removed.

Option --show 
    Shows the available EEPROMs equiped on platform.
	     Use -d for showing the GuF content (serial number...) of a specific EEPROM device.
         Use -fc for showing the full content of a specific EEPROM device.
         Use -df for writing the full content of a specific EEPROM device to the specified file.
         Use -bd for dumping the binary content of the EEPROM area.
         Use -hs for dumping the binary content as hex number: for example: gfeeprom --show -ea 2 -hs -> 4f48430021000000
         Use -ea for getting information of a specific EEPROM area.

Option --program
        Programs the GuF area or a custom area. If the erea is present it will be overwritten, if not the area will be created
        Use -ea for setting information of this EEPROM area.
		Use -hs use hex numbers as data format: 4f48430021000000
        Use -sf for loading the EEPROM area content from this file
		Use -ss for using the following command line string as raw data packet.

Note: changing the size of an area will erase the areas behind it, so it is advisable to keep the size of the areas the same.
Note: the format options are not available for erea 0. To write area 0 use the following:

echo "[BEGIN]
Controller Index = 1
Device = /dev/gfeeprom/bus0eepromA0
EEPROM Address = 0xA0
Major Version = 1
Minor Version = 3
Article Number = 900-2907R
Serial Number = 01589746
Component = Santaro
Comment = No comment
Verify = 0x1
Force write = 0x1
[END]" > eeprom.txt

gfeeprom --program -sf eeprom.txt


Call gfeeprom from other programms:

The best way to call from other programs is to pass the data on the commands line in the hex number format.

Write the data "Hello" to area 50:
gfeeprom --program -q -ea 50 -hs -ss 48656c6c6f0a

To read the data again use:
gfeeprom --show -q -ea 50 -hs
48656c6c6f0a

*/


#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>
#include <dirent.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <stdint.h>
#include <linux/errno.h>
#include <linux/i2c/gfeeprom.h>

#include "Version.h"
#if 0
//FIXME defined in linux header
#define ERROR_SUCCESS                    0
#define ERROR_PATH_NOT_FOUND             3
#define ERROR_NOT_ENOUGH_MEMORY          8
#define ERROR_INVALID_DATA               13
#define ERROR_CRC                        23
#define ERROR_BAD_ARGUMENTS              160
#define ERROR_INSUFFICIENT_BUFFER        122
#define ERROR_NO_VOLUME_LABEL            125
#define ERROR_BAD_ARGUMENTS              160
#define ERROR_I2C_FAILURE				 244
#define ERROR_LIST_INCONSISTENT          255
#define ERROR_MOD_NOT_FOUND              126
#define ERROR_EEPROM_HEAD_READ_FAILURE	 275
#define ERROR_OLD_VERSION            	 1150
#endif

#define _QUOTEME(x) #x
#define QUOTEME(x) _QUOTEME(x)

#define GF_SVN_VERSION "$Revision: 4590 $"
#define TOOL_NAME "gfeeprom"

#define ERROR_NOT_IMPLEMENTED -10
#define ERROR_AREA_NOT_FOUND  -11
#define ERROR_UNKNOWN_ARG	  -12
#define ERROR_HWINFO_SIZE_MISMATCH -13
#define ERROR_NO_CUSTOM_AREAS -14

#define STANDARD_EEPROM_FILE_NAME "bus0eepromA0"
#define STANDARD_EEPROM_PATH "/dev/gfeeprom"
#define STANDARD_EEPROM_FILE STANDARD_EEPROM_PATH "/" STANDARD_EEPROM_FILE_NAME

#define GFEEPROM_OP_CLEAN 			"--clean"
#define GFEEPROM_OP_SHOW 			"--show"
#define GFEEPROM_OP_PROGRAM 		"--program"
#define GFEEPROM_OP_VERSION 		"--version"
#define GFEEPROM_OP_HELP 			"--help"
#define GFEEPROM_OP_DEVICE			"-d"
#define GFEEPROM_OP_FULL_CONTENT	"-fc"
#define GFEEPROM_OP_EEPROM_EA		"-ea"
#define GFEEPROM_OP_BIN_DUMP		"-bd"
#define GFEEPROM_OP_HEX_STRING		"-hs"
#define GFEEPROM_OP_BIN_DUMP_TO_FILE "-df"
#define GFEEPROM_OP_BIN_SRC_FILE	"-sf"
#define GFEEPROM_OP_BIN_SRC_STRING	"-ss"
#define GFEEPROM_OP_QUIET			"-q"

#define DUMP_FULL_CONTENT			0xFFFFFFFF

#define PRINTF(s) gf_print("%s\n", s)

static bool quiet = false;
void gf_print( const char * fmt, ...)
{
	va_list args;

	if(quiet) return;
	
	va_start(args, fmt);
    vprintf(fmt, args);
    va_end(args);
}

static const char wszVersion[] = {
	TOOL_NAME " version " MAJOR_VERSION "." MINOR_VERSION " [" GF_SVN_VERSION "] (" __DATE__ ") <grimm@garz-fricke.com>\n"
};
static const char wszHelp[] =
{
	"usage:\n"
	"  " GFEEPROM_OP_CLEAN " [" GFEEPROM_OP_DEVICE " <device file>] [" GFEEPROM_OP_FULL_CONTENT "]\n"
	"                                        Resets an EEPROM content to empty.\n"
	"                                          Use " GFEEPROM_OP_DEVICE " for not using\n"
	"                                            the standard file '" STANDARD_EEPROM_FILE "'.\n"
	"                                          Use " GFEEPROM_OP_FULL_CONTENT " for cleaning\n"
	"                                            the full eeprom. If not set only all areas above\n"
	"                                            area id " QUOTEME(GFEEPROM_MIN_AREA_TYPE_NUMBER) " will be removed.\n"
	"  " GFEEPROM_OP_SHOW " [" GFEEPROM_OP_DEVICE " <device file>] [" GFEEPROM_OP_FULL_CONTENT "] [" GFEEPROM_OP_BIN_DUMP " | " GFEEPROM_OP_HEX_STRING " ] [" GFEEPROM_OP_BIN_DUMP_TO_FILE " <filen name>] [" GFEEPROM_OP_EEPROM_EA " <area number>]\n"
	"                                        Shows the available EEPROMs equiped on platform.\n"
	"                                          Use " GFEEPROM_OP_DEVICE " for showing the\n"
	"                                            GuF content of a specific EEPROM device.\n"
	"                                          Use " GFEEPROM_OP_FULL_CONTENT " for showing\n"
	"                                            the full content of a specific EEPROM device.\n"
	"                                          Use " GFEEPROM_OP_BIN_DUMP_TO_FILE " for writing\n"
	"                                            the full content of a specific EEPROM device to the specified file.\n"
	"                                          Use " GFEEPROM_OP_BIN_DUMP " for dumping\n"
	"                                            the binary content of the EEPROM area.\n"
	"                                          Use " GFEEPROM_OP_EEPROM_EA " for getting \n"
	"                                            information of a specific EEPROM area.\n"
	"  " GFEEPROM_OP_PROGRAM " [" GFEEPROM_OP_DEVICE " <device file>] [" GFEEPROM_OP_EEPROM_EA " <area number>] [" GFEEPROM_OP_BIN_SRC_FILE " <source file>] [" GFEEPROM_OP_BIN_SRC_STRING " <string>] [ " GFEEPROM_OP_HEX_STRING " ]\n"
	"                                        Programs the GuF area or a custom area. If the erea is present\n"
	"                                        the are will overwritten if not the area will be created\n"
	"                                          Use " GFEEPROM_OP_DEVICE " for showing the\n"
	"                                            GuF content of a specific EEPROM device.\n"
	"                                          Use " GFEEPROM_OP_EEPROM_EA " for setting \n"
	"                                            information of this EEPROM area.\n"
	"                                          Use " GFEEPROM_OP_BIN_SRC_FILE " for loading \n"
	"                                            the EEPROM area content from this file.\n"
	"                                          Use " GFEEPROM_OP_BIN_SRC_STRING " for using \n"
	"                                            the following command line string as raw data packet.\n"
	"  " GFEEPROM_OP_VERSION "\n"
	"                                       Print Version\n"
	"  " GFEEPROM_OP_HELP "\n"
	"                                       Print this help.\n"
};
static const char wszUsage[] =
{
	"Options:\n"
	"  " GFEEPROM_OP_CLEAN " [" GFEEPROM_OP_DEVICE " <device file>] [" GFEEPROM_OP_FULL_CONTENT "]\n"
	"  " GFEEPROM_OP_SHOW " [" GFEEPROM_OP_DEVICE " <device file>] [" GFEEPROM_OP_FULL_CONTENT "] [" GFEEPROM_OP_BIN_DUMP " | " GFEEPROM_OP_HEX_STRING " ] [" GFEEPROM_OP_BIN_DUMP_TO_FILE " <filen name>] [" GFEEPROM_OP_EEPROM_EA " <area number>]\n"
	"  " GFEEPROM_OP_PROGRAM " [" GFEEPROM_OP_DEVICE " <device file>] [" GFEEPROM_OP_EEPROM_EA " <area number>] [" GFEEPROM_OP_BIN_SRC_FILE " <source file>] [" GFEEPROM_OP_BIN_SRC_STRING " <string>] [ " GFEEPROM_OP_HEX_STRING "]\n"
	"  " GFEEPROM_OP_VERSION "\n"
	"  " GFEEPROM_OP_HELP
};

static void printVersion(void)
{
	gf_print(wszVersion);
}
static void printHelp(void)
{
	printVersion();
	gf_print(wszHelp);
}
static void printUsage(void)
{
	printVersion();
	gf_print(wszUsage);
}

static int get_area_no_from_binary(char*data)
{
	return data[0] | data[1] << 8;
}
static int get_area_length_from_binary(char*data)
{
	return 	 data[2] << 0
			|data[3] << 8
			|data[4] <<16 
			|data[5] <<24;
}
static int hexNibbleToVal(char c)
{
	int v = 0;
	if( c >= '0' && c <= '9' )
		v = c - '0';
	else if( c >= 'A' && c <= 'F' )
		v = c - 'A' + 10;
	else if( c >= 'a' && c <= 'f' )
		v = c - 'a' + 10;
	return v;
}


static int clean(char* device_name, bool full_content)
{
	int hDev;
	int ret;
	struct gfeeprom_access ea;

	hDev = open(device_name, O_RDWR);
	if(hDev < 0)
	{
		gf_print(TOOL_NAME " couldn't open device %s because of error %u.\n", device_name, errno);
		return -errno;
	}

	ea.size_of_struct = sizeof(struct gfeeprom_access);
	ea.access_type = full_content ? EEPROM_ACCESS_TYPE_CLEAN : EEPROM_ACCESS_TYPE_DELETE_ALL;
	ea.force_write = false;
	ea.data = NULL;
	ea.data_length = 0;
	ea.verify_buffer = NULL;
	ea.verify_buffer_length = 0;

	ret = ioctl(hDev, GFEEPROM_IOCTL_ACCESS_EEPROM_AREA, &ea);
	if (ret < 0)
	{
		gf_print(TOOL_NAME " couldn't %s the EEPROM. Error: 0x%X\n", full_content ? "clean" : "remove the custom areas from" , errno);
		close(hDev);
		return ret;
	}
	close(hDev);
	gf_print(TOOL_NAME " successful %s the EEPROM %s.\n", full_content ? "cleaned" : "removed the custom areas from", device_name);
	return ERROR_SUCCESS;
}

static int show_area(int hDev, u32 eara_number, bool bin_dump, bool hex_string, char* dump_file)
{
	ssize_t read_bytes;
	char* data_ptr;
	u16 length_buffer;
	int ret;
	int i, j;

	read_bytes = pread(hDev, &length_buffer, sizeof(u16), eara_number);
	if (!((-1 == read_bytes) && (EFBIG == errno) && length_buffer))
	{
		gf_print("  No custom area %u found on EEPROM.\n", eara_number);
		ret = -ENXIO;
		return ret;
	}

	gf_print("    Found Area: %u [%u Byte%s] %s\n", eara_number, length_buffer, (length_buffer > 1) ? "s" : "", (bin_dump && !dump_file) ? "dumping binary:" : "");

	data_ptr = (char*)malloc(sizeof(char)*length_buffer);
	if (!data_ptr)
	{
		gf_print("      Failed allocating enough RAM for EEPROM area.\n");
		ret = -ENOMEM;
		goto error;
	}
	read_bytes = pread(hDev, data_ptr, sizeof(char)*length_buffer, eara_number);
	if (read_bytes != length_buffer)
	{
		gf_print("      Failed reading the complete EEPROM area. Only %u byte%s read.\n", read_bytes, (read_bytes > 1) ? "s" : "");
		ret = -EPIPE;
		goto error_free;
	}
	ret = ERROR_SUCCESS;
	if( hex_string)
	{
		for(i = 0; i < length_buffer; i++)
			printf("%02x",data_ptr[i]);
		printf("\n");
	}
	else if (bin_dump && dump_file)
	{
		int hdump;

		hdump = open(dump_file, O_WRONLY | O_CREAT | O_EXCL);
		if(hdump >= 0)
		{
			ret =write(hdump, data_ptr, length_buffer);
			if( ret < 0 )
				gf_print("      Failed writing dump file '%s'. Error %u.\n", dump_file, errno);
			else
				gf_print("      Successful dumped EEPROM area to file '%s'.\n", dump_file);
			close(hdump);
		}
		else
		{
			gf_print("      Failed opening dump file '%s'. Error %u.\n", dump_file, errno);
			ret = ENXIO;
		}
	}
	else if(bin_dump)
	{
		for (i = 0; i < (length_buffer>>4); i++)
		{
			printf("      0x%04X : 0x%02X 0x%02X 0x%02X 0x%02X  0x%02X 0x%02X 0x%02X 0x%02X  0x%02X 0x%02X 0x%02X 0x%02X  0x%02X 0x%02X 0x%02X 0x%02X\n",
				(i<<4),
				data_ptr[(i<<4)], data_ptr[(i<<4)+1], data_ptr[(i<<4)+2], data_ptr[(i<<4)+3],
				data_ptr[(i<<4)+4], data_ptr[(i<<4)+5], data_ptr[(i<<4)+6], data_ptr[(i<<4)+7],
				data_ptr[(i<<4)+8], data_ptr[(i<<4)+9], data_ptr[(i<<4)+10], data_ptr[(i<<4)+11],
				data_ptr[(i<<4)+12], data_ptr[(i<<4)+13], data_ptr[(i<<4)+14], data_ptr[(i<<4)+15]);
		}
		if (length_buffer & 0xF)
		{
			printf("      0x%04X :", (i<<4));
			for (j = 0; j < (length_buffer & 0xF); j++)
			{
				if (j && !(j & 0x3))
					printf(" ");
				printf(" 0x%02X", data_ptr[(i<<4)+j]);
			}
			printf("\n");
		}
	}
	else{
		gf_print( "Not supported, use -bd flag to get a binary dump\n");
		ret = ERROR_NOT_IMPLEMENTED;
		goto error_free;
	}

error_free:
	free(data_ptr);
error:
	return ret;

}

static int show_eeprom(char* szFileName, u32 eara_number, bool bin_dump, bool hex_string, char* dump_file)
{
	int hDev;
	ssize_t read_bytes;
    struct dirent *ep;
    struct gfeeprom_hw_normal_version_data hw_info;
	int ret = ERROR_SUCCESS;

	hDev = open(szFileName, O_RDONLY);
	if(hDev < 0)
	{
		gf_print(" failed to open EEPROM '%s' with error: %d.\n", szFileName, errno);
		return hDev;
	}

	if (!bin_dump && (!eara_number || (DUMP_FULL_CONTENT == eara_number)))
	{
		read_bytes = pread(hDev, &hw_info, sizeof(struct gfeeprom_hw_normal_version_data), EEPROM_TYPE_EHW);
		if(read_bytes <= sizeof(struct gfeeprom_hw_small_version_data))
		{
			gf_print(" Reading hardware info from EEPROM '%s' failed.\n", szFileName);
		}
		else
		{
			if (sizeof(struct gfeeprom_hw_normal_version_data) == hw_info.size_of_struct)
			{
				gf_print("  EEPROM: %s -> component: '%s' | article number(s): '%s' | serial number(s): '%s' | major: '%s' | minor: '%s' | comment: '%s'\n",
					szFileName,
					hw_info.component,
					hw_info.article_number,
					hw_info.serial_numbers,
					hw_info.major,
					hw_info.minor,
					hw_info.comment);
			}
			else if (sizeof(struct gfeeprom_hw_small_version_data) == hw_info.size_of_struct)
			{
				struct gfeeprom_hw_small_version_data* hw_info_small = (struct gfeeprom_hw_small_version_data*)&hw_info;

				gf_print("  EEPROM: %s -> component: '%s' | article number(s): '%s' | major: '%s' | minor: '%s' | comment: '%s'\n",
					szFileName,
					hw_info_small->component,
					hw_info_small->article_number,
					hw_info_small->major,
					hw_info_small->minor,
					hw_info_small->comment);
			}
			else{
				gf_print("  Skipping EEPROM: %s because the hardware info struct size isn't matching.\n", szFileName);
				ret = ERROR_HWINFO_SIZE_MISMATCH;
				goto error;
			}
		}
	}

	if (DUMP_FULL_CONTENT == eara_number)
	{
		int i;
		u16 areas = 0;

		if (ioctl(hDev, GFEEPROM_IOCTL_GET_CUSTOM_AREA_COUNT, &areas, sizeof(u16)) < 0)
		{
			gf_print("  Failed getting area count scanning manually.\n");
			gf_print("  Dumping custom area(s) of EEPROM content:\n");
			for (i = EEPROM_TYPE_EHW+1; i < 0xFFFF; i++)				
				ret |= show_area(hDev, i, bin_dump, hex_string, NULL);
		}
		else if (areas)
		{
			struct gfeeprom_area_array param;
			u16* used_area_list = NULL;

			used_area_list = malloc(sizeof(u16)*areas);
			if (!used_area_list)
			{
				gf_print("     Failed allocating enough RAM for EEPROM area list.\n");
				ret = ERROR_NOT_ENOUGH_MEMORY;
				goto error;
			}
			param.size_of_struct = sizeof(struct gfeeprom_area_array);
			param.elements = areas;
			param.areas = used_area_list;
			if (ioctl(hDev, GFEEPROM_IOCTL_GET_USED_CUSTOM_AREAS, &param, sizeof(struct gfeeprom_area_array)) < 0)
			{
				gf_print("     failed to open retrieve used area list with error: %d.\n", errno);
				free(used_area_list);
				ret = ERROR_INVALID_DATA;
				goto error;
			}
			gf_print("  Dumping %u custom area%s of EEPROM content:\n", areas, (areas > 1) ? L"s" : L"");
			for (i = 0; i < areas; i++)
				ret |= show_area(hDev, used_area_list[i], bin_dump, hex_string, NULL);
			free(used_area_list);
		}
		else
		{
			gf_print("  No additional custom area stored on EEPROM.\n");
			ret = ERROR_NO_CUSTOM_AREAS;
			goto error;
		}
	}
	else if (eara_number)
	{
		ret = show_area(hDev, eara_number, bin_dump, hex_string, dump_file);
		if ( ret < 0)
		{
			goto error;
		}
	}
error:
	close(hDev);

	return ret;
}

static int show_all(void)
{
	int hDev;
	ssize_t read_bytes;
    DIR *dp;
    struct dirent *ep;
    struct gfeeprom_hw_normal_version_data hw_info;
    char file_name_buffer[256];

	dp = opendir(STANDARD_EEPROM_PATH);
	if (dp == NULL)
	{
		gf_print(TOOL_NAME ", couldn't open eeprom path %s\n", STANDARD_EEPROM_PATH);
		return ERROR_PATH_NOT_FOUND;
	}

	gf_print(TOOL_NAME " the following EEPROMs are present in the system:\n");
	while (ep = readdir(dp))
	{
		if ((ep->d_name[0] != '\0') && strcmp(ep->d_name, ".") && strcmp(ep->d_name, ".."))
		{
			strcpy(file_name_buffer, STANDARD_EEPROM_PATH);
			strcat(file_name_buffer, "/");
			strcat(file_name_buffer, ep->d_name);
			gf_print("Try to open EEPROM: %s ...", file_name_buffer);
			hDev = open(file_name_buffer, O_RDONLY);
			if(hDev < 0)
			{
				gf_print(" failed with error: %d.\n", errno);
				continue;
			}
			gf_print(" successful.\n");

			gf_print("reading hardware info from EEPROM: %s...", file_name_buffer);
			read_bytes = pread(hDev, &hw_info, sizeof(struct gfeeprom_hw_normal_version_data), EEPROM_TYPE_EHW);
			if(read_bytes <= sizeof(struct gfeeprom_hw_small_version_data))
			{
				gf_print(" failed.\n");
				close(hDev);
				continue;
			}
			gf_print(" successful.\n");

			if (sizeof(struct gfeeprom_hw_normal_version_data) == hw_info.size_of_struct)
			{
				gf_print("  EEPROM: %s -> component: '%s' | article number(s): '%s' | serial number(s): '%s' | major: '%s' | minor: '%s' | comment: '%s'\n",
					file_name_buffer,
					hw_info.component,
					hw_info.article_number,
					hw_info.serial_numbers,
					hw_info.major,
					hw_info.minor,
					hw_info.comment);
			}
			else if (sizeof(struct gfeeprom_hw_small_version_data) == hw_info.size_of_struct)
			{
				struct gfeeprom_hw_small_version_data* hw_info_small = (struct gfeeprom_hw_small_version_data*)&hw_info;

				gf_print("  EEPROM: %s -> component: '%s' | article number(s): '%s' | major: '%s' | minor: '%s' | comment: '%s'\n",
					file_name_buffer,
					hw_info_small->component,
					hw_info_small->article_number,
					hw_info_small->major,
					hw_info_small->minor,
					hw_info_small->comment);
			}
			else
				gf_print("  Skipping EEPROM: %s because the hardware info struct size isn't matching.\n", file_name_buffer);
			close(hDev);
		}
	}

	closedir (dp);
	return ERROR_SUCCESS;
}

static int program_guf_hw(char* data_file)
{
	int								hFile = -1;
	int 	  						hDev = -1;
	struct stat						f_stat;
	int	 							dwFileSize = 0;
	char							szBuffer[GFV_COMMENT_LENGTH+1];
	char* 							szEEPROMConfig;
	char* 							szToken;
	u_int32_t 						dwBytesRead;
	u_int32_t  						uiLineNr;
	struct gfeeprom_program_area	hw_info;
	bool							bInEEPROMSection = false;
	u_int32_t						uiEEPROMCounter = 0;

	hFile = open(data_file, O_RDWR);
	if(hFile < 0)
	{
		gf_print(TOOL_NAME " error opening EEPROM config file %s. (0x%X)\n", data_file, errno);
		return -5;
	}

	dwFileSize = fstat(hFile, &f_stat);
	if (-1 == dwFileSize)
	{
		gf_print(TOOL_NAME " error couldn't retrieve file attributes.\n");
		close(hFile);
		return -5;
	}

	szEEPROMConfig = (char*) malloc(f_stat.st_size);
	if (!szEEPROMConfig)
	{
		gf_print(TOOL_NAME " error couldn't allocate memory.\n");
		close(hFile);
		return -5;
	}

	if(!read(hFile, szEEPROMConfig, sizeof(char)*f_stat.st_size))
	{
		gf_print(TOOL_NAME " error reading EEPROM config file %s. (0x%X)\n", data_file, errno);
		close(hFile);
		return -6;
	}

	szToken = strtok(szEEPROMConfig, "\n" );
	uiLineNr = 0;
	while( szToken != NULL )
	{
		uiLineNr++;

		if(strstr(szToken, "[BEGIN]") == szToken)
		{
			bInEEPROMSection = true;
			hw_info.size_of_struct = sizeof(struct gfeeprom_program_area);
			hw_info.verify = false;
			hw_info.force_write = false;
			hw_info.hw_version.size_of_struct = sizeof(struct gfv_hw_version_data);
			hw_info.hw_version.major[0] = '\0';
			hw_info.hw_version.minor[0] = '\0';
			hw_info.hw_version.article_number[0] = '\0';
			hw_info.hw_version.serial_numbers[0] = '\0';
			hw_info.hw_version.component[0] = '\0';
			hw_info.hw_version.comment[0] = '\0';
			hw_info.hw_version.address_of_component = 0xFFFFFFFF;
			hw_info.hw_version.phys_mem_size = 0;
		}
		else if (strstr(szToken, "[END]") == szToken)
		{
			if (bInEEPROMSection)
			{
				bInEEPROMSection = false;

				if(hDev >= 0)
				{
					if (ioctl(hDev, GFEEPROM_IOCTL_PROGRAM_HW_INFO, &hw_info, sizeof(struct gfeeprom_program_area)) < 0)
					{
						gf_print(TOOL_NAME " EEPROM %d: error couldn't write the EEPROM information. Error: 0x%X\n", ++uiEEPROMCounter, errno);
					}
					else
					{
						gf_print(TOOL_NAME " EEPROM %d: data has been flashed successful.\n\n", ++uiEEPROMCounter);
					}
					close(hDev);
				}
			}
			else
			{
				gf_print(TOOL_NAME " error encountered [END] before begin at line %d\n", uiLineNr);
				close(hFile);
				return -7;
			}
		}
		else if(strstr(szToken, "Device") == szToken)
		{
			if (bInEEPROMSection)
			{
				if(1 != sscanf(szToken, "Device = %" QUOTEME(GFV_COMMENT_LENGTH) "s", szBuffer))
				{
					gf_print(TOOL_NAME " error in EEPROM config file line %d\n", uiLineNr);
					close(hFile);
					return -7;
				}
				hDev = open(szBuffer, O_RDWR);
				if(hDev < 0)
					gf_print(TOOL_NAME " couldn't open device %s.\n", szBuffer);
			}
		}
		else if(strstr(szToken, "Major Version") == szToken)
		{
			if (bInEEPROMSection)
			{
				if(1 != sscanf(szToken, "Major Version = %" QUOTEME(GFV_NUMBER_LENGTH) "s", hw_info.hw_version.major))
				{
					gf_print(TOOL_NAME " error in EEPROM config file line %d\n", uiLineNr);
					close(hFile);
					return -7;
				}
			}
		}
		else if(strstr(szToken, "Minor Version") == szToken)
		{
			if (bInEEPROMSection)
			{
				if(1 != sscanf(szToken, "Minor Version = %" QUOTEME(GFV_NUMBER_LENGTH) "s", hw_info.hw_version.minor))
				{
					gf_print(TOOL_NAME " error in EEPROM config file line %d\n", uiLineNr);
					close(hFile);
					return -7;
				}
			}
		}
		else if(strstr(szToken, "Article Number") == szToken)
		{
			if (bInEEPROMSection)
			{
				if(1 != sscanf(szToken, "Article Number = %" QUOTEME(GFV_ARTICLE_NUMBER_LENGTH) "s", hw_info.hw_version.article_number))
				{
					gf_print(TOOL_NAME " error in EEPROM config file line %d\n", uiLineNr);
					close(hFile);
					return -7;
				}
			}
		}
		else if(strstr(szToken, "Serial Number") == szToken)
		{
			if (bInEEPROMSection)
			{
				if(1 != sscanf(szToken, "Serial Number = %" QUOTEME(GFV_SERIAL_NUMBER_LENGTH) "s", hw_info.hw_version.serial_numbers))
				{
					gf_print(TOOL_NAME " error in EEPROM config file line %d\n", uiLineNr);
					close(hFile);
					return -7;
				}
			}
		}
		else if(strstr(szToken, "Component") == szToken)
		{
			if (bInEEPROMSection)
			{
				if(1 != sscanf(szToken, "Component = %" QUOTEME(GFV_COMPONENT_LENGTH) """[^\n]*1[\n]", hw_info.hw_version.component))
				{
					gf_print(TOOL_NAME " error in EEPROM config file line %d\n", uiLineNr);
					close(hFile);
					return -7;
				}
			}
		}
		else if(strstr(szToken, "Comment") == szToken)
		{
			if (bInEEPROMSection)
			{
				if(1 != sscanf(szToken, "Comment = %" QUOTEME(GFV_COMMENT_LENGTH) """[^\n]*1[\n]", hw_info.hw_version.comment) )
				{
					gf_print(TOOL_NAME " error in EEPROM config file line %d\n", uiLineNr);
					close(hFile);
					return -7;
				}
			}
		}
		else if(strstr(szToken, "Physical Address of component") == szToken)
		{
			if (bInEEPROMSection)
			{
				if(1 != sscanf(szToken, "Physical Address of component = 0x%x", &(hw_info.hw_version.address_of_component)))
				{
					gf_print(TOOL_NAME " error in EEPROM config file line %d\n", uiLineNr);
					close(hFile);
					return -7;
				}
			}
		}
		else if(strstr(szToken, "Physical memory size of component") == szToken)
		{
			if (bInEEPROMSection)
			{
				if(1 != sscanf(szToken, "Physical memory size of component = 0x%x", &(hw_info.hw_version.phys_mem_size)))
				{
					gf_print(TOOL_NAME " error in EEPROM config file line %d\n", uiLineNr);
					close(hFile);
					return -7;
				}
			}
		}
		else if(strstr(szToken, "Verify") == szToken)
		{
			if (bInEEPROMSection)
			{
				if(1 != sscanf(szToken, "Verify = 0x%x", &(hw_info.verify)))
				{
					gf_print(TOOL_NAME " error in EEPROM config file line %d\n", uiLineNr);
					close(hFile);
					return -7;
				}
			}
		}
		else if(strstr(szToken, "Force write") == szToken)
		{
			if (bInEEPROMSection)
			{
				if(1 != sscanf(szToken, "Force write = 0x%x", &(hw_info.force_write)))
				{
					gf_print(TOOL_NAME " error in EEPROM config file line %d\n", uiLineNr);
					close(hFile);
					return -7;
				}
			}
		}
		szToken = strtok(NULL, "\n" );
	}
	close(hFile);
	return ERROR_SUCCESS;
}

static int program(char* device_name, u32 area_number, char* data, int length)
{
	int 	  hDev;
	size_t	  size = length;
	int ret	= ERROR_SUCCESS;

	hDev = open(device_name, O_RDWR);
	if(hDev < 0)
	{
		gf_print(TOOL_NAME " couldn't open device %s.\n", device_name);
		return hDev;
	}

	if (-1 == pwrite(hDev, data, size, area_number))
	{
		gf_print(TOOL_NAME " failed writing area %u to device %s with error %u.\n", area_number, device_name, errno);
		ret =  -errno;
	}
	close(hDev);
	return ERROR_SUCCESS;
}

static int read_stdin( char ** buf, int * len)
{
	int ret = ERROR_SUCCESS;
	int buffer_len = 128; // for now
	int buffer_read = 0;
	int n;
	char * buffer = malloc(sizeof(char) * buffer_len);

	if(!buffer)
	{
		ret = ERROR_NOT_ENOUGH_MEMORY;
		goto error;
	}
	*buf = buffer;

	do{
		while((n = read( STDIN_FILENO, &buffer[buffer_read], buffer_len - buffer_read - 1) ) > 0)
		{
			buffer_read += n;
			if(0 == buffer_len - buffer_read - 1) break;
		}
		if(n > 0)
		{
			// Need more memory
			buffer_len *= 2;
			char * extended_buffer = realloc( buffer, buffer_len);
			if( !extended_buffer)
			{
				ret = ERROR_NOT_ENOUGH_MEMORY;
				goto error;
			}
			if( buffer != extended_buffer)
			{
				ret = ERROR_NOT_ENOUGH_MEMORY;
				goto error;
			}
		}
	}while(n>0);	// Only stop on eof or on failure
	if( n < 0)
	{
		ret = n;
		goto error;
	}
	buffer[buffer_read] = 0;
	*len=buffer_read;

error:
	return ret;
}

int read_infile( char ** buf, int * len, const char * data_file)
{
	int 	  		hData_file, file_size;
	struct stat		f_stat;
	char * buffer;

	hData_file = open(data_file, O_RDONLY);
	if(hData_file < 0)
	{
		gf_print(TOOL_NAME " error opening source file %s. (0x%X)\n", data_file, errno);
		return -errno;
	}

	file_size = fstat(hData_file, &f_stat);
	if (-1 == file_size)
	{
		gf_print(TOOL_NAME " error couldn't retrieve file attributes.\n");
		close(hData_file);
		return -errno;
	}

	buffer = (char*) malloc(sizeof(char)*f_stat.st_size);
	if (!buffer)
	{
		gf_print(TOOL_NAME " error couldn't allocate memory.\n");
		close(hData_file);
		return -ENOMEM;
	}
	*buf = buffer;

	file_size = read(hData_file, buffer, sizeof(char)*f_stat.st_size);
	if(file_size < 0)
	{
		gf_print(TOOL_NAME " error reading source file %s. (0x%X)\n", data_file, errno);
		close(hData_file);
		return -errno;
	}
	close(hData_file);
	*len = sizeof(char)*f_stat.st_size;
}

int main(int argc, char** argv)
{
	u_int32_t dwReturn = 0;
	u_int32_t counter;
	char* 	  device = STANDARD_EEPROM_FILE;

	if(argc < 2)
	{
		printUsage();
		return 0;
	}

	if(!strcmp(argv[1], GFEEPROM_OP_CLEAN))
	{
		bool full_content = false;

		for (counter = 1; counter < argc; counter++)
		{
			if (!strcmp(argv[counter], GFEEPROM_OP_DEVICE) && (counter+1 < argc))
				device = argv[++counter];
			else if (!strcmp(argv[counter], GFEEPROM_OP_FULL_CONTENT))
				full_content = true;
		}
		dwReturn = clean(device, full_content);
	}
	else if(!strcmp(argv[1], GFEEPROM_OP_SHOW))
	{
		bool	show_all_elements = true;
		bool	bin_dump = false;
		bool	hex_string = false;
		char*	file_name = NULL;
		u32 	eara_number = 0;

		for (counter = 2; counter < argc; counter++)
		{
			if (!strcmp(argv[counter], GFEEPROM_OP_DEVICE) && (counter+1 < argc))
			{
				device = argv[++counter];
				show_all_elements = false;
			}
			else if (!strcmp(argv[counter], GFEEPROM_OP_EEPROM_EA) && (counter+1 < argc))
			{
				eara_number = atol(argv[++counter]);
				show_all_elements = false;
			}
			else if (!strcmp(argv[counter], GFEEPROM_OP_BIN_DUMP))
			{
				bin_dump = true;
			}
			else if (!strcmp(argv[counter], GFEEPROM_OP_HEX_STRING))
			{
				hex_string = true;
			}
			else if (!strcmp(argv[counter], GFEEPROM_OP_QUIET))
			{
				quiet = true;
			}
			else if (!strcmp(argv[counter], GFEEPROM_OP_BIN_DUMP_TO_FILE) && (counter+1 < argc))
			{
				file_name = argv[++counter];
				bin_dump = true;
			}
			else if (!strcmp(argv[counter], GFEEPROM_OP_FULL_CONTENT))
			{
				eara_number = DUMP_FULL_CONTENT;
			}
			else{
				gf_print("Unknown argument for " GFEEPROM_OP_SHOW ": %s\n", argv[counter]);
				dwReturn = ERROR_UNKNOWN_ARG;
				goto error;
			}
		}
		if (show_all_elements)
			dwReturn = show_all();
		else
			dwReturn = show_eeprom(device, eara_number, bin_dump, hex_string, file_name);
	}
	else if (!strcmp(argv[1], GFEEPROM_OP_HELP))
	{
		printHelp();
	}
	else if (!strcmp(argv[1], GFEEPROM_OP_VERSION))
	{
		printVersion();
	}
	else if (!strcmp(argv[1], GFEEPROM_OP_PROGRAM))
	{
		u32 	eara_number = EEPROM_TYPE_EHW;
		char*	src_file = NULL;
		char*	src_string = NULL;
		bool	hex_string = false;
		char * buffer  = NULL;
		int		length = 0;

		for (counter = 2; counter < argc; counter++)
		{
			if (!strcmp(argv[counter], GFEEPROM_OP_DEVICE) && (counter+1 < argc))
			{
				device = argv[++counter];
			}
			else if (!strcmp(argv[counter], GFEEPROM_OP_EEPROM_EA) && (counter+1 < argc))
			{
				eara_number = atol(argv[++counter]);
			}
			else if (!strcmp(argv[counter], GFEEPROM_OP_BIN_SRC_FILE) && (counter+1 < argc))
			{
				src_file = argv[++counter];
			}
			else if (!strcmp(argv[counter], GFEEPROM_OP_BIN_SRC_STRING) && (counter+1 < argc))
			{
				src_string = argv[++counter];
			}
			else if (!strcmp(argv[counter], GFEEPROM_OP_HEX_STRING) )
			{
				hex_string = true;
			}
			else if (!strcmp(argv[counter], GFEEPROM_OP_QUIET))
			{
				quiet = true;
			}
			else{
				gf_print("Unknown argument for " GFEEPROM_OP_PROGRAM": %s\n", argv[counter]);
				dwReturn = ERROR_UNKNOWN_ARG;
				goto error;
			}
		}

		if (EEPROM_TYPE_EHW == eara_number)	// only supports src_file currently
		{
			dwReturn = program_guf_hw(src_file);
			if (ERROR_SUCCESS == dwReturn)
				gf_print(TOOL_NAME " programming finished successful.\n");
			goto finish;
		}

		// Read input source
		if( src_file == NULL && src_string == NULL)
		{
			dwReturn = read_stdin( &buffer, &length);
			if( dwReturn)
				goto error_free_buffer;
			src_string = buffer;
		}else if( src_file)
		{
			dwReturn = read_infile( &buffer, &length, src_file);
			if( dwReturn)
				goto error_free_buffer;
			src_string = buffer;
		}else if( src_string)
		{
			length = strlen(src_string);
		}
		// Convert input source
		if( hex_string)
		{
			int k,l=0;
			char * new_buffer = malloc(sizeof(char) * (length/2 + 2));
			if(!new_buffer)
			{
				dwReturn = ERROR_NOT_ENOUGH_MEMORY;
				goto error_free_buffer;
			}
			for( k = 0; k < length/2; k++)
			{
				new_buffer[k] = hexNibbleToVal(src_string[2*k]) << 4 
							|hexNibbleToVal(src_string[2*k+1]) << 0;
			}
			if(buffer) free(buffer);
			buffer = new_buffer;
			src_string = buffer;
			length = k;
		}
		
		dwReturn = program(device, eara_number, src_string, length);
		if (ERROR_SUCCESS == dwReturn)
			gf_print(TOOL_NAME " programming finished successful.\n");
error_free_buffer:
	if(buffer)
		free(buffer);
	}
	else
	{
		printUsage();
		dwReturn = -1;
	}

error:
finish:
	return dwReturn;
}
