﻿#!/usr/bin/python
#-------------------------------------------------------------------------------#
# gffirmware.py                                                                 #
# Author: Janik Hasche <janik.hasche@garz-fricke.com>                           #
#			                                                                    #
# Copyright 2014 Garz & Fricke GmbH                                             #
#-------------------------------------------------------------------------------#

# documentation-----------------------------------------------------------------#
"""
Searches Database for available firmware packages and lists/downloads them
"""

# defines ----------------------------------------------------------------------#
CONFIG_PATH = '/etc/gffirmware.conf'
REVISION = '$Revision: 4028 $'
NAME = 'gffirmware'
VERSION = '1.0'

# imports-----------------------------------------------------------------------#
import os
import sys
import urllib
import optparse
import ConfigParser
import xml.etree.ElementTree as etree

# functions---------------------------------------------------------------------#
def print_elem(elem_name, elem_value):
    """
    prints element formatted
    """
    print('\033[4m' + '\033[96m' + elem_name + ':' + '\033[0m')
    print(elem_value)


# classes-----------------------------------------------------------------------#
class Gf_Html_Getter():
    """
    Calculates a HTML-GET link and parses Information out of xml-data
    """
    def __init__(self, host, item_url, number):
        # initialize object's variables
        self.host = host
        self.item_url =item_url
        self.number = number

        # calculate prefix-string for xml elements
        self.xml_element_prefix = '{' + host + '}'

        # get an xml tree
        self.xml_tree = self.__get_xml_tree()


    def __get_xml_tree(self):
        """
        calculate http-get link and get an ElementTree-Object out of it
        """
        url = self.host + self.item_url + self.number

        try:
            xml_string = urllib.urlopen(url).read()
        # handle errors
        except IOError:
            show_error_and_close('Error: can\'t open url: ' + url )

        # try to parse xml from string
        try:
            xml_tree = etree.fromstring(xml_string)
        except etree.ParseError:
            show_error_and_close('Error: Problems parsing xml-tree(the given URL might be invalid)')

        return xml_tree


    def get_elements_texts(self, element_name):
        """
        get texts from elements out of the xml_tree. The element_name specifies the element
        (e.g.: ['Artikel', 'Artikelnummer'] for 'Artikel/Artikelnummer')
        """
        # calculate a path-string and get the texts of a requested element out of it
        element_string = self.xml_element_prefix + element_name

        # get all available texts and store them in a list
        elements = self.xml_tree.iter(element_string)
        texts = []
        for element in elements:
            text = element.text
            # workaround for problematic texts which could cause exceptions
            if text:
                text = text.encode('ascii', 'ignore').decode('ascii')
            texts.append(text)

        return texts



class Article():
    def __init__(self, article_number, description, device_text, device_warranty_end):
        self.article_number = article_number
        self.description = description
        self.device_text = device_text
        self.device_warranty_end = device_warranty_end



class FirmWarePackage():
    def __init__(self, article_number, description, usage, state, archive):
        self.article_number = article_number
        self.description = description
        self.usage = usage
        self.state = state
        self.archive = archive



class PartListFRM():
    def __init__(self, article_number, description):
        self.article_number = article_number
        self.description = description



class ArticleFRM():
    def __init__(self, type, path, md5):
        self.type = type
        self.path = path
        self.md5 = md5

        # parse filename out of path
        self.file_name = path.split('\\')[-1]



class Gffirmware():
    """
    provides functions for getting or showing information about the device and available firmware-packages
    """
    def __init__(self):
        # get serial number from device
        self.serial_number = self.get_serial_number()

        try:
            # get parameters for database out of config-file
            config = ConfigParser.ConfigParser()
            config.read(CONFIG_PATH)
            self.host = config.get('database', 'host')
            self.article_url = config.get('database', 'article_url')
            self.firmware_packages_url = config.get('database', 'firmware_packages_url')
            self.partlist_frm_url = config.get('database', 'partlist_frm_url')
            self.article_frm_url = config.get('database', 'article_frm_url')
            self.download_data_article_frm_url = config.get('database', 'download_data_article_frm_url')

            # get xml element names out of config-file
            self.article_number_xml = config.get('element_xml_names', 'article_number')
            self.description_xml = config.get('element_xml_names', 'description')
            self.device_text_xml = config.get('element_xml_names', 'device_text')
            self.device_warranty_end_xml = config.get('element_xml_names', 'device_warranty_end')
            self.usage_xml = config.get('element_xml_names', 'usage')
            self.state_xml = config.get('element_xml_names', 'state')
            self.archive_xml = config.get('element_xml_names', 'archive')
            self.type_xml = config.get('element_xml_names', 'type')
            self.path_xml = config.get('element_xml_names', 'path')
            self.md5_xml = config.get('element_xml_names', 'md5')

            # get xml element output names out of config-file
            self.article_number_output = config.get('element_output_names', 'article_number')
            self.description_output = config.get('element_output_names', 'description')
            self.device_text_output = config.get('element_output_names', 'device_text')
            self.device_warranty_end_output = config.get('element_output_names', 'device_warranty_end')
            self.type_output = config.get('element_output_names', 'type')
            self.path_output = config.get('element_output_names', 'path')
            self.md5_output = config.get('element_output_names', 'md5')
            self.file_name_output = config.get('element_output_names', 'file_name')

        except ConfigParser.NoSectionError:
            self.show_error_and_close('Error: can\'t find configuration file: ' + CONFIG_PATH)


    def get_serial_number(self):
        """
        get serial number using sconfig
        """
        # get serial number and clean string with strip()
        serial_number = os.popen('sconfig serial').read()
        serial_number = serial_number.strip()
        return serial_number


    def get_md5(self, file_name):
        """
        wrapper for md5 tool, returns md5 sum of given file(file_name)
        """
        # parse md5 sum out of output
        md5 = os.popen('md5sum ' + file_name).read()
        md5 = md5.split()[0]
        return md5


    def get_articles(self):
        """
        Gets all articles from serial number and stores information in a list of Article objects
        """
        # create a HTML getter an parse information with it
        gf_html_getter = Gf_Html_Getter(self.host, self.article_url, self.serial_number)
        article_numbers = gf_html_getter.get_elements_texts(self.article_number_xml)
        descriptions = gf_html_getter.get_elements_texts(self.description_xml)
        device_texts = gf_html_getter.get_elements_texts(self.device_text_xml)
        device_warranty_ends = gf_html_getter.get_elements_texts(self.device_warranty_end_xml)

        # create a list with Article objects out of information and return it
        artikel_list = []
        for article_number, description, device_text, device_warranty_end in zip(
            article_numbers, descriptions, device_texts, device_warranty_ends):
            artikel_list.append(Article(article_number, description, device_text, device_warranty_end))
        return artikel_list


    def get_firmwarepackages(self, article_number):
        """
        Gets all firmware packages from a given article number(article_number)
        and stores information in a list of FirmWarePackage objects
        """
        # create a HTML getter an parse information with it
        gf_html_getter = Gf_Html_Getter(self.host, self.firmware_packages_url, article_number)
        article_numbers = gf_html_getter.get_elements_texts(self.article_number_xml)
        descriptions = gf_html_getter.get_elements_texts(self.description_xml)
        usages = gf_html_getter.get_elements_texts(self.usage_xml)
        states = gf_html_getter.get_elements_texts(self.state_xml)
        archives = gf_html_getter.get_elements_texts(self.archive_xml)

        # create a list with FirmWarePackage objects out of information and return it
        firmwarepackage_list = []
        for article_number, description, usage, state, archive in zip(
            article_numbers, descriptions, usages, states, archives):
            firmwarepackage_list.append(FirmWarePackage(article_number, description, usage, state, archive))
        return firmwarepackage_list


    def get_partlist_frm(self, article_number):
        """
        Gets all part lists from a given article number(article_number)
        and stores information in a list of PartListFRM objects
        """
        # create a HTML getter an parse information with it
        gf_html_getter = Gf_Html_Getter(self.host, self.partlist_frm_url, article_number)
        article_numbers = gf_html_getter.get_elements_texts(self.article_number_xml)
        descriptions = gf_html_getter.get_elements_texts(self.description_xml)

        # create a list with PartListFRM objects out of information and return it
        partlist_frm_list = []
        for article_number, description in zip(article_numbers, descriptions):
            partlist_frm_list.append(PartListFRM(article_number, description))
        return partlist_frm_list


    def get_article_frm(self, article_number):
        """
        Gets all article frms from a given article number(article_number)
        and stores information in a list of ArticleFRM objects
        """
        # create a HTML getter an parse information with it
        gf_html_getter = Gf_Html_Getter(self.host, self.article_frm_url, article_number)
        types = gf_html_getter.get_elements_texts(self.type_xml)
        paths = gf_html_getter.get_elements_texts(self.path_xml)
        md5s = gf_html_getter.get_elements_texts(self.md5_xml)

        # create a list with ArticleFRM objects out of information and return it
        article_frm_list = []
        for type, path, md5 in zip(types, paths, md5s):
            article_frm_list.append(ArticleFRM(type, path, md5))
        return article_frm_list


    def show_error_and_close(self, error_msg):
        """
        Shows error with given error message(error_msg) and exits the script
        """
        print(error_msg)
        sys.exit()


    def show_device_information(self, verbose):
        """
        Gets all available articles for serial number and shows information
        """
        # get all articles and check content
        articles = self.get_articles()
        if len(articles) < 1:
            self.show_error_and_close('Error: Can\'t find any information for this serial number!')

        # go through articles and print information
        for article in articles:
            print_elem(self.article_number_output, article.article_number)
            print_elem(self.description_output, article.description)
            if verbose:
                print_elem(self.device_text_output,article.device_text)
                print_elem(self.device_warranty_end_output, article.device_warranty_end)
            print('')


    def show_available_firmware_packages(self):
        """
        Shows all available firmware packages for serial number
        """
        # get through all articles
        articles = self.get_articles()
        for article in articles:
            # print available firmware packages for each article
            print('Firmwarepackages for:' + article.description + '(' + article.article_number + '):')
            firmwarepackages = self.get_firmwarepackages(article.article_number)
            for firmwarepackage in firmwarepackages:
                print('- ' + (firmwarepackage.article_number).ljust(10) + ' : ' + firmwarepackage.description)


    def show_partlist_frm(self, article_number):
        """
        Shows part list to a given article number(article_number)
        """
        # get part list and check content
        partlist_frm = self.get_partlist_frm(article_number)
        if len(partlist_frm) < 1:
            self.show_error_and_close('Error: Can\'t find any part list for this article number')

        # go through part list and show article number and description
        print('Available files:')
        for part in partlist_frm:
            print((part.article_number).ljust(10) + ' : ' + part.description)


    def show_article_frm(self, article_number, verbose):
        """
        Shows information about an article to a given article number(article_number)
        """
        # get article frm and check content
        article_frm = self.get_article_frm(article_number)
        if len(article_frm) < 1:
            self.show_error_and_close('Error: Can\'t find an article for this article number!')

        # show information
        for article in article_frm:
            print_elem(self.article_number_output, article_number)
            print_elem(self.type_output, article.type)
            print_elem(self.file_name_output, article.file_name)
            print_elem(self.md5_output, article.md5)
            if verbose:
                print_elem(self.path_output, article.path)


    def download_article_frm(self, article_number, file_dir = ''):
        """
        Requests and downloads a file to a given frm article number(article_number)
        """
        # get articles filename and md5
        article_frm = self.get_article_frm(article_number)
        # check if there is an article and get information out of it
        if len(article_frm) < 1 :
            self.show_error_and_close('Error: Can\'nt find requested article')
        else:
            article = article_frm[0]
            file_name = article.file_name
            md5 = article.md5

        # calculate url and try to download the file
        url = self.host + self.download_data_article_frm_url + article_number
        file_path = file_dir + file_name
        try:
            f = urllib.urlopen(url)
            print('downloading file: ' + file_name + '...')
            # Open local file for writing
            local_file = open(file_path, 'wb')
            # Write to our local file
            local_file.write(f.read())
            local_file.close()
        # handle errors
        except IOError:
            self.show_error_and_close('Error: can\'t open url: ' + url )

        # check md5 sums
        if self.get_md5(file_path) != md5:
            self.show_error_and_close('Error: MD5sums don\'t match!')
        print('sucessful!')


# script functionality----------------------------------------------------------#
# don't execute this if script is imported
if __name__ == "__main__":
    # parse arguments
    version = NAME + ' ' + VERSION + ' (' + REVISION + ')'
    parser = optparse.OptionParser(description=(version+ '\n' +__doc__))
    parser.add_option("-i", "--information",
                      action="store_true", dest="information", default=False,
                      help="show device information")
    parser.add_option("-f", "--firmware",
                      action="store_true", dest="firmware", default=False,
                      help="show available firmware packages")
    parser.add_option("-p", "--partlist", dest="package_no",
                      help="show partlist of a firmware package", metavar="PACKAGE_NO(FWR-xxxx)")
    parser.add_option("-a", "--article", dest="article_no",
                      help="show information about a firmware article", metavar="ARTICLE-NO(FWR-xxxx)")
    parser.add_option("-v", "--verbose",
                      action="store_true", dest="verbose", default=False,
                      help="show detailed information")
    (options, args) = parser.parse_args()

    # execute specified action
    gffirmware = Gffirmware()
    if options.information:
        gffirmware.show_device_information(options.verbose)
    elif options.firmware:
        gffirmware.show_available_firmware_packages()
    elif options.package_no:
        gffirmware.show_partlist_frm(options.package_no)
    elif options.article_no:
        gffirmware.show_article_frm(options.article_no, options.verbose)
    else:
        gffirmware.show_error_and_close('No action specified!')