#!/bin/sh
### BEGIN INIT INFO
# Provides:          hdmi.sh
# Required-Start:    early-xml-mnt
# Required-Stop:     
# Should-Start:      
# Should-Stop:       
# Default-Start:     S
# Default-Stop:      0 6
# Short-Description: Unblanks and configures monitor connected to HDMI
### END INIT INFO

SVNVERSION='1.0'
USR_MODE_CNT=0
STD_MODE_CNT=0
MODE_CNT=0
MODES=""
USR_MODES=""
STD_MODES=""

usage()
{
    version
	echo "Usage: $0 {start | stop | show | list | set | blank | unblank | help }"
}

version()
{
    echo "hdmiconfig version $SVNVERSION <jonas.hoeppner@garz-fricke.com>"
}

usage_fault()
{
	usage >&2
	exit 1
}

help()
{
	usage

	echo "     start              Setup HDMI from xml settings, used during startup" 
	echo "     stop               Used during shutdown to blank the monitor" 
	echo "     list               List available modes"
	echo "     show [ -v | -vv ]  Show current mode, -v enables verbose output"
	echo "     set default   Switch to the default mode"
    echo "                        (normally highest resolution read from edid)"
	echo "     set <mode number>  Switch to mode specified by the number"
    echo "                        as displayed with 'list'"
	echo "     set <xres> <yres>  [<framerate>] [<depth>]"
	echo "                        Set HDMI to the given resolution, if not specified, "
	echo "                        framerate and depth are set to default values"
	echo "     blank              Blank the monitor" 
	echo "     unblank            Unblank the monitor" 

	exit 1
}


function pixclock_from_pixcels()
{
	local xres_total=$1
   	local yres_total=$2
	local framerate=$3

	local pixclk_hz=$( expr $xres_total \* $yres_total \* $framerate )
	local pixclk_ps=$( expr 1000000000000 \/ $pixclk_hz )

	echo "$pixclk_ps"
}

function pixclock_from_framerate()
{
	local framerate=$1; 
	local param=$2;
	local xres_total
   	local yres_total
	local pixclk_ps

	# Use fbset to get the configured or default margins
	read xres_total yres_total  <<<$( \
		fbset --test -fb $HDMI_FB_DEV $param -s -x  \
		| grep "HTimings\|VTimings" \
		| awk '{print $5}' \
		)
	pixclk_ps=$( pixclock_from_pixcels $xres_total $yres_total $framerate )
	
	echo "$pixclk_ps"
	return 0
}

function set_hdmi_from_xml()
{
	# Read all attributes from 'hdmi' node
	# Special attributes are name and refresh, 
	# all others are passed to fbset
	# If pixclock is setn it overwrites refresh
    # if none from refresh and 
	# pixclock are set, refresh=60 is assumed
	# pixclock is calculated from refresh and xres and yres
	#
	local fbset_parameter=""
	local framerate=""
	local pixclock
	local xres
	local yres
	local usedefault=0

	for attr in $(xconfig listattributes -p /variables/hdmi)
	do 
		if [ $attr == "default" ]; then usedefault=1; continue; fi

		value=$( xconfig getattribute -p /variables/hdmi -n $attr )

		if [ $attr == "refresh" ]; then framerate=$value; continue; fi
		if [ $attr == "xres" ]; then xres=$value; continue; fi
		if [ $attr == "yres" ]; then yres=$value; continue; fi

	done

	for attr in $(xconfig listattributes -p /variables/hdmi/fbset)
	do 

		value=$( xconfig getattribute -p /variables/hdmi/fbset -n $attr )
		fbset_parameter="$fbset_parameter -$attr $value "
	done

	if [ 1 == $usedefault ]; then
		set_hdmi_default
		unblank_hdmi
		return $?
	fi

	if [ -z $xres ] || [ -z $yres ]; then
		# Incomplete resolution settings, ignore
		# echo "Incomplete Settings: X $xres Y $yres"
		exit 1
	fi

	set_hdmi_from_xyres $xres $yres "$framerate" "$fbset_parameter"
	unblank_hdmi

	return $?
}

function set_hdmi_from_xyres()
{
	local pixclock
	local matching_mode
	local mmode; local mxres; local myres; local minterlace;
	local xres=$1
	local yres=$2
	local framerate=$3
	local fbset_parameter="$4"

	# echo " $xres $yres $framerate |$fbset_parameter|"

	if [ -z "$xres" ]; then usage_fault; fi
	if [ -z "$yres" ]; then usage_fault; fi

	read_hdmi_modes matching_mode $xres $yres $framerate
	if [ -z "$framerate" ]; then
		if [ -z $matching_mode ]; then
			#echo Framerate not set, use default
			framerate=60
		else
			#echo "Framerate not set, looking in mode $matching_mode ${MODES[$matching_mode]}"
			read_mode mmode mxres myres minterlace framerate ${MODES[$matching_mode]}
		fi
	fi
	
	if [ ! -z "$fbset_parameter" ]; then
		#echo " Force usage of fbset"
		matching_mode=""
	fi
	
	if [ ! -z $matching_mode ]; then
		set_hdmi_by_mode $matching_mode
	else
		fbset_parameter=" -xres $xres -yres $yres $fbset_parameter"
	
		#echo " $xres $yres Rate: $framerate |$fbset_parameter|"

		if ! echo "$fbset_parameter" | grep -q "pixclock";then
			# calculate the pixclock
			pixclock=$( pixclock_from_framerate $framerate "$fbset_parameter" ) 
			fbset_parameter=" $fbset_parameter -pixclock $pixclock "
		fi
		#echo "fbset -fb $HDMI_FB_DEV $fbset_parameter "
		fbset -fb $HDMI_FB_DEV $fbset_parameter 
	fi
	
	return $?
}

function set_hdmi_by_modestr()
{
	local modestr=$1
	
	echo $modestr > /sys/class/graphics/$HDMI_FB/mode
	return $?
}

function set_hdmi_by_mode()
{
	local number=$1
	local modestr

	read_hdmi_modes

	echo "Using HDMI mode $number"

	if [ -z $number ]; then usage_fault; fi
	if [ $number -gt $MODE_CNT ]; then
		echo "No such mode with the number $number"
		exit 1
	fi
	modestr=${MODES[$number]}
	if [ -z $modestr ]; then
		echo "No such mode with the number $number"
		exit 1
	fi

	set_hdmi_by_modestr $modestr
	return $?
}

function set_hdmi_default()
{
	local modestr

	read_hdmi_modes

	modestr=${MODES[$DFT_MODE]}
	#echo "default mode |$modestr|"
	if [ -z $modestr ]; then
		echo "Could not find default mode"
		set_hdmi_from_xyres 1920 1080 60
	else
		set_hdmi_by_modestr $modestr
	fi
	return $?
}

function set_hdmi()
{
	if [ -z $1 ]; then usage_fault; fi

	case "$1" in
		default)
			set_hdmi_default
			;;
		xml)
			set_hdmi_from_xml
			;;
		*)
			if [ -z $2 ]; then
				set_hdmi_by_mode $1 
			else
				if [ -z $4 ];then
					set_hdmi_from_xyres $1 $2 $3
				else
					set_hdmi_from_xyres $1 $2 $3 " -depth $4"
				fi
			fi
			;;
	esac
	return $?
}


function read_mode()
{
	local _mode=$1; 
	local _xres=$2;
   	local _yres=$3;
   	local _interlace=$4;
   	local _framerate=$5;

	set -- $(echo $6 | sed -e 's/\([UDVS]\):\([0-9]\+\)x\([0-9]\+\)\([ipd]\)-\([0-9]\+\).*/\1 \2 \3 \4 \5/')

	eval $_mode=\$1; 
	eval $_xres=\$2;
	eval $_yres=\$3;
   	eval $_interlace=\$4;
   	eval $_framerate=\$5;
}

function read_current_mode()
{
	local modestr=$( cat /sys/class/graphics/$HDMI_FB/mode)
	echo $modestr
}

function determine_default
{
	local number=$1
	local xres=$2
	local yres=$3
	local framerate=$4
	if [ $xres -gt $DFT_XRES ]; then
		DFT_XRES=$xres
		DFT_YRES=$yres
		DFT_RATE=$framerate
		DFT_MODE=$number
	else
		if [ $xres -eq $DFT_XRES ] && [ $yres -gt $DFT_YRES ]; then
			DFT_XRES=$xres
			DFT_YRES=$yres
			DFT_RATE=$framerate
			DFT_MODE=$number
		else
			if [ $xres -eq $DFT_XRES ] && [ $yres -eq $DFT_YRES ] && [ $framerate -gt $DFT_RATE ]; then
				DFT_XRES=$xres
				DFT_YRES=$yres
				DFT_RATE=$framerate
				DFT_MODE=$number
			fi
		fi
	fi
}

function read_hdmi_modes()
{
	local mode; local xres; local yres; local interlace; local framerate;
	local modestr;
	local t_framerate=0
	local s_mode=""
	local _out=$1
	local s_xres=$2
	local s_yres=$3
	local s_framerate=$4

	DFT_MODE=-1
	DFT_XRES=0
	DFT_YRES=0
	DFT_RATE=0

	USR_MODE_CNT=0
	STD_MODE_CNT=0
	MODE_CNT=0

	for modestr in $( cat /sys/class/graphics/$HDMI_FB/modes ); do
		read_mode mode xres yres interlace framerate $modestr
		
		if [ ! -z $s_xres ] && [ $xres -eq $s_xres ] && [ $yres -eq $s_yres ]; then
			if [ -z $s_framerate ] || [ $s_framerate -eq $framerate ]; then
				if [ $framerate -gt $t_framerate  ]; then
					s_mode=$MODE_CNT
					t_framerate=$framerate
				fi
			fi
		fi

		case "$mode" in
			S|V|D)
				determine_default $MODE_CNT $xres $yres $framerate
				STD_MODES[$STD_MODE_CNT]=$MODE_CNT
				STD_MODE_CNT=$(($STD_MODE_CNT+1))
				;;
			U)
				USR_MODES[$USR_MODE_CNT]=$MODE_CNT
				USR_MODE_CNT=$(($USR_MODE_CNT+1))
				;;
		esac
		MODES[$MODE_CNT]=$modestr
		MODE_CNT=$(($MODE_CNT+1))
	done

	if [ ! -z $s_mode ]; then
		eval $_out=\$s_mode
	fi
}

function print_mode()
{
	local modestr=$1
	local type=""
	local mode; local xres; local yres; local interlace; local framerate;
	read_mode mode xres yres interlace framerate $modestr
	case "$mode" in
		D)	type="(detailed)"	;;
		U)  type="(user)" ;;
		V)  type="(vesa)" ;;
	esac

	printf "X:% 5d Y:% 5d Framerate:% 4dHz %s" $xres $yres $framerate $type
}

function print_mode_by_number()
{
	local number=$1
	local modestr=${MODES[$number]}
	printf "% 3d) %s" $number "$(print_mode $modestr )"
}
	
function print_current_mode()
{
	echo "Current mode:"
	echo "       $(print_mode $(read_current_mode))"
}
function list_hdmi_modes()
{
	local i

	read_hdmi_modes

	print_current_mode

	if [ ! -z $DFT_MODE ]; then
		echo "Default mode:"
		echo "    $(print_mode_by_number $DFT_MODE)"
	fi
	if [ $STD_MODE_CNT -gt 0 ]; then
		echo "Standard monitor modes:"
		for (( i=0; i-$STD_MODE_CNT; i=(($i+1)) )); do
			echo "    $(print_mode_by_number ${STD_MODES[$i]})"
		done
	fi

	if [ $USR_MODE_CNT -gt 0 ]; then
		echo "User modes:"
		for (( i=0; i-$USR_MODE_CNT; i=(($i+1)) )); do
			echo "    $(print_mode_by_number ${USR_MODES[$i]})"
		done
	fi
}

function show_hdmi_mode()
{
	print_current_mode
	if [ "x$1" == "x-v" ]; then
		fbset -fb $HDMI_FB_DEV -s
	fi
	if [ "x$1" == "x-vv" ]; then
		fbset -fb $HDMI_FB_DEV -s -i
	fi
}

function get_hdmi_framebuffer
{
	local i
	FRAMEBUFFERS=$(ls /sys/class/graphics/)
	HDMI_FB=""
	for i in $FRAMEBUFFERS
	do 
		if [ "xhdmi" == "x$(cat /sys/class/graphics/$i/fsl_disp_dev_property)" ]; 
		then 
			HDMI_FB=$i 
		fi 
	done
	if [ "x" == "x$HDMI_FB" ]
	then
		echo No framebuffer for hdmi found.
		exit 1
	fi
	echo $HDMI_FB
	return 0
}

function blank_hdmi()
{
	echo 4 > /sys/class/graphics/$HDMI_FB/blank
}
function unblank_hdmi()
{
	echo 0 > /sys/class/graphics/$HDMI_FB/blank
}

function show_hdmi()
{
	HDMI_FB=$(get_hdmi_framebuffer)
	fbset -fb $HDMI_FB_DEV

	return 0
}
	
HDMI_FB=$(get_hdmi_framebuffer)
HDMI_FB_DEV="/dev/$HDMI_FB"

case "$1" in
	start)
		/etc/init.d/hdmi.sh background &
		;;
	background|restart|force-reload)
		set_hdmi_from_xml
		# Always return successful on init
		true
		;;

	unblank|-u)
		unblank_hdmi
		;;
	stop|blank|-b)
		blank_hdmi
		;;

	list|-l)
		list_hdmi_modes
		;;
	show|-s)
		show_hdmi_mode $2
		;;
	set)
		set_hdmi $2 $3 $4 $5
		;;
	--version|-v)
		version
		;;
	help|-h|-?|?)
		help
		;;

	*)
		usage;;
esac

exit $?
