/*
 * Copyright (C) 2013 Carsten Behling, Garz & Fricke GmbH
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>

#define MAP_PAGESIZE 4096UL

enum access_type { ACCESS_UNDEF = 0, ACCESS_8_BIT, ACCESS_16_BIT, ACCESS_32_BIT };

void usage(void)
{
	fprintf(stderr, "Usage: iopoke [-b location] [-1 | -2 | -4] [-v value]\n");
	return;
}

int main(int argc, char *argv[])
{
	int opt;
	int access = ACCESS_UNDEF;
	off_t location = -1;
	off_t page_offset = -1;
	off_t address_offset = -1;
	void *p_mem;
	volatile unsigned char *p_mem_8;
	volatile unsigned short *p_mem_16;
	volatile unsigned int *p_mem_32;
	unsigned int value = 0;
	int fd = -1;
	

	while ((opt = getopt(argc, argv, "b:124v:")) != -1) {
		switch (opt) {
			case 'b':
				location = strtol(optarg, NULL, 0);
				break;
			case '1':
				access = ACCESS_8_BIT;
				break;
			case '2':
				access = ACCESS_16_BIT;
				break;
			case '4':
				access = ACCESS_32_BIT;
				break;
			case 'v':
				value = strtol(optarg, NULL, 0);
				break;
			default: /* '?' */
				usage();
				exit(EXIT_FAILURE);
		}
	}

	if(location == -1) {
		usage();
		exit(EXIT_FAILURE);
	}
	
	if(access == ACCESS_32_BIT && (location % 4)) {
		fprintf(stderr, "Error: 32 bit access on non-32-bit-alligned address\n");
		exit(EXIT_FAILURE);	
	}
	
	if(access == ACCESS_16_BIT && (location % 2)) {
		fprintf(stderr, "Error: 16 bit access on non-16-bit-alligned address\n");
		exit(EXIT_FAILURE);	
	}
	
	if (geteuid() != 0) {
		fprintf(stderr, "Error: No root access rights\n");
		exit(EXIT_FAILURE);
    }

	page_offset = location & (off_t)~(MAP_PAGESIZE - 1);
	address_offset = location - page_offset;

	/* Open mem device for read/write */
	fd = open("/dev/mem", O_RDWR | O_SYNC);

	if (fd < 0) {
		fprintf(stderr, "Error: open of '/dev/mem' fail\n");
		exit(EXIT_FAILURE);
    }

	/* Get pointer to memory */
	p_mem = mmap(NULL,
		(off_t)MAP_PAGESIZE,
		(PROT_READ | PROT_WRITE),
		MAP_SHARED,
		fd,
		page_offset);

	if ((p_mem == MAP_FAILED) || (p_mem == NULL)) {
		fprintf(stderr, "Error: mmap of /dev/mem fail\n");
		close(fd);
		exit(EXIT_FAILURE);
	}
	
	switch(access) {
		case ACCESS_8_BIT:
		default:
			p_mem_8 = (volatile unsigned char *)p_mem;
			(*(p_mem_8 + address_offset)) = (unsigned char)value;
			break;
		case ACCESS_16_BIT:
			p_mem_16 = (volatile unsigned short *)p_mem;
			(*(p_mem_16 + address_offset / 2)) = (unsigned short)value;
			break;
		case ACCESS_32_BIT:
			p_mem_32 = (volatile unsigned int *)p_mem;
			(*(p_mem_32 + address_offset / 4)) = value;
			break;
	}
	
	if(munmap(p_mem, (off_t)MAP_PAGESIZE)) {
		fprintf(stderr, "Error: munmap of /dev/mem fail\n");
		close(fd);
		exit(EXIT_FAILURE);	
	}

	close(fd);
	
	exit(EXIT_SUCCESS);
}
