/******************************************************************************/
/* Copyright (C) 2011 Garz&Fricke GmbH						 			      */
/*		No use or disclosure of this information in any form without	      */
/*		the written permission of the author							      */
/******************************************************************************/

/******************************************************************************/
/* INCLUDES													 			      */
/******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <time.h>
#include <pthread.h>
#include <ncurses.h>
#include <sys/utsname.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/time.h>


/******************************************************************************/
/* DEFINES													 			      */
/******************************************************************************/
#define VERSION						"2.0"
#define SVN_REV						"$Revision: 3627 $"
#define RESULTS_LOGFILE_DEFAULT		"results.log"
#define TESTCASES_LOGFILE_DEFAULT	"testcases.log"
#define LTP_PATH_DEFAULT			"/opt/ltp"
#define COMMAND_BUFFER_LENGTH		256

// The testcases are stored in a linked list
typedef struct tc_list_item_t
{
	char name[64];
	char command[COMMAND_BUFFER_LENGTH];
	int result;
	int pass_count;
	int fail_count;
	pthread_t thread;
	int thread_running;
	struct tc_list_item_t *next;
} tc_list_item;


/******************************************************************************/
/* GLOBAL VARIABLES											 			      */
/******************************************************************************/
unsigned int	number_of_testcases = 0;
unsigned int	runs = 1;
unsigned int	run = 0;
unsigned int	parallel = 0;
unsigned int	cleanup_table_x = 0, cleanup_table_y = 0;

char			results_logfile[256]    = RESULTS_LOGFILE_DEFAULT;
char			testcases_logfile[256]  = TESTCASES_LOGFILE_DEFAULT;
char			ltp_path[256]           = LTP_PATH_DEFAULT;
char			configfile[256]         = "";
char			version[10]             = VERSION;

FILE			*p_logfile;
struct utsname	sysinfo;
pthread_mutex_t	mutex = PTHREAD_MUTEX_INITIALIZER;

// We have three linked list: testcases, preparation, cleanup
tc_list_item	testcase_top, preparation_top, cleanup_top;


/******************************************************************************/
/* FUNCTIONS												 			      */
/******************************************************************************/
void usage()
{
	printf("Built-In-Test v%s\n", version);
	printf("Syntax: bit [-r <runs>] [-l <logfile>] [-h] <scriptfile>\n");
	printf("	-r	: number of runs (default: 1), 0 is infinite\n");
	printf("	-l	: results logfile name (default: %s)\n", RESULTS_LOGFILE_DEFAULT);
	printf("	-t	: testcase output logfile name (default: %s)\n", TESTCASES_LOGFILE_DEFAULT);
	printf("	-f	: path to LTP installation (default: %s)\n", LTP_PATH_DEFAULT);
	printf("	-p	: run testcases in parallel instead of sequentially\n");
	printf("	-h	: this help\n");
	exit(1);
}


void parse_arguments(int argc, char **argv)
{
	int c;
	extern char *optarg;
	extern int optind;

	while ((c = getopt(argc, argv, ":l:t:r:f:hp")) != -1)
	{
		switch(c)
		{
		case 'h':
			usage();
			break;
		case 'r':
			runs = atoi(optarg);
			break;
		case 'l':
			strcpy(results_logfile, optarg);
			break;
		case 't':
			strcpy(testcases_logfile, optarg);
			break;
		case 'f':
			strcpy(ltp_path, optarg);
			break;
		case 'p':
			parallel = 1;
			break;
		case '?':
			usage();
			break;
		}
	}

	// Use the first non-option parameter as scriptfile
	if (optind == argc)
		usage();
	strcpy(configfile, argv[optind]);
}


int create_logfile()
{
	mkdir("results", S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);

	chdir("./results");

	// Open and close testcases logfile just to overwrite an existing one
	p_logfile = fopen(testcases_logfile, "w");
	fclose(p_logfile);

	p_logfile = fopen(results_logfile, "w");
	chdir("..");

	if (!p_logfile)
		return errno;

	return 0;
}


void close_logfile()
{
	fclose(p_logfile);
}


int fail_count(tc_list_item *p_testcase_top)
{
	int sum = 0;
	tc_list_item *p_testcase = p_testcase_top;

	while (p_testcase != NULL)
	{
		if (p_testcase->result != 0)
			sum++;
		p_testcase = p_testcase->next;
	}

	return sum;
}


void log_header()
{
	time_t rawtime;
	struct tm *timeinfo;

	time ( &rawtime );
	timeinfo = localtime ( &rawtime );

	fprintf(p_logfile, "Test Start Time: %s\n", asctime (timeinfo));
	fprintf(p_logfile, "----------------------------------------------------\n");
	fprintf(p_logfile, "Testcase                       Result     Exit Value\n");
	fprintf(p_logfile, "--------                       ------     ----------\n");
}


void log_summary(int mseconds)
{
	fprintf(p_logfile, "\n----------------------------------------------------\n");
	fprintf(p_logfile, "Total Tests: %d\n", number_of_testcases);
	fprintf(p_logfile, "Total Failures: %d\n", fail_count(&testcase_top));
	fprintf(p_logfile, "Kernel Version: %s\n", sysinfo.release);
	fprintf(p_logfile, "Machine Architecture: %s\n", sysinfo.machine);
	fprintf(p_logfile, "Hostname: %s\n\n", sysinfo.nodename);
	fprintf(p_logfile, "Elapsed time: %d.%d seconds\n\n\n", mseconds/1000, mseconds%1000/100);
}


void log_result(char *name, int value)
{
	char *result_string = "PASS";
	if (value != 0)
		result_string = "FAIL";
	fprintf(p_logfile, "%-30s %-10s %d\n", name, result_string, value);
}


int read_configfile()
{
	FILE *fp;
	static char line[COMMAND_BUFFER_LENGTH + 50];
	static char temp[COMMAND_BUFFER_LENGTH];
	static char *line_ptr, *command, *name;
	tc_list_item *p_current_testcase = &testcase_top;
	tc_list_item *p_current_preparation = &preparation_top;
	tc_list_item *p_current_cleanup = &cleanup_top;
	tc_list_item **pp_list_item = &p_current_testcase;

	chdir("./runtest");
	fp = fopen(configfile, "r");
	chdir("..");

	if (!fp)
		return errno;

	// Read configfile line by line
	while (fgets(line, sizeof(line), fp))
	{
		// Check for separation line
		if (line[0]=='#' && line[1]==' ')
		{
			// Switch to the according linked list
			if (line[2]=='P')
				pp_list_item = &p_current_preparation;
			if (line[2]=='T')
				pp_list_item = &p_current_testcase;
			if (line[2]=='C')
				pp_list_item = &p_current_cleanup;
		}

		// Skip comment and empty lines
		if (line[0]=='#' || line[0]==' ' || line[0]=='\n')
			continue;

		// Split line into name and command
		line_ptr = line;
		name = line_ptr;
		while (*line_ptr != ' ' && *line_ptr != '\t')
			line_ptr++;
		*(line_ptr++) = 0;
		while (*line_ptr == ' ' || *line_ptr == '\t')
			line_ptr++;
		command = line_ptr;

		// Store command in linked list
		strcpy((*pp_list_item)->name, name);
		strcpy((*pp_list_item)->command, command);
		// Redirect testcase output to logfile
		strcpy(temp, "(");
		strcat(temp, (*pp_list_item)->command);
		temp[strlen(temp)-1] = 0;
		strcat(temp, ") >>");
		strcat(temp, ltp_path);
		strcat(temp, "/results/");
		strcat(temp, testcases_logfile);
		strcat(temp, " 2>&1");
		strcpy((*pp_list_item)->command, temp);
		(*pp_list_item)->next = calloc(1, sizeof(tc_list_item));
		*pp_list_item = (*pp_list_item)->next;

		if (*pp_list_item == p_current_testcase)
			number_of_testcases++;
	}
	fclose(fp);

	return 0;
}


void free_linked_lists()
{
	tc_list_item *p_testcase;

	p_testcase = testcase_top.next;
	while (p_testcase != NULL)
	{
		tc_list_item *temp = p_testcase->next;
		free(p_testcase);
		p_testcase = temp;
	}
	p_testcase = preparation_top.next;
	while (p_testcase != NULL)
	{
		tc_list_item *temp = p_testcase->next;
		free(p_testcase);
		p_testcase = temp;
	}
	p_testcase = cleanup_top.next;
	while (p_testcase != NULL)
	{
		tc_list_item *temp = p_testcase->next;
		free(p_testcase);
		p_testcase = temp;
	}
}


inline void testcase_update_results(tc_list_item *p_testcase, int retval)
{
	p_testcase->result = retval;

	if (p_testcase->result != 0) {
		p_testcase->fail_count++;
	} else {
		p_testcase->pass_count++;
	}
}


int total_pass_count(tc_list_item *p_testcase_top)
{
	int sum = 0;
	tc_list_item *p_testcase = p_testcase_top;

	while (p_testcase != NULL)
	{
		sum += p_testcase->pass_count;
		p_testcase = p_testcase->next;
	}

	return sum;
}


int total_fail_count(tc_list_item *p_testcase_top)
{
	int sum = 0;
	tc_list_item *p_testcase = p_testcase_top;

	while (p_testcase != NULL)
	{
		sum += p_testcase->fail_count;
		p_testcase = p_testcase->next;
	}

	return sum;
}


void result_table_header(char *headline)
{
	printw("\n----------------------------------------------------------------\n");
	printw("%-30s Result   RetVal   Passed   Failed\n", headline);
	printw("----------------------------   ------   ------   ------   ------\n");
	refresh();
}


void result_table_entry_start(tc_list_item *p_testcase)
{
	printw("%-30s ", p_testcase->name);
	refresh();
}


void result_table_entry_tail(tc_list_item *p_testcase)
{
	printw("%-6s %8d %8d %8d\n",
		p_testcase->result == 0 ? "PASS" : "FAIL",
		p_testcase->result,
		p_testcase->pass_count,
		p_testcase->fail_count
	);
	if (p_testcase->result != 0)
	{
		beep();
	}
	refresh();
}


void result_table_footer(int mseconds, int passed, int failed)
{
	printw("----------------------------------------------------------------\n");
	printw("Elapsed time: %2d.%1d seconds %28d %8d\n", mseconds/1000, mseconds%1000/100, passed, failed);
	printw("----------------------------------------------------------------\n");
	refresh();
}


inline int mseconds_between_timevals(struct timeval start, struct timeval end)
{
	return ((end.tv_sec - start.tv_sec) * 1000 + (end.tv_usec - start.tv_usec)/1000.0) + 0.5;
}


void run_cleanup()
{
	tc_list_item *p_cleanup = &cleanup_top;
	struct timeval start, end;
	int ret;

	pthread_mutex_lock(&mutex);

	// Destroy all parallel threads
	if (parallel)
	{
		tc_list_item *p_testcase = &testcase_top;
		while (p_testcase->name[0] != 0)
		{
			if (!pthread_equal(pthread_self(), p_testcase->thread) && p_testcase->thread_running)
			{
				pthread_cancel(p_testcase->thread);
				pthread_join(p_testcase->thread, NULL);
			}
			p_testcase = p_testcase->next;
		}
	}

	if (cleanup_table_y != 0)
		move(cleanup_table_y, cleanup_table_x);

	result_table_header("Cleanup");
	gettimeofday(&start, NULL);
	chdir("./testcases/bin");

	while (p_cleanup->name[0] != 0)
	{
		result_table_entry_start(p_cleanup);
		ret = system(p_cleanup->command);
		testcase_update_results(p_cleanup, WEXITSTATUS(ret));
		result_table_entry_tail(p_cleanup);

		// Check for Ctrl+C
		if (WIFSIGNALED(ret) && (WTERMSIG(ret) == SIGINT || WTERMSIG(ret) == SIGQUIT))
			break;

		p_cleanup = p_cleanup->next;
	}

	// Print table footer
	gettimeofday(&end, NULL);
	result_table_footer(
		mseconds_between_timevals(start, end),
		total_pass_count(&cleanup_top),
		total_fail_count(&cleanup_top)
	);

	chdir("../..");

	// Destroy the linked lists of testcases/preparations/cleanups
	free_linked_lists();

	endwin();

	exit(0);

	pthread_mutex_unlock(&mutex);
}


void run_preparations()
{
	tc_list_item *p_preparation = &preparation_top;
	struct timeval start, end;
	int ret;

	result_table_header("Preparation");
	gettimeofday(&start, NULL);
	chdir("./testcases/bin");

	p_preparation = &preparation_top;
	while (p_preparation->name[0] != 0)
	{
		result_table_entry_start(p_preparation);
		ret = system(p_preparation->command);
		testcase_update_results(p_preparation, WEXITSTATUS(ret));
		result_table_entry_tail(p_preparation);

		// Check for Ctrl+C
		if (WIFSIGNALED(ret) && (WTERMSIG(ret) == SIGINT || WTERMSIG(ret) == SIGQUIT))
			run_cleanup();

		if (WEXITSTATUS(ret) != 0 && WEXITSTATUS(ret) != 255)
		{
			printw("\nStep \"%s\" reported an error(%d).\nPreparation failed!\n", p_preparation->name, WEXITSTATUS(ret));
			refresh();
			run_cleanup();
		}

		p_preparation = p_preparation->next;
	}

	// Print table footer
	gettimeofday(&end, NULL);
	result_table_footer(
		mseconds_between_timevals(start, end),
		total_pass_count(&preparation_top),
		total_fail_count(&preparation_top)
	);

	chdir("../..");
}


void *run_test(void *arg)
{
	tc_list_item *p_testcase = (tc_list_item *)arg;
	int ret;

	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	ret = system(p_testcase->command);
	testcase_update_results(p_testcase, WEXITSTATUS(ret));

	// Check for Ctrl+C
	if (WIFSIGNALED(ret) && (WTERMSIG(ret) == SIGINT || WTERMSIG(ret) == SIGQUIT))
		run_cleanup();

	return NULL;
}


void run_tests()
{
	tc_list_item *p_testcase = &testcase_top;
	struct timeval start, end;
	long mseconds;
	int ret;

	chdir("./testcases/bin");
	log_header();
	gettimeofday(&start, NULL);

	// Print table header for test results
	result_table_header("Testcase");

	// Run tests
	p_testcase = &testcase_top;
	while (p_testcase->name[0] != 0)
	{
		if (parallel)
		{
			p_testcase->thread_running = 1;
			pthread_create(&p_testcase->thread, NULL, run_test, p_testcase);
		}
		else
		{
			result_table_entry_start(p_testcase);

			ret = system(p_testcase->command);
			testcase_update_results(p_testcase, WEXITSTATUS(ret));
			p_testcase->result = WEXITSTATUS(ret);

			// Check for Ctrl+C
			if (WIFSIGNALED(ret) && (WTERMSIG(ret) == SIGINT || WTERMSIG(ret) == SIGQUIT))
				run_cleanup();

			result_table_entry_tail(p_testcase);
		}

		p_testcase = p_testcase->next;
	}

	// Log results
	p_testcase = &testcase_top;
	while (p_testcase->name[0] != 0)
	{
		if (parallel)
		{
			// Don't join more threads if we're cleaning up already
			pthread_mutex_lock(&mutex);
			pthread_mutex_unlock(&mutex);

			result_table_entry_start(p_testcase);

			pthread_join(p_testcase->thread, NULL);
			p_testcase->thread_running = 0;

			result_table_entry_tail(p_testcase);
		}

		log_result(p_testcase->name, p_testcase->result);
		p_testcase = p_testcase->next;
	}

	gettimeofday(&end, NULL);
	mseconds = ((end.tv_sec  - start.tv_sec) * 1000 + (end.tv_usec - start.tv_usec)/1000.0) + 0.5;

	// Print table footer
	result_table_footer(
		mseconds_between_timevals(start, end),
		total_pass_count(&testcase_top),
		total_fail_count(&testcase_top)
	);
	getyx(curscr, cleanup_table_y, cleanup_table_x);

	log_summary(mseconds);
	chdir("../..");
}


int main(int argc, char **argv)
{
	char path[512];
	char *oldpath;
	unsigned int testcase_table_x, testcase_table_y;

	// Construct version string
	strcat(version, "r");
	strncat(version, strchr(SVN_REV, ' ')+1, strrchr(SVN_REV, ' ')-strchr(SVN_REV, ' ')-1);

	// Parse command line arguments
	parse_arguments(argc, argv);

	// Change working directory to the LTP root
	chdir(ltp_path);

	// Add current directory to the search PATH, so that
	// testcases can be executed without absolute path.
	oldpath = getenv("PATH");
	strcpy(path, oldpath);
	strcat(path, ":.");
	setenv("PATH", path, 1);

	// Read configfile and create a linked list of all testcases and preparation/cleanup steps
	if (read_configfile() != 0)
	{
		printf("ERROR: Could not read configfile \"%s\"\n", configfile);
		return -1;
	}

	// Create logfile
	if (create_logfile() != 0)
	{
		printf("ERROR: Could not create logfile \"%s\"\n", results_logfile);
		return -1;
	}

	// Read system information for logfile summary
	uname(&sysinfo);

	// Start the curses mode
	initscr();

	// Display header
	move(0, 0);
	printw("****************************************************************\n");
	printw("* Garz & Fricke - Linux Built-In-Test                v%s *\n", version);
	printw("****************************************************************\n");
	refresh();

	// Run preparations
	run_preparations();

	// Save current cursor position for testcase table
	getyx(curscr, testcase_table_y, testcase_table_x);

	// Run all tests in the list
	for (run = 0; run < runs || runs == 0; run++)
	{
		move(testcase_table_y, testcase_table_x);
		run_tests();
	}

	// Close logfile
	close_logfile();

	// Run cleanup
	run_cleanup();

	endwin();

	return 0;
}
