/******************************************************************************/
/* Copyright (C) 2013 Garz&Fricke GmbH						 			      */
/*		No use or disclosure of this information in any form without	      */
/*		the written permission of the author							      */
/******************************************************************************/

/******************************************************************************/
/*                                                                            */
/* File:        can.c                                                         */
/*                                                                            */
/* Description: This is a test for can busses.                                */
/*                                                                            */
/* Total Tests: 1                                                             */
/*                                                                            */
/* Test Name:   can                                                           */
/*                                                                            */
/* Test Assertion                                                             */
/* & Strategy:  Random can payload will be sent in all of the 8 data bytes    */
/*              of the CAN message with a configurable ID.                    */
/*              The remote peer will be    */
/*              a PC running the Garz & Fricke 'TestCAN_W32_2.0.2.0' tool     */
/*				from the DOS prompt or a batch file as follows:               */
/*                TestCAN_W32_2.0.2.0.exe -s -v1 -x4						  */
/*				The above command starts the TestCAN tool in receive mode.    */
/*              To enable the echo mode, we have to send a special            */
/*              "ECHO_ENABLE" message first from the device with the ID 0x12  */
/*              and the first data byte 0x01 to enable the echo               */
/*              (likewise 0x00 to disable the echo)                           */
/*              TestCAN runs with above configuration at a baudrate of 125k.  */
/*              After enabling the echo mode we'll send random payload data   */
/*              an check if the messages are properly echoed.                 */
/*                                                                            */
/* Author:      Carsten Behling <carsten.behling@garz-fricke.com>             */
/*                                                                            */
/******************************************************************************/

/* Standard Include Files */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <time.h>
#include <sys/ioctl.h>
#include <sys/select.h>
#include <net/if.h>
#include <linux/can.h>
#include <linux/can/raw.h>
#include <libsocketcan.h>
#include <unistd.h>
#include <pthread.h>
#include <poll.h>

/* Harness Specific Include Files. */
#include "test.h"
#include "usctest.h"

/* Local Defines */
#if !defined(TRUE) && !defined(FALSE)
#define TRUE  1
#define FALSE 0
#endif

/*If yocto is used (ltp since git d59a659cd639ca2780b00049d102acd2a783d585) */
#ifdef YOCTO
#define Tst_count tst_count
#endif

#define MSG_ID_ECHO 			0x12 /* same as in TestCAN_W32_2.0.2.0 */
#define DISABLE_ECHO			0x00 /* has to reside in data[0] of MSG_ID_ECHO message */
#define ENABLE_ECHO				0x01 /* has to reside in data[0] of MSG_ID_ECHO message */

#define DEFAULT_DEVICE_NAME		"can0"
#define DEFAULT_BITRATE			125000
#define DEFAULT_CAN_ID			0x100
#define DEFAULT_MESSAGE_COUNT	10
#define DEFAULT_WAIT_US			10000
#define DEFAULT_TIMEOUT_MS		1000

/* Extern Global Variables */
extern int  Tst_count;             	/* counter for tst_xxx routines.         */
extern char *TESTDIR;              	/* temporary dir created by tst_tmpdir() */

/* Global Variables */
char *TCID     = "can            ";	/* test program identifier.              */
int  TST_TOTAL = 1;                 /* total number of tests in this file.   */

char *conf_device_name          = DEFAULT_DEVICE_NAME;      /* name of the interface */
unsigned int conf_bitrate       = DEFAULT_BITRATE;          /* bitrate */
unsigned int conf_can_id        = DEFAULT_CAN_ID;           /* CAN ID to use */
unsigned int conf_message_count = DEFAULT_MESSAGE_COUNT;    /* message count */
unsigned int conf_wait_us       = DEFAULT_WAIT_US;          /* microseconds to wait between writing and reading */
unsigned int conf_timeout_ms    = DEFAULT_TIMEOUT_MS;       /* receive timeout */
unsigned int conf_quiet         = 0;                        /* quiet mode */

int sockfd;
struct sockaddr_can addr;
struct ifreq ifr;
struct can_frame send_message;
struct can_frame receive_message;
time_t t;                      /* used for the random generator */
pthread_mutex_t count_mutex     = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t  buffer_full     = PTHREAD_COND_INITIALIZER;
pthread_cond_t  buffer_empty    = PTHREAD_COND_INITIALIZER;
unsigned int data_sent          = 0;

/* Extern Global Functions */
/******************************************************************************/
/*                                                                            */
/* Function:    cleanup                                                       */
/*                                                                            */
/* Description: Performs all one time clean up for this test on successful    */
/*              completion,  premature exit or  failure. Closes all temporary */
/*              files, removes all temporary directories exits the test with  */
/*              appropriate return code by calling tst_exit() function.       */
/*                                                                            */
/* Input:       None.                                                         */
/*                                                                            */
/* Output:      None.                                                         */
/*                                                                            */
/* Return:      On failure - Exits calling tst_exit(). Non '0' return code.   */
/*              On success - Exits calling tst_exit(). With '0' return code.  */
/*                                                                            */
/******************************************************************************/
extern void cleanup()
{
	struct can_frame disable_echo_message;
	ssize_t nbytes = 0;


	/* reset TestCAN_W32 echo mode */
	disable_echo_message.can_id = MSG_ID_ECHO;
	disable_echo_message.can_dlc = 1;
	disable_echo_message.data[0] = DISABLE_ECHO;
	nbytes = write(sockfd, &disable_echo_message, sizeof(disable_echo_message));
	if(nbytes != sizeof(disable_echo_message)) {
		tst_resm(TWARN,
			"Unable to send echo disable message for can interface %s. Error: %d, %s",
			conf_device_name, errno, strerror(errno));
	}

	if(sockfd != -1)
		close(sockfd);

	/* stop can interface */
	can_do_stop(conf_device_name);
	tst_exit();
}


/* Local Functions */
/******************************************************************************/
/*                                                                            */
/* Function:    setup                                                         */
/*                                                                            */
/* Description: Performs all one time setup for this test. This function is   */
/*              typically used to capture signals, create temporary dirs      */
/*              and temporary files that may be used in the course of this    */
/*              test.                                                         */
/*                                                                            */
/* Input:       None.                                                         */
/*                                                                            */
/* Output:      None.                                                         */
/*                                                                            */
/* Return:      On failure - Exits by calling cleanup().                      */
/*              On success - Nothing.                                         */
/*                                                                            */
/******************************************************************************/
void setup()
{
	int err = 0;
	struct can_frame enable_echo_message;
	ssize_t nbytes = 0;

	/* create send socket */
	sockfd = socket(PF_CAN, SOCK_RAW, CAN_RAW);
	if(sockfd == -1) {
		tst_brkm(TBROK, cleanup,
			"Unable to create a socket for can interface %s. Error: %d, %s",
			conf_device_name, errno, strerror(errno));
	}

	/* stop can interface, otherwise the bitrate cannot be set */
	can_do_stop(conf_device_name);

	/* set bitrate */
	err = can_set_bitrate(conf_device_name, conf_bitrate);
	if (err < 0) {
		tst_brkm(TBROK, cleanup,
			"Unable to set bitrate for the can interface %s. Error: %d, %s",
			conf_device_name, errno, strerror(errno));
	}

	/* start can interface */
	can_do_start(conf_device_name);

	/* initialize socket */
	addr.can_family = AF_CAN;
	addr.can_ifindex = if_nametoindex(conf_device_name);

	bind(sockfd, (struct sockaddr*)&addr, sizeof(addr));

	/* set TestCAN_W32 echo mode */
	enable_echo_message.can_id = MSG_ID_ECHO;
	enable_echo_message.can_dlc = 1;
	enable_echo_message.data[0] = ENABLE_ECHO;
	nbytes = write(sockfd, &enable_echo_message, sizeof(enable_echo_message));
	if(nbytes != sizeof(enable_echo_message)) {
		tst_brkm(TBROK, cleanup,
			"Unable to send echo enable message for can interface %s. Error: %d, %s",
			conf_device_name, errno, strerror(errno));
	}

	/* initialize random number generator */
	time(&t);
	srand((int)t);
}

/******************************************************************************/
/*                                                                            */
/* Function:    receive_thread                                                */
/*                                                                            */
/* Description: This is the receive thread, that ...                          */
/*                                                                            */
/* Input:       None.                                                         */
/*                                                                            */
/* Output:      None.                                                         */
/*                                                                            */
/* Return:      None.                                                         */
/*                                                                            */
/******************************************************************************/
void *receive_thread(void* dummy)
{
	unsigned int i = 0;
	int status = 0;
	struct pollfd fds[1];
	int ret = 0;

	fds[0].fd = sockfd;
	fds[0].events = POLLIN;

	for(i = 0; i < conf_message_count; i++) {
		/* wait for data */
		pthread_mutex_lock(&count_mutex);
		while (!data_sent)
			pthread_cond_wait(&buffer_full, &count_mutex);

		ret = poll(fds, 1, conf_timeout_ms);
		if(ret == -1) {
			tst_resm(TFAIL, "Poll error while reading CAN message %d.", i);
			status = -1;
		} else if(ret == 0) {
			tst_resm(TFAIL, "Timeout while reading CAN message.");
			status = -1;
		} else {

			if(read(sockfd, &receive_message, sizeof(receive_message))
				!= sizeof(receive_message)) {
				tst_resm(TFAIL, "CAN message could no be read.");
				status = -1;
			}

			if (receive_message.can_id & CAN_EFF_FLAG) {
				receive_message.can_id &= CAN_EFF_MASK;
			} else {
				receive_message.can_id &= CAN_SFF_MASK;
			}

			if (!conf_quiet) {
				tst_resm(TINFO,
					"R: ID: 0x%04x DLC: %d Data: 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x",
					receive_message.can_id,
					receive_message.can_dlc,
					receive_message.data[0],
					receive_message.data[1],
					receive_message.data[2],
					receive_message.data[3],
					receive_message.data[4],
					receive_message.data[5],
					receive_message.data[6],
					receive_message.data[7]);
			}

			/* compare send and receive message */
			if(receive_message.data[0] != send_message.data[0] ||
			   receive_message.data[1] != send_message.data[1] ||
			   receive_message.data[2] != send_message.data[2] ||
			   receive_message.data[3] != send_message.data[3] ||
			   receive_message.data[4] != send_message.data[4] ||
			   receive_message.data[5] != send_message.data[5] ||
			   receive_message.data[6] != send_message.data[6] ||
			   receive_message.data[7] != send_message.data[7]) {
				tst_resm(TINFO, "Compare error in CAN message %d:", i);
				tst_resm(TINFO, "S: 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x",
					send_message.data[0], send_message.data[1], send_message.data[2], send_message.data[3],
					send_message.data[4], send_message.data[5], send_message.data[6], send_message.data[7]);
				tst_resm(TINFO, "R: 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x\n",
					receive_message.data[0], receive_message.data[1], receive_message.data[2], receive_message.data[3],
					receive_message.data[4], receive_message.data[5], receive_message.data[6], receive_message.data[7]);
				status = -1;
			}
		}
		data_sent = 0;
		pthread_mutex_unlock(&count_mutex);
		/* signal send thread that buffer is empty */
		pthread_cond_signal(&buffer_empty);
	}
	pthread_exit((void*)status);
}

/******************************************************************************/
/*                                                                            */
/* Function:    test                                                          */
/*                                                                            */
/* Description: This function ...                                             */
/*                                                                            */
/* Input:       None.                                                         */
/*                                                                            */
/* Output:      None.                                                         */
/*                                                                            */
/* Return:      On failure - returns -1.                                      */
/*              On success - returns 0.                                       */
/*                                                                            */
/******************************************************************************/
int test()
{
	int ret = 0;
	pthread_t receive_thread_handle;
	int status = 0;
	unsigned int i = 0;
	unsigned int j = 0;

	/* create the receive thread */
	ret = pthread_create(&receive_thread_handle, NULL, receive_thread, NULL);
	if (ret) {
		tst_brkm(TBROK, cleanup,
			"Receive thread could not be started.");
	}

	for(i = 0; i < conf_message_count; i++) {
		/* wait until last data has been received */
		pthread_mutex_lock(&count_mutex);
		while (data_sent)
			pthread_cond_wait(&buffer_empty, &count_mutex);

		/* send the test message */
		send_message.can_id = conf_can_id;
		send_message.can_dlc = 8;

		for(j = 0; j < 8; j++) {
			send_message.data[j] = rand() % 256;
		}
		if (!conf_quiet) {
			tst_resm(TINFO,
				"S: ID: 0x%04x DLC: %d Data: 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x",
				send_message.can_id,
				send_message.can_dlc,
				send_message.data[0],
				send_message.data[1],
				send_message.data[2],
				send_message.data[3],
				send_message.data[4],
				send_message.data[5],
				send_message.data[6],
				send_message.data[7]);
		}
		if(write(sockfd, &send_message, sizeof(send_message))
			!= sizeof(send_message)) {
			ret = -1;
		}
		/* signal receive thread that data has been sent */
		data_sent = 1;
		pthread_mutex_unlock(&count_mutex);
		pthread_cond_signal(&buffer_full);
	}

	/* wait for receive thread to finish */
	pthread_join(receive_thread_handle, (void*)&status);
	if(status < 0) {
		tst_resm(TFAIL, "CAN message could no be sent.");
		ret = -1;
	}

	return ret;
}

/******************************************************************************/
/*                                                                            */
/* Function:    help                                                          */
/*                                                                            */
/* Description: This function is called when the test is started with         */
/*              parameter -h. It displays all parameter options specifically  */
/*              available for this test.                                      */
/*                                                                            */
/* Input:       None.                                                         */
/*                                                                            */
/* Output:      None.                                                         */
/*                                                                            */
/* Return:      Nothing.                                                      */
/*                                                                            */
/******************************************************************************/
void help()
{
	printf("  -d s    Use CAN bus device s (default: %s)\n", DEFAULT_DEVICE_NAME);
	printf("  -i x    Use CAN bus id x (default: 0x%x)\n", DEFAULT_CAN_ID);
	printf("  -b x    Use CAN bus bitrate x (default: %u)\n", DEFAULT_BITRATE);
	printf("  -n x    Send and receive x CAN messages (default: %u)\n", DEFAULT_MESSAGE_COUNT);
	printf("  -w x    Wait x us between reading and writing (default: %u)\n", DEFAULT_WAIT_US);
	printf("  -o x    Timeout after x ms (default: %u)\n", DEFAULT_TIMEOUT_MS);
	printf("  -q      Suppress info messages\n");
}


/******************************************************************************/
/*                                                                            */
/* Function:    main                                                          */
/*                                                                            */
/* Description: Entry point to this test-case. It parses all the command line */
/*              inputs, calls the global setup and executes the test. It logs */
/*              the test status and results appropriately using the LTP API's */
/*              On successful completion or premature failure, cleanup() func */
/*              is called and test exits with an appropriate return code.     */
/*                                                                            */
/* Input:       Command line parameters (see help())                          */
/*                                                                            */
/* Exit:        On failure - Exits by calling cleanup().                      */
/*              On success - exits with 0 exit value.                         */
/*                                                                            */
/******************************************************************************/
int main(int argc, char **argv)
{
	int lc;		/* loop counter */
	int opt_d = 0, opt_i = 0, opt_b = 0, opt_n = 0, opt_w = 0, opt_o = 0, opt_q = 0, ret = -1;
	char *str_d,  *str_i, *str_b, *str_n, *str_w, *str_o, *msg;

	option_t options[] = {
		{"d:", &opt_d, &str_d},
		{"i:", &opt_i, &str_i},
		{"b:", &opt_b, &str_b},
		{"n:", &opt_n, &str_n},
		{"w:", &opt_w, &str_w},
		{"o:", &opt_o, &str_o},
		{"q", &opt_q, NULL},
		{NULL, NULL, NULL}
	};

	/* parse options */
	if ((msg = parse_opts(argc, argv, options, &help)) != NULL)
		tst_brkm(TBROK, cleanup, "OPTION PARSING ERROR - %s", msg);

	if (opt_d)
		conf_device_name = str_d;

	if (opt_i) {
		if (sscanf(str_i, "0x%x", &conf_can_id) != 1)
			tst_resm(TWARN,
				"Hex numbers should begin with '0x'. Using default CAN ID 0x%x.", DEFAULT_CAN_ID);
	}

	if (opt_b) {
		if (sscanf(str_b, "%u", &conf_bitrate) != 1) {
			tst_resm(TWARN,
				"Bitrate could not be detected. Using default bitrate %u bit/s.", DEFAULT_BITRATE);
		}
	}

	if (opt_n) {
		if (sscanf(str_n, "%u", &conf_message_count) != 1) {
			tst_resm(TWARN,
				"Message count could not be detected. Using default message count %u.", DEFAULT_MESSAGE_COUNT);
		}
	}

	if (opt_w) {
		if (sscanf(str_w, "%u", &conf_wait_us) != 1)
			tst_resm(TWARN,
				"Wait parameter could not be detected. Using default wait time %u us.", DEFAULT_WAIT_US);
	}

	if (opt_o) {
		if (sscanf(str_o, "%u", &conf_timeout_ms) != 1)
			tst_resm(TWARN,
				"Timeout parameter could not be detected. Using default timeout %u ms.", DEFAULT_TIMEOUT_MS);
	}

	if (opt_q)
		conf_quiet = 1;

	/* do initialization */
	setup();

	/* start the test */
	for (lc=0; TEST_LOOPING(lc); lc++) {
		ret = test();

		/* always tell the test result */
		if (ret != 0)
			tst_resm(TFAIL, "Test failed.");
		else
			tst_resm(TPASS, "Test passed.");
	}

	cleanup();
	tst_exit();
}
