#include <time.h>
#include <stdio.h>
#include <pthread.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>

#define VERSION						"1.0"
#define SVN_REV						"$Revision: 3627 $"

#define USEC(a) ((a) * 1000)
#define MSEC(a) ( USEC(a) * 1000)
#define SEC(a) ( MSEC(a) * 1000)

struct clk_stats {
	clockid_t clock_id;
	char * name;
	struct timespec start;
	struct timespec last;
	struct timespec now;
	struct timespec running;

	unsigned int count;
	
	int max_offset_period;
	int min_offset_period;
	int sum_offset_period;

	int max_offset_reference;
	int min_offset_reference;
	int sum_offset_reference;
};

static struct clk_stats clock_stats[] = 
{
	{ .clock_id = CLOCK_MONOTONIC		, .name = "monotonic"			},
	{ .clock_id = CLOCK_REALTIME		, .name = "realtime"			},
#ifdef CLOCK_MONOTONIC_RAW
	{ .clock_id = CLOCK_MONOTONIC_RAW	, .name = "monotonic raw"		},
#endif
#ifdef CLOCK_MONOTONIC_COARSE
	{ .clock_id = CLOCK_MONOTONIC_COARSE, .name = "monotonic coarse"	},
#endif
#ifdef CLOCK_REALTIME_COARSE
	{ .clock_id = CLOCK_REALTIME_COARSE	, .name = "realtime coarse"		},
#endif
};

static const int clock_count = sizeof(clock_stats)/sizeof(clock_stats[0]);
static volatile int thread_run;
static int sleeptime = MSEC(1);
static int runs = 20;
static char	version[10]             = VERSION;
static int verb = 0;

 
static struct timespec diff(struct timespec start, struct timespec end)
{
	struct timespec temp;
	if ((end.tv_nsec-start.tv_nsec)<0) {
		temp.tv_sec = end.tv_sec-start.tv_sec-1;
		temp.tv_nsec = 1000000000+end.tv_nsec-start.tv_nsec;
	} else {
		temp.tv_sec = end.tv_sec-start.tv_sec;
		temp.tv_nsec = end.tv_nsec-start.tv_nsec;
	}
	return temp;
}

static unsigned int diff_ns(struct timespec start, struct timespec end)
{
	struct timespec d = diff( start, end);

	return d.tv_nsec + SEC(d.tv_sec );
}



static struct timespec add(struct timespec a, struct timespec b)
{
	struct timespec temp = a;

	temp.tv_nsec += b.tv_nsec;
	temp.tv_sec += b.tv_sec;
	if( temp.tv_nsec >= SEC(1))
	{
		temp.tv_nsec -= SEC(1);
		temp.tv_sec += 1;
	}
	return temp;
}



static void update_clocknow( struct clk_stats * stats)
{
	if( clock_gettime( stats->clock_id, & stats->now))
	{
		printf("clock_gettime ID %d failed: %d %s\n", stats->clock_id, errno, strerror(errno));
		return;
	}
}
static void update_clockstats( struct clk_stats * stats, int default_period, struct timespec ref_now)
{
	int d, dref;

	stats->running = diff(stats->start, stats->now);	

	dref = diff_ns( stats->running, ref_now);
	d = diff_ns( stats->last, stats->now);
	d -= default_period;
	stats->count++;

	if( 1 == stats->count)
	{
		stats->start = stats->now;
		stats->last = stats->now;
		stats->sum_offset_period = 0;
		stats->sum_offset_reference = 0;
		return;
	}
	if( 2 == stats->count)
	{
		stats->max_offset_period = d;
		stats->min_offset_period = d;
		stats->max_offset_reference = dref;
		stats->min_offset_reference = dref;
	}

	if( d > stats->max_offset_period)
		stats->max_offset_period = d;
	if( d < stats->min_offset_period)
		stats->min_offset_period = d;
	stats->sum_offset_period += d;

	if( dref > stats->max_offset_reference)
		stats->max_offset_reference = dref;
	if( dref < stats->min_offset_reference)
		stats->min_offset_reference = dref;
	stats->sum_offset_reference += dref;
		
	stats->last = stats->now;
}
static void print_clockstats( struct clk_stats * stats)
{
	printf(" %d | %3d:%09d | %11.3f | % 11d | % 11d| %11.3f | % 11d| % 11d | %s\n" 
			, stats->clock_id, stats->running.tv_sec, stats->running.tv_nsec, 
			(float)stats->sum_offset_period / stats->count, stats->min_offset_period, stats->max_offset_period,
			(float)stats->sum_offset_reference/stats->count, stats->min_offset_reference, stats->max_offset_reference,
			stats->name
		);
}

 
/* this function is run by the second thread */
void * thread_entry(void * x)
{
	clockid_t clock_id = CLOCK_MONOTONIC;
	struct timespec wakeup, now, running, sleep;
	int ret, i;

	sleep.tv_sec = 0;
	sleep.tv_nsec = sleeptime;

	if( clock_gettime( clock_id, &now))
	{
		printf("clock_gettime failed: %d %s\n", errno, strerror(errno));
		return NULL;
	}
	wakeup = add(now, sleep);

	while(thread_run)
	{
		ret = clock_nanosleep( clock_id, TIMER_ABSTIME,  &wakeup  ,0);
		if( ret == EINTR )
			continue;
		if( ret)
		{
			printf(" clock_nanosleep returned: %d %s\n", ret, strerror(ret));
			return NULL;
		}
		wakeup = add(wakeup, sleep);

		for(i = 0; i < clock_count; i++)
			update_clocknow( &clock_stats[i]);

		running = diff(clock_stats[0].start, clock_stats[0].now);	
		for(i = 0; i < clock_count; i++)
			update_clockstats( &clock_stats[i], sleep.tv_nsec, running);
	}

	/* the function must return something - NULL will do */
	return NULL;

}

static void usage()
{
	printf("Time Test v%s\n", version);
	printf("Syntax: timerapi-test [-r <seconds to run>] [-w <wakeup period >] [-v]\n");
	printf("	-r	: number seconds the test runs (default: 10), 0 is infinite\n");
	printf("	-w	: wakeup interval in nanoseconds\n");
	printf("	-v	: print intermediate results\n");
	printf("	-h	: this help\n");
	exit(1);
}


static void parse_arguments(int argc, char **argv)
{
	int c;
	extern char *optarg;
	extern int optind;

	while ((c = getopt(argc, argv, ":h:w:r:")) != -1)
	{
		switch(c)
		{
		case 'h':
			usage();
			break;
		case 'w':
			sleeptime = atoi(optarg);
			break;
		case 'r':
			runs = atoi(optarg);
			break;
		case 'v':
			verb = 1;
			break;
		case '?':
			usage();
			break;
		}
	}
}

int main(int argc, char **argv)
{
	pthread_t thread;
	struct timespec time1, time2;
	int temp, i, j, ret;
	
	// Construct version string
	strcat(version, "r");
	strncat(version, strchr(SVN_REV, ' ')+1, strrchr(SVN_REV, ' ')-strchr(SVN_REV, ' ')-1);

	// Parse command line arguments
	parse_arguments(argc, argv);

	printf( "Thread wakeup period: %d ns ( %.3f ms)\n", sleeptime, (float) sleeptime / MSEC(1));
	printf( "Averaging %d seconds ", runs );
	fflush(stdout);

	thread_run = 1;

	if(pthread_create(&thread, NULL, thread_entry, NULL)) {

		fprintf(stderr, "Error creating thread\n");
		return 1;
	}
	

	for(j = 0; j < runs; j++)
	{
		sleep(1);
		if(verb)
		{
			printf("\n");
			for(i = 0; i < clock_count; i++)
				print_clockstats( &clock_stats[i]);
		}else
		{
			printf(".");
			fflush(stdout);
		}
	}

	printf("\nThread woke up %d times\n", clock_stats[0].count);

	printf("ID |   since start |    Offset from the given period (ns)   |Offset total time compared to Clk 0 (ns)| Clock name\n");
	printf("                   |    Avg      |    Min      |     Max    |     Avg     |    Min     |    Max      |           \n");

	for(i = 0; i < clock_count; i++)
		print_clockstats( &clock_stats[i]);

	/* wait for the second thread to finish */
	thread_run = 0;
	if(pthread_join(thread, NULL)) {

		fprintf(stderr, "Error joining thread\n");
		return 2;
	}

	return 0;
}
 
