#!/usr/bin/python3
#------------------------------------------------------------------------------#
#                                                                              #
#                                                                              #
#			                                                                   #
#                                                                              #
#                                                                              #
#                                                                              #
#------------------------------------------------------------------------------#

# documentation----------------------------------------------------------------#
DOC = """
=======================================
Garz und Fricke operating hours counter
=======================================

This tool implements and operating hours counter with the 
gfeeprom tool or a file as storage backend

The operating hours counter is a python program that updates one of two counter in the eeprom using the gfeeprom tool.
The data format of the eeprom area containing one counter:
The length is 8 bytes, plain display with the gfeeprom tool looks:
root@sdc-cspu:~# gfeeprom --show -q -ea 2 -hs
4f48430027000000

The meaning of the bytes:
0	4f	O	ID Tag
1	48	H	ID Tag
2	43	C	ID Tag
3	00	0	Version, currently 0, but may be changed if the data format has to be changed
4	29	41	LSB of 32 bit counter value
5	00	0	
6	00	0
7	00	0	MSB of 32 bit counter value

The increment of the counter happens in a interval configured in  /etc/ohc/ohc.conf. 
Default is 15 minutes. The above value means 10 hours, 15 minutes.

There are two counters used, one in area 2 and one in area 3.
The counter with the higher value is the valid one.

root@sdc-cspu:~# gfeeprom --show -q -ea 2 -hs
4f48430029000000
root@sdc-cspu:~# gfeeprom --show -q -ea 3 -hs
4f48430028000000

--> area 2 contains the actual counter value: 0x29.

"""
# defines ---------------------------------------------------------------------#
TOOL_NAME = 'ohc'
TOOL_LONG_NAME = 'Operating hours counter'
DEFAULT_CONFIG_FILE = '/etc/ohc/ohc.conf'
SVNVERSION='1.0'
AUTHOR="Jonas Hoeppner"
EMAIL="jonas.hoeppner@garz-fricke.com"
COUNTER_MAX=2
MAX_RETRY=COUNTER_MAX*4

# program modes
SERVICE=1
POLL_COUNTER=2
POLL_COUNTER_RAW=3
POLL_CONFIG=4
CLEAR_ALL=5


# gfeeprom return types
ERROR_AREA_NOT_FOUND=250
ERROR_CRC_FAILURE=224

verbose=False

# imports----------------------------------------------------------------------#

import sys
import os
import os.path
import subprocess
import re
import binascii
import getopt
import time
import signal
import configparser

#==============================================================================
#==============================================================================
class eepromExceptionType:
    ToolNotFound = 0
    ParameterError = 1
    WrongVersion = 2
    AreaNotFound = 3
    CrcFailure = 4

class eepromException(Exception):
    """Exception raised for errors eeprom tool access.

    Attributes:
        type -- input expression in which the error occurred
        msg  -- explanation of the error
    """
    def __init__(self, typeNo, msg):
        self.typeNo = typeNo
        self.msg = msg

class eepromAccess:
    minVersionMajor=1
    minVersionMinor=3
    defaultPath='gfeeprom'
    major=0
    minor=0
    max_areas=2

    def __init__(self, path=0):
        if(path!=0):
            self.__tool=path
        else:
            self.__tool=self.defaultPath

        self.__check_tool_version()
        
    def __check_tool_version(self):
        toolversion = self.__call_tool("--version")
        version_major = re.search('(?<=version)\s*\d*', toolversion.decode())
        version_minor = re.search('(?<=version'+version_major.group(0)+'\.)\d*', toolversion.decode())
        revision_m = re.search('(?<=\[\$Revision:)\s*\d*', toolversion.decode())
        self.revision = int(revision_m.group(0));
        self.major=int(version_major.group(0))
        self.minor=int(version_minor.group(0))
        #print("Found " + self.__tool + " with Version "+str(self.major)+ "." + str(self.minor) + " Revision: " + str(self.revision))
        if( self.major < self.minVersionMajor or ( self.major == self.minVersionMinor and self.minor < self.minVersionMinor)):
            raise eepromException(eepromExceptionType.WrongVersion, "Required Version: " + self.minVersionMajor + "." + self.minVersionMinor)
        return 0
        
    def __call_tool(self, args="", data=0):
        #print(self.__tool +" " +  args)
        arg_list = [self.__tool];
        for arg in args.split(" "):
            arg=arg.lstrip()
            if(len(arg)):
                arg_list.append(arg)

        #print(arg_list)
        try:
            ret = subprocess.check_output(arg_list)
        except subprocess.CalledProcessError as e:
            if(e.returncode == ERROR_AREA_NOT_FOUND):
                raise eepromException(eepromExceptionType.AreaNotFound, "Area does not exist: Return value: " + str(e.returncode) +".");
            elif(e.returncode == ERROR_CRC_FAILURE):
                raise eepromException(eepromExceptionType.CrcFailure, "CRC Failure while reading area: Return value: " + str(e.returncode) +".");
            else:
                raise eepromException(eepromExceptionType.ParameterError, "Call to " + self.__tool + " " + args + " failed: Return value: " + str(e.returncode) +".");
        except Exception as e:
            raise eepromException(eepromExceptionType.ToolNotFound, "Tool: " + self.__tool  + " not found.")
        return ret;

    def __indexToArea(self, index):
        if(index >= self.max_areas or index < 0):
            raise IndexError()
        return index + 2

    def read_data(self, index):
        area=self.__indexToArea( index);
        data = self.__call_tool("--show -hs -q -ea " + str(area))
        #print("Read 1:" + data.decode('ascii'))
        data = data.decode('ascii').strip()
        #print("Read 2:" + data)
        if( len(data) % 2):
            data = data[:len(data)-1]
        try:
            data = binascii.a2b_hex(data)
        except TypeError as e:
            print("Error in data format while reading from gfeeprom")
            data = ""
        except binascii.Error as e:
            print("Error in data format while reading from gfeeprom")
            data = ""
        return data;

    def write_data(self, index, data):
        area=self.__indexToArea( index);
        #print("Write 1:"+ data.decode('ascii'))
        data = binascii.b2a_hex(data)
        #print("Write 2:" + data.decode('ascii'))
        data = self.__call_tool("--program -hs -q -ea " + str(area) + " -ss " + data.decode('ascii') + "")
        return 0

#========================================================================================
#========================================================================================

class CounterException(Exception):
    """Exception raised for counter data format errors

    Attributes:
        msg  -- explanation of the error
    """
    def __init__(self, msg):
        self.msg = msg

class Usage(Exception):
    """Exception raised for counter data format errors

    Attributes:
        msg  -- explanation of the error
    """
    def __init__(self, msg):
        self.msg = msg

#========================================================================================
#========================================================================================

class Counter:
    """ Defines the structure of a counter value and the its serialization
    """
    __counter = 0
    __version=0
    __tag=b'OHC'
    __bytelen=8

    def __init__(self, data=0):
        """ Init counter from byte array as generated by toBytes() or initilize counter to 0
        """        
        self.__counter = 0;
        if( data != 0):
            self.fromBytes(data)

    def increment(self):
        self.__counter+=1

    def setCounter(self, value):
        self.__counter=value

    def getCounter(self):
        return self.__counter

    def toBytes(self):
        data = []
        data.append(self.__tag[0]& 0xff)
        data.append(self.__tag[1]& 0xff)
        data.append(self.__tag[2]& 0xff)
        data.append(self.__version & 0xff)
        data.append(((self.__counter >> 0 )& 0xff))
        data.append(((self.__counter >> 8 )& 0xff))
        data.append(((self.__counter >>16 )& 0xff))
        data.append(((self.__counter >>24 )& 0xff))
        if len(data) < self.__bytelen:
            raise(CounterException("Data length mismatch"))
        return bytes(data)

    def fromBytes(self,data):
        if len(data) < self.__bytelen:
            raise(CounterException("Data length mismatch"))
        for i in range(0,3):
            if(data[i] != self.__tag[i]):
                raise(CounterException("Tag mismatch"))
        if(data[3] != self.__version):
            raise(CounterException("Version mismatch"))
        self.__counter = data[4] | data[5] << 8 | data[6] << 16 | data[7] << 24

class StoredCounter(Counter):
    """ Defines on instance of a counter with the related storage functions
    """
    def __init__(self, storage, index):
        self.__index = index
        self.__storage = storage
        counter = self.__openCounter()
        Counter.__init__(self)
        self.setCounter(counter.getCounter())
        
    def __openCounter(self ):
        """ Opens the stored counter representation, initializes it if it fails
        """
        assert( self.__index < self.__counter_max)

        try:
            data = self.__storage.read_data(self.__index);
        except eepromException as e:
            if e.typeNo == eepromExceptionType.AreaNotFound or e.typeNo == eepromExceptionType.CrcFailure:
                c = Counter();
                self.__storage.write_data( self.__index , c.toBytes())
                data = self.__storage.read_data( self.__index );
            else:
                raise(e)
        try:
            counter = Counter(data)
        except CounterException as e:
            # TODO may add force flag to not just override the data that didn't matched
            c = Counter();
            self.__storage.write_data( self.__index , c.toBytes())
            data = self.__storage.read_data(self.__index );
            counter = Counter(data)
        return counter

    def __write(self):
        self.__storage.write_data( self.__index , self.toBytes())

    def __read(self):
        self.fromBytes( self.__storage.read_data( self.__index))

    def writeAndVerify(self, value):
        self.setCounter(value)
        val = self.getCounter()
        self.__write()
        self.__read()
        if(val != self.getCounter()):
            return 1
        return 0

    def getIndex(self):
        return self.__index
            
        

#========================================================================================
#========================================================================================

class OHCounter:
    def __init__(self,storage, counter_max=2):

        self.__counter_max=counter_max
        # Open counters in storage, 
        self.__counter=[]
        for i in range(0, self.__counter_max):
            self.__counter.append( StoredCounter(storage, i ))

    def __getHighestCounter(self):
        cm=self.__counter[0]
        for c in self.__counter:
            if cm.getCounter() < c.getCounter():
                cm=c
        return cm

    def __getLowestCounter(self):
        cm=self.__counter[0]
        for c in self.__counter:
            if cm.getCounter() > c.getCounter():
                cm=c
        return cm

    def __getNextLowestCounter(self, previous):
        cm=self.__counter[0]
        if(cm == previous):
            cm=self.__counter[1]
        for c in self.__counter:
            if(c == previous):
                continue
            if cm.getCounter() > c.getCounter():
                cm=c
        return cm
                
    def getCounterValue(self):
        return self.__getHighestCounter().getCounter()
                
    def incrementValue(self):
        current = self.getCounterValue()
        nextCounter = self.__getLowestCounter()
        i=0
        while(1):
            if( 0 == nextCounter.writeAndVerify(current + 1)):
                break
            i=i+1
            if(i>MAX_RETRY):
                raise CounterException("Failed to update counter, too many retries")
            print("Failed to write counter " + str(nextCounter.getIndex()) + ". Try next one.")
            nextCounter = self.__getNextLowestCounter(nextCounter)
        
    def clearAll(self):
        self.__counter=[]
        for c in self.__counter:
            if(not 0 == c.writeAndVerify(0)):
                raise CounterException("Counter " + str(c.getIndex()) + "couldn't be cleared")
        

#========================================================================================
#========================================================================================

def display_time(seconds, granularity=2):
    result = []
    intervals = (
        ('weeks', 604800),  # 60 * 60 * 24 * 7
        ('days', 86400),    # 60 * 60 * 24
        ('hours', 3600),    # 60 * 60
        ('minutes', 60),
        ('seconds', 1),
        )
    if seconds == 0:
        return "0 seconds"

    for name, count in intervals:
        value = seconds // count
        if value:
            seconds -= value * count
            if value == 1:
                name = name.rstrip('s')
            result.append("{} {}".format(value, name))
    return ', '.join(result[:granularity])

def printVersion():
    print(TOOL_NAME + " '" + TOOL_LONG_NAME +  "' Version " + str(SVNVERSION) + " <" + EMAIL + ">")


def printUsage():
    print("Usage:")
    print("\tpython3 ohc.py [-d|-c|-h|-v] [-V] [-f <conf_file>]")

def printHelp():
    print(TOOL_LONG_NAME)
    printVersion()
    printUsage()
    print("Called without parameter returns current operating hours value")
    print("-d\t--deamon                     starts the deamon ")
    print("-c\t--config                     returns overview of the configuration")
    print("-h\t--help                       print this help")
    print("-v\t--version                    print tool version")
    print("-V\t--verbose                    more output")
    print("-f\t--configfile=<configfile>    use given configfile instead of" + str(DEFAULT_CONFIG_FILE) )
    print("  \t--clearall                   Initialize counter to 0" )
    print("-r\t--raw                        Prints the raw counter value" )
        
def signal_handler(signal, frame):
    print('Operating hours counter is exiting')
    # Cleanup code
    sys.exit(0)

def print_verb(txt):
    global verbose
    if verbose:
        print(txt)

def main(argv=None):
    configfile=DEFAULT_CONFIG_FILE
    mode=POLL_COUNTER
    global verbose
    verbose=False

    if argv is None:
        argv = sys.argv
    try:
        try:
            opts, args = getopt.gnu_getopt(argv[1:], "hf:dcvVr", ["help", "configfile=","deamon","config","version","clearall", "verbose", "raw"])
        except getopt.GetoptError as e:                
             raise Usage(e)
        
        for o, a in opts:
            if o in ("-h", "--help"):
                printHelp()
                return 0
            elif o in ("-v", "--version"):
                printVersion ()
                return 0
            elif o in ( "-V", "--verbose"):
                verbose = True
            elif o in ("-f", "--configfile"):
                configfile=a
            elif o in ("-d", "--deamon"):
                mode=SERVICE
            elif o in ("-c", "--config"):
                mode=POLL_CONFIG
            elif o in ("  ", "--clearall"):
                mode=CLEAR_ALL
            elif o in ("-r", "--raw"):
                mode=POLL_COUNTER_RAW

        # parse config file
        config = configparser.ConfigParser()
        print_verb('Reading config file: ' + configfile)
        try:
            files_read = config.read(configfile)
            if configfile not in files_read:
                print("Failed to open config file: " + configfile)
                print("Using default values")
        except configparser.Error as e:
            print("Failed to parse config file: " + configfile)
            print("Using default values")
        
        gfeeprom = config.get('Options','gfeeprom_tool',fallback='/usr/bin/gfeeprom')
        counterInterval = config.getint('Options','counter_interval',fallback=900)
        
        try:
            storage=eepromAccess(gfeeprom)
        except eepromException as e:
            print("Initialize EEPROM access failed:")
            print(e.msg)
            exit(1)

        try:

            try:
                ohc = OHCounter(storage,COUNTER_MAX)
            except CounterException as e:
                print("Failed to read or initialize counters")
                print(e.msg)
                exit(1)

            if mode == POLL_COUNTER:
                    val = ohc.getCounterValue()
                    val = counterInterval * val
                    print( display_time(val) )
                    return(0)
            if mode == POLL_COUNTER_RAW:
                    print(str(ohc.getCounterValue()))
                    return(0)
            if mode == POLL_CONFIG:
                    print("Storage used:       " + gfeeprom)
                    print("Counter Interval:   " + display_time(counterInterval) )
                    print("Config file:        " + configfile )
                    return(0)
            elif mode == CLEAR_ALL:
                    ohc.clearAll()
                    return(0)
            
            if not mode == SERVICE:
                print("This function is not implemented yet");
                return(1)

            # Loop that builds the service
            signal.signal(signal.SIGINT, signal_handler)

            print_verb ("Start : %s" % time.ctime())
            while(1):
                time.sleep(counterInterval)
                ohc.incrementValue()
                print_verb( "Woke up: %s" % time.ctime() + " Clock %s" % time.time() + " New counter:" + str(ohc.getCounterValue()))
            
        except eepromException as e:
            print("EEPROM access failed:")
            print(e.msg)
            return(1)



        # done-------------------------------------------------------------------------#
	
    except Usage as err:
        print (err.msg)
        printUsage()
        return 2

if __name__ == "__main__":
    sys.exit(main())

