/******************************************************************************/
/*                                                                            */
/* pman - power management for Garz & Fricke LIVIUS                           */
/*                                                                            */
/* This program is free software; you can redistribute it and/or modify it    */
/* under the terms of the GNU General Public License version 2 as published   */
/* by the Free Software Foundation.                                           */
/*                                                                            */
/* This power management tool is meant to run on Garz & Fricke LIVIUS.        */
/* The tool provides an adaptive and predictive power management based on     */
/* weather forecast data.                                                     */
/*                                                                            */
/* Author:      Richard Guenther <richard.guenther@haw-hamburg.de             */
/*                                                                            */
/******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <float.h>
#include <sys/stat.h>
#include <getopt.h>

#include "pman.h"
#include "sysfs_utils.h"
#include "weather.h"

#define _GNU_SOURCE /* Must be defined to enable getopt_long functinality. */

/* default buffer sizes */
#define MAX_PATH_LENGTH			75
#define MAX_MSG_LENGTH			75

/* GPIOs */
#define RUN_VBAK5V			(1 * 32 + 11)
#define CHARGE_STAT1		(0 * 32 + 8)
#define CHARGE_DONE			(0 * 32 + 9)
#define CHARGE_DISABLE		(1 * 32 + 30)
#define VCC_CAN_EN			(2 * 32 + 2)
#define GSM_EN				(2 * 32 + 4)
#define VBATT_EN			(2 * 32 + 3)
#define VBAK5V_EN			(1 * 32 + 17)
#define KB_LED_EN			(1 * 32 + 10)

/* ADC channels */
#define ADC_CHANNEL_VIN				1
#define ADC_CHANNEL_BATTERY			5				
#define ADC_CHANNEL_VIN_CURRENT		6
#define ADC_CHANNEL_RTC_BATTERY		7

/* battery data */
#define BAT_VOLTAGE_MAX			13.8	/* [V] */
#define BAT_VOLTAGE_MIN			10.5	/* [V] */
#define BAT_CAPACITY			7200	/* [Ah] */
#define BAT_ENERGY_MAX			86400	/* [Wh] */
#define BAT_PERCENT_BASE		(BAT_VOLTAGE_MAX - BAT_VOLTAGE_MIN)	/ 100

/* solar module data */
#define SOLAR_POWER_RATING		17.5		/* [V] */
#define SOLAR_SURFACE_AREA		0.146475	/* [m^2] */
#define SOLAR_MUE				0.04096		

/* prediction defines */
#define PREDICTION_HORIZON		14400	/* [sec] */
#define LEVEL_CHANGE_THRESHOLD	42.4	/* [%] */

/* Log file paths */
#define LOG_PATH_STATES 		"/home/root/current_state"
#define LOG_PATH_MEASUREMENT	"/home/root/measurements"
#define LOG_PATH_PREDICTION		"/home/root/prediction"

/* defines and function for printing messages */
#define V_INFO	3
#define V_DEBUG	2
#define V_WARN	1
#define V_ERROR 0 

/* print function for verbose messages */
#define PRINT_INFO(verbosity_level, fmt, args...) do { 		\
	if (verbosity >= verbosity_level) 				\
		fprintf(stderr, "%s: " fmt, __progname, ## args);  	\
} while (0)

/* print function for error messages */
#define PRINT_ERROR(errno, fmt, args...) \
	fprintf(stderr, "%s: ERROR %d: %s: " fmt, __progname, errno, __func__, ## args);

/******************************************************************************/
/*                                                                            */
/* global variables                                                           */
/*                                                                            */
/******************************************************************************/
int verbosity = 1;			/* warnings enabled by default */
int logging = 0;			/* logging disabled by default */

/******************************************************************************/
/*                                                                            */
/* struct definitions                                                         */
/*                                                                            */
/******************************************************************************/
struct system_state {
	struct power_lvl *		plevel;				/* power level */ 
	time_t					time;				/* current system time (UNIX format) */
	float					bat_energy;			/* available energy in main battery [Wh] */ 
	float					bat_percent;		/* main battery state of charge [%] */ 
	float					v_backup_bat;		/* backup battery voltage [V] */  
	float					v_main_bat;			/* main battery voltage [V] */ 
	float					i_main_bat;			/* main battery current [A] */ 
	
	struct system_state * 	next;				/* pointer to the next system state */
	struct system_state * 	prev;				/* pointer to the previous system state */
};

struct power_lvl {
	int		level;				/* level */
	int 	active_interval;	/* activity cycle interval */
	int 	power_consumption;	/* power consumption */
};

/******************************************************************************/
/*                                                                            */
/* enumerations                                                               */
/*                                                                            */
/******************************************************************************/

/* available power levels */
enum powerlevels {P0, P1, P2, P3, P4, P5, P6, P7, P8, P9, P10};

/******************************************************************************/
/*                                                                            */
/* tables                                                                     */
/*                                                                            */
/******************************************************************************/
/* 
 * ----------------------------------------
 * linear sleep time characteristic  
 * step: 342 s
 * ----------------------------------------
 */
struct power_lvl power_level_table1[] = {
/*  | Level | sleep | power |
 *  |       | [sec] | [mW]  |*/	
	{ P0, 	  3600,   671 },
	{ P1, 	  3258,   676 },	
	{ P2, 	  2916,   682 },
	{ P3, 	  2574,   689 },
	{ P4, 	  2232,   699 },
	{ P5, 	  1890,   716 },
	{ P6, 	  1548,	  732 },
	{ P7, 	  1206,   762 },
	{ P8, 	  864, 	  817 },
	{ P9, 	  522, 	  943 },
	{ P10, 	  180, 	  1550 },
};

/* 
 * -----------------------------------------------
 * mostly linear power consumption characteristic:
 * step: 0.085 mW
 * -----------------------------------------------
 */
struct power_lvl power_level_table2[] = {
/*  | Level | sleep | power |
 *  |       | [sec] | [mW]  |*/	
	{ P0,     3600  ,  656  },
	{ P1,     1449  ,  702  },
	{ P2,     813   ,  762  },
	{ P3,     565   ,  822  },
	{ P4,     480   ,  856  },
	{ P5,     420   ,  889  },
	{ P6,     360   ,  934  },
	{ P7,     300   ,  995  },
	{ P8,     240   ,  1088 },
	{ P9,     180   ,  1242 },
	{ P10,    120   ,  1550 },
};

/* 
 * --------------------------------------------------
 * lower power consumption, linear sleep time characteristic
 * step: 300 s
 * --------------------------------------------------
 */
struct power_lvl power_level_table3[] = {
/*  | Level | sleep | power |
 *  |       | [sec] | [mW]  |*/	
	{ P0,     3600,   656   }, 
	{ P1,     3300,   659   }, 
	{ P2,     3000,   662   }, 
	{ P3,     2700,   666   }, 
	{ P4,     2400,   671   }, 
	{ P5,     2100,   678   }, 
	{ P6,     1800,   687   }, 
	{ P7,     1500,   699   }, 
	{ P8,     1200,   718   }, 
	{ P9,     900,    748   }, 
	{ P10,    600,    810   }, 
};

/******************************************************************************/
/*                                                                            */
/* Function:    get_next_active_time                                          */
/*                                                                            */
/* Description: Returns the next active time that is scheduled for the given  */
/*              system state.                                                 */
/*                                                                            */
/* Input:       @state: pointer to the system state                           */
/*                                                                            */
/* Return:      @next_active_time: time value of next activity (UNIX format)  */
/*                                                                            */
/******************************************************************************/
time_t get_next_active_time(struct system_state *state)
{
	int i;
	time_t next_active_time;

	if(state == NULL) {
		PRINT_ERROR(EINVAL, "invalid pointer to system state\n");
		return -EINVAL;
	}
	
	/* get next active time from power level */
	next_active_time = (state->time + state->plevel->active_interval);
	
	PRINT_INFO(V_INFO, "next active time will be: %s", ctime(&next_active_time));
	
	return next_active_time;
}

/******************************************************************************/
/*                                                                            */
/* Function:    log_to_file                                                   */
/*                                                                            */
/* Description: writes a message to a file at the given path with a time      */
/*              stamp.                                                        */
/*                                                                            */
/* Input:       * message to write                                            */
/*              * path to the file                                            */
/*                                                                            */
/* Return:      On failure - negative errno return code                       */
/*              On success - '0' return code.                                 */
/*                                                                            */
/******************************************************************************/
static int log_to_file(char *file_path, char *msg)
{
	FILE *log_file;
	time_t raw_time;
	struct tm *ptm;
	
	char *log_msg = malloc(MAX_MSG_LENGTH + strlen(msg));
	if (log_msg == NULL) {
		PRINT_ERROR(ENOMEM, "log message memory allocation failed");
		return -ENOMEM;
	}
	
	time(&raw_time);
	ptm = localtime(&raw_time);
	
	/* generate timestamp */
	snprintf(log_msg, MAX_MSG_LENGTH, "[%.2d/%.2d/%d %.2d:%.2d:%.2d] ",
						ptm->tm_mday, (ptm->tm_mon)+1, (ptm->tm_year)+1900, 
						ptm->tm_hour, ptm->tm_min, ptm->tm_sec);
	
	/* concatenate log message */
	strcat(log_msg, msg);
	
	/* write message to logfile */
	log_file = fopen(file_path, "ab+");
	if(!log_file)
		return -EIO;
	fprintf(log_file, log_msg);
	fclose(log_file);
	
	return 0;
}
	
/******************************************************************************/
/*                                                                            */
/* Function:    log_system_state                                              */
/*                                                                            */
/* Description: writes data of the given system state to the specified log    */
/*              file path.                                                    */
/*                                                                            */
/* Input:       @file_path: path to the location of the log file              */
/*              @state    : pointer to the system state to log                */
/*                                                                            */
/* Return:      On failure - negative errno return code                       */
/*              On success - '0' return code.                                 */
/*                                                                            */
/******************************************************************************/
static int log_system_state(char *file_path, struct system_state *state)
{
	int ret;
	time_t system_time, next_active_time;
	
	char *log_msg = malloc(MAX_MSG_LENGTH);
	if (log_msg == NULL) {
		PRINT_ERROR(ENOMEM, "log message memory allocation failed");
		return -ENOMEM;
	}
	
	system_time = state->time;
	next_active_time = get_next_active_time(state);
	
	ret = log_to_file(file_path, "--------------------------------------------------\n");
	/* system time */
		snprintf(log_msg, MAX_MSG_LENGTH,  "system time: %s", ctime(&system_time));
	ret = log_to_file(file_path, log_msg);
	
	/* power level */
		snprintf(log_msg, MAX_MSG_LENGTH,  "power level: %d\n", state->plevel->level);
	ret = log_to_file(file_path, log_msg);
		snprintf(log_msg, MAX_MSG_LENGTH,  "  power consumption: %d\n", state->plevel->power_consumption);
	ret = log_to_file(file_path, log_msg);
		snprintf(log_msg, MAX_MSG_LENGTH,  "  active interval: %d\n", state->plevel->active_interval);
	ret = log_to_file(file_path, log_msg);
		snprintf(log_msg, MAX_MSG_LENGTH,  "  next active time: %s", ctime(&next_active_time));
	ret = log_to_file(file_path, log_msg);

	/* main battery */
		snprintf(log_msg, MAX_MSG_LENGTH,  "main battery percent: %.4f%%\n", state->bat_percent);
	ret = log_to_file(file_path, log_msg);
		snprintf(log_msg, MAX_MSG_LENGTH,  "main battery energy: %.4f mWh\n", state->bat_energy);
	ret = log_to_file(file_path, log_msg);
		snprintf(log_msg, MAX_MSG_LENGTH,  "main battery voltage: %.4f V\n", state->v_main_bat);
	ret = log_to_file(file_path, log_msg);
		snprintf(log_msg, MAX_MSG_LENGTH,  "main battery current: %.4f mA\n", (state->i_main_bat)*1000);
	ret = log_to_file(file_path, log_msg);
	
	/* backup battery */
	snprintf(log_msg, MAX_MSG_LENGTH,  "backup battery voltage: %.4f V\n", state->v_backup_bat);
	ret = log_to_file(file_path, log_msg);
	ret = log_to_file(file_path, "--------------------------------------------------\n");

	free(log_msg);
	return ret;
}

/******************************************************************************/
/*                                                                            */
/* Function:    log_system_state_list                                         */
/*                                                                            */
/* Description: prints out the time of all list elements                      */
/*                                                                            */
/* Input:       @head: pointer to the head of the list                        */
/*                                                                            */
/* Return:      On failure - negative errno return code                       */
/*              On success - '0' return code.                                 */
/*                                                                            */
/******************************************************************************/
int log_system_state_list(char *file_path, struct system_state *head) {
    struct system_state *current;
	time_t time;
	
	if(head == NULL) {
		PRINT_ERROR(EINVAL, "invalid pointer to list head\n");
		return -EINVAL;
	}
	
	current = head;
	
    while (current != NULL) {
		log_system_state(file_path, current);
        current = current->next;
    }
}

/******************************************************************************/
/*                                                                            */
/* Function:    get_current_system_state                                      */
/*                                                                            */
/* Description: update values of the given system state by reading supply     */
/*              voltages and GPIOs                                            */
/*                                                                            */
/* Input:       @state: pointer to the system state to update                 */
/*                                                                            */
/* Return:      On failure - negative errno return code                       */
/*              On success - '0' return code.                                 */
/*                                                                            */
/******************************************************************************/
static int get_current_system_state(struct system_state *state)
{
	float v_backup_bat, v_main_bat, i_main_bat, bat_percent, bat_energy;
	float v_bat_1, v_bat_2;
	
	if(state == NULL) {
		PRINT_ERROR(EINVAL, "invalid pointer to system state\n");
		return -EINVAL;
	}
	
	/* read ADC values */
	v_backup_bat = sysfs_adc_read_channel(ADC_CHANNEL_BATTERY);
	i_main_bat = sysfs_adc_read_channel(ADC_CHANNEL_VIN_CURRENT);
	
	
	v_bat_1 = sysfs_adc_read_channel(ADC_CHANNEL_VIN);
	sleep(1);
	v_bat_2 = sysfs_adc_read_channel(ADC_CHANNEL_VIN);
	
	v_main_bat = (v_bat_1 + v_bat_2) / 2;
	
	/* calculate voltage/current (due to hardware) */
	v_main_bat = v_main_bat * 1.386;
	i_main_bat = i_main_bat / (0.022*100);
	
	/* calculate percent and energy values */
	bat_percent = ((v_main_bat - BAT_VOLTAGE_MIN) / (BAT_VOLTAGE_MAX - BAT_VOLTAGE_MIN)) * 100;
	bat_energy = BAT_ENERGY_MAX * (bat_percent / 100);
	
	/* write new values*/
	state->v_backup_bat = v_backup_bat;
	state->v_main_bat = v_main_bat;
	state->i_main_bat = i_main_bat;		
	state->bat_percent = bat_percent;
	state->bat_energy = bat_energy;
	
	return 0;
}

/******************************************************************************/
/*                                                                            */
/* Function:    freeze_system                                                 */
/*                                                                            */
/* Description: sets the device into "freeze" mode                            */
/*                                                                            */
/******************************************************************************/
int freeze_system(void)
{
	return system("echo freeze > /sys/power/state");
}

/******************************************************************************/
/*                                                                            */
/* Function:    do_active_work                                                */
/*                                                                            */
/* Description: simulates active work by generating CPU load for 85 seconds   */
/*                                                                            */
/******************************************************************************/
int do_active_work(void)
{
	system("dd if=/dev/zero of=/dev/null &");
	sleep(56);
	system("killall dd");
}

/******************************************************************************/
/*                                                                            */
/* Function:    set_power_level                                               */
/*                                                                            */
/* Description: writes the given power level to the given system state.       */
/*                                                                            */
/* Input:       @state: pointer to the system state to update                 */
/*              @level: power level to write                                  */
/*                                                                            */
/* Return:      On failure - negative errno return code                       */
/*              On success - '0' return code.                                 */
/*                                                                            */
/******************************************************************************/
int set_power_level(struct system_state *state, int level)
{
	if(state == NULL) {
		PRINT_ERROR(EINVAL, "invalid pointer to system state\n");
		return -EINVAL;
	}
	
	/* do not exceed array borders */
	if (level > P10) {
		level = P10;
	}
	else if (level < P0) {
		level = P0;
	}
	
	PRINT_INFO(V_INFO, "setting power level to %d\n", level);

	//state->plevel = &power_level_table1[level];
	//state->plevel = &power_level_table2[level];
	state->plevel = &power_level_table3[level];
}

/******************************************************************************/
/*                                                                            */
/* Function:    gsm_startup                                                   */
/*                                                                            */
/* Description: boot the GSM modem by pulling the ignition pin to high for    */
/*              5 seconds                                                     */
/*                                                                            */
/* Input:       None                                                          */
/*                                                                            */
/* Return:      On failure - negative errno return code                       */
/*              On success - '0' return code.                                 */
/*                                                                            */
/******************************************************************************/
int gsm_startup(void)
{
	int ret;
	
	ret = sysfs_gpio_write_value(GSM_EN, 1);
	sleep(5);
	ret = sysfs_gpio_write_value(GSM_EN, 0);
	
	return ret;
}

/******************************************************************************/
/*                                                                            */
/* Function:    toggle_active_mode                                            */
/*                                                                            */
/* Description: apply a power level to the device. This will mainly toggle    */
/*              peripheral devices to reduce or gain power consumption        */
/*                                                                            */
/* Input:       @toggle: switch active mode on or off                         */
/*                                                                            */
/* Return:      On failure - negative errno return code                       */
/*              On success - '0' return code.                                 */
/*                                                                            */
/******************************************************************************/
int toggle_active_mode(int toggle)
{
	int ret;

	/* peripheral control switches */	
	ret = sysfs_gpio_write_value(CHARGE_DISABLE, !toggle);
	ret = sysfs_gpio_write_value(VCC_CAN_EN, toggle);
	ret = sysfs_gpio_write_value(VBATT_EN, toggle);
	
	if(toggle) {
		ret = gsm_startup();
	}
	
	return ret;
}

/******************************************************************************/
/*                                                                            */
/* Function:    predict_next_system_state                                     */
/*                                                                            */
/* Description: Predicts values for the next system state.                    */
/*              A model of the solar panel and the main battery are used to   */
/*              calculate expected battery and main supply values for the     */
/*              next active time                                              */
/*                                                                            */
/* Input:       @current_state: pointer to the system state to start from     */
/*              @weather_data : pointer to the weather forecast data array    */
/*                                                                            */
/* Return:      On failure - negative errno return code                       */
/*              On success - '0' return code.                                 */
/*                                                                            */
/******************************************************************************/
int predict_next_system_state(struct system_state *current_state, struct weather_entry *weather_data)
{
	struct system_state *next_state;
	struct weather_entry *w_entry;
	
	float bat_energy_charge = 0, bat_energy_discharge = 0; 
	float solar_power, time_interval;
	float next_vbat, next_bat_energy, next_bat_percent, next_ibat;	
	int harvest_start_time, harvest_end_time;
	int i;
	
	if(current_state == NULL) {
		PRINT_ERROR(EINVAL, "invalid pointers to system state\n");
		return -EINVAL;
	}
	
	if(current_state->next == NULL) {
		PRINT_ERROR(EINVAL, "invalid pointers to next system state (list corrupted)\n");
		return -EINVAL;
	}
	
	if(weather_data == NULL) {
		PRINT_ERROR(EINVAL, "invalid pointers to weather data array\n");
		return -EINVAL;
	}
	
	next_state = current_state->next;
	w_entry = weather_data;
	
	/* update time for future system state */
	next_state->time = get_next_active_time(current_state);
	
	/* 
	 * calculate energy harvest from solar panel:
	 * ========================================== 
	 */
	next_bat_energy = current_state->bat_energy;
	next_bat_percent = current_state->bat_percent;
	next_vbat = current_state->v_backup_bat;
	
	/* start and end time for energy harvesting */
	harvest_start_time = current_state->time;
	harvest_end_time = next_state->time - 1;

	/* parse weather data array */
	while((w_entry+1)->time != 0) {
		
		/* interval is negative: exit */
		if(w_entry->time > harvest_end_time) {
			break;
		}
		/* interval is between two data entries */
		else if ((harvest_start_time > w_entry->time) && (harvest_end_time < (w_entry+1)->time)) {
			time_interval = (float)(harvest_end_time - harvest_start_time);
		}
		/* complete data entry within the interval */
		else if ((harvest_start_time == w_entry->time) && (harvest_end_time >= (w_entry+1)->time)) {
			time_interval = (float)((w_entry+1)->time - w_entry->time);
			harvest_start_time = (w_entry+1)->time;
		}
		/* interval: start time -> next data entry time */
		else if((harvest_start_time > w_entry->time) && (harvest_start_time < (w_entry+1)->time) && (harvest_end_time >= (w_entry+1)->time)) {
			time_interval = (float)((w_entry+1)->time - harvest_start_time);
			harvest_start_time = (w_entry+1)->time;
		}
		/* interval: last data entry time -> end time */
		else if((harvest_start_time == w_entry->time) && (harvest_end_time < (w_entry+1)->time)) {		
			time_interval = (float)(harvest_end_time - (w_entry->time));		
		}
		/* data entry is not within the interval */
		else {
			if((w_entry+1)->time > harvest_start_time)
				harvest_start_time = (w_entry+1)->time;
			w_entry++;
			continue;
		}
		
		/* calculate solar supply power */
		solar_power = ((float)(w_entry->condition) * SOLAR_SURFACE_AREA) * SOLAR_MUE;
		
		/* determine charging energy (integration) */
		bat_energy_charge += (solar_power * (time_interval / 3600));	
		
		/* move to the next data entry */
		w_entry++;
	}
	
	/* 
	 * calculate changes on battery charge 
	 * ==================================== 
	 */	
	bat_energy_discharge = ((float)(current_state->plevel->power_consumption * current_state->plevel->active_interval)) / 3600;
	
	PRINT_INFO(V_INFO, " total charge energy    : %.2f mWh\n", bat_energy_charge);
	PRINT_INFO(V_INFO, " total discharge energy : %.2f mWh\n", bat_energy_discharge);
	PRINT_INFO(V_INFO, " energy balance         : %.2f mWh\n", bat_energy_charge - bat_energy_discharge);
	
	/* calculate values for the next state */
	next_bat_energy += (float)(bat_energy_charge - bat_energy_discharge); 

	/* battery cannot carry more than max. capacity */
	if(next_bat_energy > BAT_ENERGY_MAX)
		next_bat_energy = BAT_ENERGY_MAX;
	
	next_bat_percent = (next_bat_energy / BAT_ENERGY_MAX) * 100;
	next_vbat = next_bat_percent * ((BAT_VOLTAGE_MAX - BAT_VOLTAGE_MIN) / 100) + BAT_VOLTAGE_MIN;
	
	/* write new system state values */
	next_state->bat_energy = next_bat_energy;
	next_state->bat_percent = next_bat_percent;
	next_state->v_main_bat = next_vbat;
	
	return 0;
}

/******************************************************************************/
/*                                                                            */
/* Function:    predict_multiple_system_states                                */
/*                                                                            */
/* Description: Predicts the future system state for the given time. For this */
/*              all other intermediate states between the start point and the */
/*              specified time need to be predicted                           */
/*                                                                            */
/* Input:       @start_state    : pointer to the system state to start from   */
/*              @prediction_time: time to predict the system state for        */
/*              @weather        : array with weather forecast data            */
/*                                                                            */
/* Return:      On failure - negative errno return code                       */
/*              On success - '0' return code.                                 */
/*                                                                            */
/******************************************************************************/
int predict_multiple_system_states(struct system_state *start_state, time_t prediction_time, struct weather_entry *weather_data)
{
	struct system_state *current_state, *next_state;
	int next_active_time;
	time_t time_predicted;
	int i, j;
	
	if(start_state == NULL) {
		PRINT_ERROR(EINVAL, "invalid pointer to list head\n");
		return -EINVAL;
	}
	
	if(prediction_time <= 0) {
		PRINT_ERROR(EINVAL, "invalid prediction horizon - must be positive integer\n");
		return -EINVAL;
	}
	
	current_state = start_state;
	
	/*
	 * This loop predicts future system states until reaching <prediction_time>
	 */
	while(prediction_time > time_predicted) {
	
		/* insert new element if list is too short */
		if(current_state->next == NULL) {
			add_list_element(current_state);
		}
		
		/* catch corrupted list */
		if(current_state == NULL || current_state->next == NULL) {
			PRINT_ERROR(EINVAL, "could not complete prediction. system state list ended unexpected\n");
			return -EINVAL;
		}
		
		/* predict future system states */
		predict_next_system_state(current_state, weather_data);	
		time_predicted = current_state->next->time;
		
		/* move pointer to next list element */
		current_state = current_state->next;
	}
	
	return 0;
}

/******************************************************************************/
/*                                                                            */
/* Function:    pop_list_element                                              */
/*                                                                            */
/* Description: Removes the first element of the list and moves the head      */
/*              pointer to the next element                                   */
/*                                                                            */
/* Input:       @head: pointer to pointer to the head of the list             */
/*                                                                            */
/* Return:      On failure - negative errno return code                       */
/*              On success - '0' return code.                                 */
/*                                                                            */
/******************************************************************************/
int pop_list_element(struct system_state **head) 
{
    struct system_state *next_element = NULL;
    
    if(*head == NULL) {
		PRINT_ERROR(EINVAL, "invalid pointer to list head\n");
		return -EINVAL;
	}

    next_element = (*head)->next;
    free(*head);
    *head = next_element;

    return 0;
}

/******************************************************************************/
/*                                                                            */
/* Function:    delete_list                                            	      */
/*                                                                            */
/* Description: deletes the whole list                                        */
/*                                                                            */
/* Input:       @head: pointer to the head of the list                        */
/*                                                                            */
/* Return:      On failure - negative errno return code                       */
/*              On success - '0' return code.                                 */
/*                                                                            */
/******************************************************************************/
int delete_list(struct system_state **head) 
{
    struct system_state *next, *deleteMe;
    
	if(*head == NULL) {
		PRINT_ERROR(EINVAL, "invalid pointer to list head\n");
		return -EINVAL;
	}
	
	deleteMe = *head;
    
	while (deleteMe) {
        next = deleteMe->next;
        free(deleteMe);
        deleteMe = next;
    }
    *head = NULL;
	
	return 0;
}

/******************************************************************************/
/*                                                                            */
/* Function:    add_list_element                                              */
/*                                                                            */
/* Description: Adds an element at the bottom of a doubly linked list         */
/*                                                                            */
/* Input:       @head: pointer to the head of the list                        */
/*                                                                            */
/* Return:      On failure - negative errno return code                       */
/*              On success - '0' return code.                                 */
/*                                                                            */
/******************************************************************************/
int add_list_element(struct system_state *head)
{
	struct system_state *newState = malloc(sizeof(struct system_state));
	struct system_state *current  = head;
	
	if (newState == NULL) {
		PRINT_ERROR(ENOMEM, "list element memory allocation failed");
		return -ENOMEM;
	}
	
	if(current == NULL) {
		PRINT_ERROR(EINVAL, "invalid pointer to list head\n");
		return -EINVAL;
	}
	
	/* first insertion */
	if(head->next == NULL) {
		head->next = newState;
		newState->prev = head;
	}
	/* find last element and insert the new one */
	else {
        while (1) {
			if(current->next == NULL)
			{
				current->next = newState;
				newState->prev = current;
				break;
			}
            current = current->next;
        };
    }
	
	/* pre set values */
	set_power_level(newState, current->plevel->level);
	newState->time = get_next_active_time(newState->prev);
	newState->bat_energy = 0;
	newState->bat_percent = 0;
	newState->i_main_bat = 0;
	newState->v_backup_bat = 0;
	newState->next = NULL;
	
    return 0;
}

/******************************************************************************/
/*                                                                            */
/* Function:    optimize_system_state_predicted                               */
/*                                                                            */
/* Description: This function optimizes the power levels of a list of system  */
/*              states by looking at the next system states.                  */
/*                                                                            */
/* Input:       @head: Pointer to the head of the list of system states       */
/*                                                                            */
/* Return:      On failure - negative errno return code                       */
/*              On success - '0' return code if the plevel was not changed    */
/*              On success - '1' return code if the plevel was changed        */
/*                                                                            */
/******************************************************************************/ 
int optimize_system_state_predicted(struct system_state *current_state, time_t prediction_time, struct weather_entry *weather_data)
{
	struct system_state *current;
	int finished = 0, optimized = 0;
	
	if(current_state == NULL) {
		PRINT_ERROR(EINVAL, "invalid pointer to list head\n");
		return -EINVAL;
	}
	
	predict_multiple_system_states(current_state, prediction_time, weather_data);

	while(finished == 0) {
		
		optimized = 0;
		/* generate future system state list */
		predict_multiple_system_states(current_state, prediction_time, weather_data);
		
		current = current_state;
		
		/* check for system state to optimize */
		while(current->next != NULL) {		
		
			optimized = optimize_power_level_recursive(current, 0);
			
			/* optimization error */
			if (optimized < 0) {
				PRINT_INFO(V_WARN, "an error occurred while optimizing\n");
				finished = 1;
				break;
			}
			/* optimization successful */
			else if(optimized == 1) {
				delete_list(&(current->next)); 
				break;
			}
			/* optimization reaches current state: cannot optimize any more */
			else if (optimized == 2) {
				PRINT_INFO(V_WARN, "opimization reached current system state\n");
				if(current->plevel->level == 0 || current->plevel->level == 7)
					finished = 1;
				break;
			}
			/* no need to optimize this state - switch to the next */
			current = current->next;
		}
		
		/* no state was optimized within the prediction horizon - we are done */
		if(optimized == 0) {
			finished = 1;
		}
	}
	
	return 0;
}

/******************************************************************************/
/*                                                                            */
/* Function:    optimize_power_level_recursive                                */
/*                                                                            */
/* Description: This function optimizes the power level of the given system   */
/*              state by looking at the next system state.                    */
/*                                                                            */
/* Input:       @state: Pointer to the system state                           */
/*                                                                            */
/* Return:      On failure - negative errno return code                       */
/*              On success - '0' return code if the plevel was not changed    */
/*              On success - '1' return code if the plevel was changed        */
/*              On success - '2' return code if state is current state and at */
/*                               power level 0                                */
/*                                                                            */
/******************************************************************************/ 
int optimize_power_level_recursive(struct system_state *state, int force) 
{
	int new_plevel, level_changed = 0;
	
	if(state == NULL || state->next == NULL) {
		PRINT_ERROR(EINVAL, "invalid pointer to system state\n");
		return -EINVAL;
	}

	new_plevel = state->plevel->level;
	
	/* ---------------------------------------------------- */
	/* prevent from going below 25% battery charge          */
	/* ---------------------------------------------------- */
	if((state->next->bat_percent < LEVEL_CHANGE_THRESHOLD) || force == 1 ) {
		new_plevel -= 1;
		
		/* optimize previous power level if current cannot be optimized any more */
		if (new_plevel < 0) {
			if (state->prev != NULL) {
				level_changed = optimize_power_level_recursive(state->prev, 1);	// recursive!
			}
			else {
				level_changed = 2;
			}
		}
		else {
			set_power_level(state, new_plevel);
			level_changed = 1;
		}
	}
	
	return level_changed;
}

/******************************************************************************/
/*                                                                            */
/* Function:    print_system_status                                           */
/*                                                                            */
/* Description: displays the values of the given system status                */
/*                                                                            */
/* Input:       @state: pointer to the system state to display                */
/*                                                                            */
/******************************************************************************/
int print_system_status(struct system_state *state)
{
	if(state == NULL) {
		PRINT_ERROR(EINVAL, "invalid pointer to system state\n");
		return -EINVAL;
	}
	
	printf("\n");
	printf("System status:\n");
	printf("  - system time             : %s", ctime(&(state->time)));
	printf("  - backup-battery          : %.2fV\n", state->v_backup_bat);
	printf("  - main supply             : %.4fV at %.2f mA \n", state->v_main_bat, state->i_main_bat * 1000);
	printf("Power level:\n");
	printf("  - current level           : %d\n", state->plevel->level);
	printf("  - activity interval       : %d seconds\n", state->plevel->active_interval);
	printf("  - avg. power consumption  : %d mW\n", state->plevel->power_consumption);
	printf("\n");
	
	return 0;
}

/******************************************************************************/
/*                                                                            */
/* Function:    show_help                                                     */
/*                                                                            */
/* Description: This function is called when the test is started with         */
/*              parameter -h. It displays all parameter options specifically  */
/*              available for this test.                                      */
/*                                                                            */
/******************************************************************************/	
void show_help()
{
	printf("\n");
	printf("Power management tool for Garz & Fricke LIVIUS.\n");
	printf("\n");
	printf("Usage: pman [OPTIONS]\n");
	printf("  -h, --help            show this help and exit\n");
	printf("  -s, --status          print current system status\n");
	printf("  -v, --verbose         increase verbosity level\n");
	printf("  -q, --quiet           decrease verbosity level\n");
}

/******************************************************************************/
/*                                                                            */
/* Function:    main                                                          */
/*                                                                            */
/* Description: Entry point to the program. It parses all the command line    */
/*              inputs, sets up the initial system state and runs the         */
/*              infinite loop of the program.                                 */                  
/*                                                                            */
/* Input:       None.                                                         */
/*                                                                            */
/* Return:      On failure - negative errno return code                       */
/*              On success - '0' return code.                                 */
/*                                                                            */
/******************************************************************************/	
int main(int argc, char* argv[])
{
	struct system_state *current_state = malloc(sizeof(struct system_state));
	struct weather_entry *weather_data_start; 
	struct weather_entry weather_data[48];
	int opt, i, show_status=0;
	time_t prediction_time, next_active_time;

	/* options definition */
	const char* const shortopts = "shvq";
	struct option longopts[] = {
		{"help"        , 0, NULL, 'h'},
		{"status"      , 0, NULL, 's'},
		{"verbose"     , 0, NULL, 'v'},
		{"quiet"       , 0, NULL, 'q'},
		{0, 0, 0, 0}  	  /* Array always terminates with a NULL option. */
	};

	/* parsing options */ 
	while((opt = getopt_long(argc, argv, shortopts, longopts, NULL)) != -1) {
		switch(opt) {
		case 's':
			show_status = 1;
			break;
		case 'h':
			show_help();
			return 1;
			break;
		case 'v':
			verbosity += 1;
			break;
		case 'q':
			verbosity -= 1;
			break;
		case ':':
			printf("ERROR: Option needs a value.\n");
			show_help();
			return 1;
			break;
		case '?':
			printf("ERROR: Unknown option: %c\n", optopt);
			show_help();
			return 1;
			break;
		default:
			show_help();
			return 1;
		break;
		}
	}
	
	/* always enable backup battery fall back */
	sysfs_gpio_write_value(VBAK5V_EN, 1);
	sysfs_gpio_write_value(KB_LED_EN, 1);
	
	/* set values for the first list element */
	get_current_system_state(current_state);
	current_state->next = NULL;
	current_state->prev = NULL;
	current_state->time = time(NULL);
	set_power_level(current_state, P10);

	/* print current status */
	if(show_status) {
		print_system_status(current_state);
		return 0;
	}
	
	/* parse weather data array */
	for(i=0; i<48; i++) {
		weather_data[i].time = weather_forecast_testing_4h[i].time + time(NULL);
		weather_data[i].condition = weather_forecast_testing_4h[i].condition;		
		//weather_data[i].time = weather_forecast_testing_night[i].time + time(NULL);
		//weather_data[i].condition = weather_forecast_testing_night[i].condition;
	}
	
	weather_data_start = &weather_data[0];

	/* main loop */ 	
	while(1) {
	
		/* update system state values and set preferred power level */
		set_power_level(current_state, P10);
		current_state->time = time(NULL);
		get_current_system_state(current_state);
		PRINT_INFO(V_DEBUG, "current time is          %s", ctime(&(current_state->time)));
		PRINT_INFO(V_DEBUG, "current battery voltage: %.4fV\n", current_state->v_main_bat);
		
		/* ------------- 
		 *  active mode
		 * ------------- */
		toggle_active_mode(1);
		
		/* do some task in active mode */
		log_to_file(LOG_PATH_MEASUREMENT, "measurement simulation\n");
		do_active_work();
		
		/* ------------------------------------
		 * this is where the prediction happens 
		 * ------------------------------------*/
		/* optimize current power level using prediction */
		prediction_time = time(NULL) + PREDICTION_HORIZON;
		optimize_system_state_predicted(current_state, prediction_time, weather_data_start);

		PRINT_INFO(V_DEBUG, "switching to power level: %d\n", current_state->plevel->level);

		log_system_state(LOG_PATH_STATES, current_state);
		log_system_state_list(LOG_PATH_PREDICTION, current_state);

		/* delete predicted states */
		delete_list(&(current_state->next));
	 
		/* -------------- 
		 *  passive mode
		 * -------------- */
		toggle_active_mode(0);
		
		next_active_time = get_next_active_time(current_state);
		PRINT_INFO(V_DEBUG, "next active time will be %s", ctime(&next_active_time));
		
		/* setting next RTC wakeup and go to sleep */
		sysfs_rtc_set_wakealarm(next_active_time);
		freeze_system();	
	}
	
	return 0;
}
