/* useful set of sysfs util functionality
 *
 * Copyright (c) 2008 Jonathan Cameron
 * Copyright (c) 2014 Richard Guenther
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 */

#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <stdio.h>
#include <stdint.h>
#include <dirent.h>
#include <errno.h>

#include "sysfs_utils.h"

/* Made up value to limit allocation sizes */
#define IIO_MAX_NAME_LENGTH 30
#define IIO_MAX_PATH_LENGTH	75

#define FORMAT_SCAN_ELEMENTS_DIR "%s/scan_elements"
#define FORMAT_TYPE_FILE "%s_type"

/* GPIO */
#define SYSFS_GPIO_PATH				"/sys/class/gpio"
#define SYSFS_GPIO_VALUE_PATH		SYSFS_GPIO_PATH "gpio%d/value"
#define SYSFS_GPIO_EXPORT_PATH		SYSFS_GPIO_PATH "export"
#define SYSFS_GPIO_UNEXPORT_PATH	SYSFS_GPIO_PATH "unexport"

/* ADC */
#define SYSFS_ADC_DEVICE			"/iio:device0"
#define SYSFS_ADC_CHANNEL_PATH		SYSFS_ADC_DEVICE "/in_voltage%d_raw"
#define SYSFS_ADC_SCALE_PATH		SYSFS_ADC_DEVICE "/in_voltage_scale"

/* RTC */
#define SYSFS_RTC_PATH				"/sys/class/rtc/rtc0/"

/* LED */
#define SYSFS_LED_PATH				"/sys/class/leds/"

char *iio_dir = "/sys/bus/iio/devices/";

inline int _write_sysfs_int(char *filename, char *basedir, int val, int verify)
{
	int ret=0;
	FILE *sysfsfp;
	int test;
	char *temp = malloc(strlen(basedir) + strlen(filename) + 2);
	if (temp == NULL)
		return -ENOMEM;
	sprintf(temp, "%s/%s", basedir, filename);
	sysfsfp = fopen(temp, "w");
	if (sysfsfp == NULL) {
		printf("failed to open %s\n", temp);
		ret = -errno;
		goto error_free;
	}
	fprintf(sysfsfp, "%d", val);
	fclose(sysfsfp);
	if (verify) {
		sysfsfp = fopen(temp, "r");
		if (sysfsfp == NULL) {
			printf("failed to open %s\n", temp);
			ret = -errno;
			goto error_free;
		}
		fscanf(sysfsfp, "%d", &test);
		fclose(sysfsfp);
		if (test != val) {
			printf("Possible failure in int write %d to %s%s\n",
				val,
				basedir,
				filename);
			ret = -1;
		}
	}
error_free:
	free(temp);
	return ret;
}

int write_sysfs_int(char *filename, char *basedir, int val)
{
	return _write_sysfs_int(filename, basedir, val, 0);
}

int write_sysfs_int_and_verify(char *filename, char *basedir, int val)
{
	return _write_sysfs_int(filename, basedir, val, 1);
}

int _write_sysfs_string(char *filename, char *basedir, char *val, int verify)
{
	int ret = 0;
	FILE  *sysfsfp;
	char *temp = malloc(strlen(basedir) + strlen(filename) + 2);
	if (temp == NULL) {
		printf("Memory allocation failed\n");
		return -ENOMEM;
	}
	sprintf(temp, "%s/%s", basedir, filename);
	sysfsfp = fopen(temp, "w");
	if (sysfsfp == NULL) {
		printf("Could not open %s\n", temp);
		ret = -errno;
		goto error_free;
	}
	fprintf(sysfsfp, "%s", val);
	fclose(sysfsfp);
	if (verify) {
		sysfsfp = fopen(temp, "r");
		if (sysfsfp == NULL) {
			printf("could not open file to verify\n");
			ret = -errno;
			goto error_free;
		}
		fscanf(sysfsfp, "%s", temp);
		fclose(sysfsfp);
		if (strcmp(temp, val) != 0) {
			printf("Possible failure in string write of %s "
				"Should be %s "
				"written to %s\%s\n",
				temp,
				val,
				basedir,
				filename);
			ret = -1;
		}
	}
error_free:
	free(temp);

	return ret;
}

/**
 * write_sysfs_string_and_verify() - string write, readback and verify
 * @filename: name of file to write to
 * @basedir: the sysfs directory in which the file is to be found
 * @val: the string to write
 **/
int write_sysfs_string_and_verify(char *filename, char *basedir, char *val)
{
	return _write_sysfs_string(filename, basedir, val, 1);
}

/**
 * write_sysfs_string() - string write
 * @filename: name of file to write to
 * @basedir: the sysfs directory in which the file is to be found
 * @val: the string to write
 **/
int write_sysfs_string(char *filename, char *basedir, char *val)
{
	return _write_sysfs_string(filename, basedir, val, 0);
}

/**
 * read_sysfs_posint() - read positive int from sysfs
 * @filename: name of file to read from
 * @basedir: the sysfs directory in which the file is to be found
 **/
int read_sysfs_posint(char *filename, char *basedir)
{
	int ret;
	FILE  *sysfsfp;
	char *temp = malloc(strlen(basedir) + strlen(filename) + 2);
	if (temp == NULL) {
		printf("Memory allocation failed");
		return -ENOMEM;
	}
	sprintf(temp, "%s/%s", basedir, filename);
	sysfsfp = fopen(temp, "r");
	if (sysfsfp == NULL) {
		ret = -errno;
		goto error_free;
	}
	fscanf(sysfsfp, "%d\n", &ret);
	fclose(sysfsfp);
error_free:
	free(temp);
	return ret;
}

/**
 * read_sysfs_posint() - read float from sysfs
 * @filename: name of file to read from
 * @basedir: the sysfs directory in which the file is to be found
 * @val: pointer to the variable in which the value is to be stored
 **/
int read_sysfs_float(char *filename, char *basedir, float *val)
{
	float ret = 0;
	FILE  *sysfsfp;
	char *temp = malloc(strlen(basedir) + strlen(filename) + 2);
	if (temp == NULL) {
		printf("Memory allocation failed");
		return -ENOMEM;
	}
	sprintf(temp, "%s/%s", basedir, filename);
	sysfsfp = fopen(temp, "r");
	if (sysfsfp == NULL) {
		ret = -errno;
		goto error_free;
	}
	fscanf(sysfsfp, "%f\n", val);
	fclose(sysfsfp);
	
error_free:
	free(temp);
	return ret;
}

/**
 * sysfs_gpio_export() - export gpio
 * @gpio: gpio number to export
 **/
int sysfs_gpio_export(unsigned int gpio)
{
	return write_sysfs_int("export", SYSFS_GPIO_PATH, gpio);
}

/**
 * sysfs_gpio_unexport() - unexport gpio
 * @gpio: gpio number to unexport
 **/
int sysfs_gpio_unexport(unsigned int gpio)
{
	return write_sysfs_int("unexport", SYSFS_GPIO_PATH, gpio);
}

/**
 * sysfs_gpio_read_value() - read gpio value
 * @gpio: gpio number to read value from
 **/
int sysfs_gpio_read_value(unsigned int gpio)
{
	int ret=0, result;
	char *temp;
	
	temp = malloc(strlen(SYSFS_GPIO_VALUE_PATH) + 2);
	if (temp == NULL) {
		printf("Memory allocation failed");
		return -ENOMEM;
	}
	sprintf(temp, "gpio%d/value", gpio);
	
	ret = write_sysfs_int("export", SYSFS_GPIO_PATH, gpio);
	result = read_sysfs_posint(temp, SYSFS_GPIO_PATH);
	ret = write_sysfs_int("unexport", SYSFS_GPIO_PATH, gpio);
	
	free(temp);
	if(ret != 0)
		return ret;
	
	return result;
}

/**
 * sysfs_gpio_write_value() - write gpio value
 * @gpio: gpio number to write value to
 **/
int sysfs_gpio_write_value(unsigned int gpio, int value)
{
	int ret, result;
	char *temp;
	
	temp = malloc(strlen(SYSFS_GPIO_VALUE_PATH) + 2);
	if (temp == NULL) {
		printf("Memory allocation failed");
		return -ENOMEM;
	}
	sprintf(temp, "gpio%d/value", gpio);
	
	ret = write_sysfs_int("export", SYSFS_GPIO_PATH, gpio);
	ret = write_sysfs_int(temp, SYSFS_GPIO_PATH, value);
	ret = write_sysfs_int("unexport", SYSFS_GPIO_PATH, gpio);
	
	free(temp);
	return ret;
}

/**
 * sysfs_adc_read_channel() - read float value from adc channel
 * @channel: iio channel number to read value from
 **/
float sysfs_adc_read_channel(unsigned int channel)
{
	int ret;
	int raw_value;
	float raw_scale;
	char adc_channel_path[IIO_MAX_PATH_LENGTH];
	
	snprintf(adc_channel_path, IIO_MAX_PATH_LENGTH, SYSFS_ADC_CHANNEL_PATH, channel);
	
	raw_value = read_sysfs_posint(adc_channel_path, iio_dir);
	ret = read_sysfs_float(SYSFS_ADC_SCALE_PATH, iio_dir, &raw_scale);
	if(ret != 0)
		return ret;
		
	return (raw_scale * raw_value);
}

/**
 * sysfs_rtc_read_time() - read time from rtc in UNIX format
 **/
int sysfs_rtc_read_time(void)
{
	return read_sysfs_posint("since_epoch", SYSFS_RTC_PATH);
}

/**
 * sysfs_rtc_read_wakealarm() - read wakealarm from rtc
 **/
int sysfs_rtc_read_wakealarm(void)
{
	return read_sysfs_posint("wakealarm", SYSFS_RTC_PATH);
}

/**
 * sysfs_rtc_reset_wakealarm() - reset the rtc wakealarm by setting it to 0
 * @channel: iio channel number to read value from
 **/
int sysfs_rtc_reset_wakealarm(void)
{
	return write_sysfs_int("wakealarm", SYSFS_RTC_PATH, 0);
}

/**
 * sysfs_rtc_set_wakealarm() - set the rtc wakealarm
 * @time: alarm time in UNIX format
 **/
int sysfs_rtc_set_wakealarm(int time)
{
	sysfs_rtc_reset_wakealarm();
	return write_sysfs_int("wakealarm", SYSFS_RTC_PATH, time);
}

/**
 * sysfs_rtc_set_wakealarm() - set the rtc wakealarm
 * @time: alarm time in UNIX format
 **/
int sysfs_led_set_brightness(char *led_name, int brightness)
{
	return write_sysfs_int(led_name, SYSFS_LED_PATH, brightness);
}
