
#include "input_device.h"

#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <linux/input.h>
#include <linux/uinput.h>
#include "debug.h"
#include "config.h"
#include <linux/fb.h>

#define UINPUT "/dev/uinput"

//if a touch reports multiple keys, selecte the one which is first in the list
static const int touch_key_precedence[] = 
{
	BTN_TOUCH,
	BTN_LEFT,
	BTN_TOOL_PEN,
};

int select_touch_key( const struct input_device * input, int key)
{
	int i;
	if(	input->touch_key == -1)
		return key;

	// touch reports multiple keys, select one as the touch down key
	for( i = 0; i < ARRAY_SIZE(touch_key_precedence); i++)
	{
		if( key == touch_key_precedence[i])
			return key;
		if( input->touch_key == touch_key_precedence[i])
			return input->touch_key;
	}
	if(input->touch_key > key)
		return key;
}

int get_device_info(struct input_device * input)
{
	int res, len, i;
	input->mt_channel_cnt = 0;
	input->touch_key = -1;

    res = ioctl(input->fd, EVIOCGID, &input->info);
	if ( res < 0) return res;
    res = ioctl(input->fd, EVIOCGNAME(INPUT_NAME_LEN), &input->name);
	if ( res < 0) return res;

	pr_dbg("%s vendor %04hx product %04hx version %04hx bus %d\n",
		input->name, input->info.vendor, input->info.product, input->info.version, input->info.bustype);


    len = ioctl(input->fd, EVIOCGBIT(0, sizeof(input->bitmask)), input->bitmask);
    if (len < 0) { pr_err("ioctl EVIOCGBIT for bitmask failed: %s\n", strerror(errno)); return len;   }
	for( i = 0; i <= 0x17; i++)
		if(EvdevBitIsSet(input->bitmask, i))
			pr_dbg( "%s: event 0x%02x %s\n", input->name, i, evdev_event_names[i]);

	if(EvdevBitIsSet(input->bitmask, EV_REL))
	{
		len = ioctl(input->fd, EVIOCGBIT(EV_REL, sizeof(input->rel_bitmask)), input->rel_bitmask);
		if (len < 0) { pr_err("ioctl EVIOCGBIT for EV_REL failed: %s\n", strerror(errno)); return len;   }
		for( i = REL_X; i <= REL_CNT; i++)
			if(EvdevBitIsSet(input->rel_bitmask, i))
				pr_dbg( "%s: rel axis 0x%02x %s\n", input->name, i, evdev_rel_names[i]);
	}else pr_dbg( "%s: no rel axis found\n", input->name);

	if(EvdevBitIsSet(input->bitmask, EV_ABS)){
		len = ioctl(input->fd, EVIOCGBIT(EV_ABS, sizeof(input->abs_bitmask)), input->abs_bitmask);
		if (len < 0) { pr_err("ioctl EVIOCGBIT for EV_ABS failed: %s\n", strerror(errno)); return len;   }

		/*
		 * Do not try to validate absinfo data since it is not expected
		 * to be static, always refresh it in evdev structure.
		 */
		for (i = ABS_X; i <= ABS_MAX; i++) {
			if (EvdevBitIsSet(input->abs_bitmask, i)) {
				len = ioctl(input->fd, EVIOCGABS(i), &input->absinfo[i]);
				if (len < 0) { pr_err("ioctl EVIOCGABSi(%d) failed: %s\n", i, strerror(errno));return len;   }

				pr_dbg( "%s: absolute axis 0x%02x %s [%d..%d]\n", input->name, i, evdev_abs_names[i], input->absinfo[i].minimum, input->absinfo[i].maximum);

				if( !input->mt_channel_cnt  && i >= ABS_MT_SLOT && i <= ABS_MT_TOOL_Y)
					input->mt_channel_cnt = 1; // At least one mt channel
			}
		}
	}else{
		pr_err( "%s: Device has no absolute axis\n", input->name, i);
		return -1;
	}

	if(EvdevBitIsSet(input->bitmask, EV_KEY)){
		len = ioctl(input->fd, EVIOCGBIT(EV_KEY, sizeof(input->key_bitmask)), input->key_bitmask);
		if (len < 0) { pr_err(  "ioctl EVIOCGBIT for EV_KEY failed: %s\n", strerror(errno)); return len;   }
		for (i = 0; i < KEY_CNT; i++)
			if (EvdevBitIsSet(input->key_bitmask, i)) 
			{
				pr_dbg( "%s: key 0x%02x \n", input->name, i);
				input->touch_key = select_touch_key( input, i);
			}
	}
	pr_dbg( "%s: touch key is 0x%02x \n", input->name, i);

	return 0;
}

#if 0
int create_duplicate( struct input_device * input)
{
	int res, i;

    input->uin = open( UINPUT, O_WRONLY | O_NONBLOCK);
    if( input->uin < 0){ pr_err( "Failed to open uinput: %d \n", input->uin ); return input->uin; }

    memset(&input->uidev, 0, sizeof(input->uidev));
    snprintf(input->uidev.name, UINPUT_MAX_NAME_SIZE, "guf touch");

    input->uidev.id.bustype = input->info.bustype;
    input->uidev.id.vendor  = input->info.vendor;
    input->uidev.id.product = input->info.product;
    input->uidev.id.version = input->info.version;

	// Copy single touch abs
	res =  ioctl (input->uin, UI_SET_EVBIT, EV_ABS);
	if (res < 0) { pr_err(  "ioctl UI_SET_EVBIT for EV_ABS failed: %s\n", strerror(errno)); return res;   }

	for (i = ABS_X; i < ABS_MT_SLOT; i++) {
		if (EvdevBitIsSet(input->abs_bitmask, i)) 
		{
			res = ioctl(input->uin, UI_SET_ABSBIT, i);
			if (res < 0) { pr_err(  "ioctl UI_SET_ABSBIT for %d failed: %s\n", i, strerror(errno)); return res;   }

			// abs axis ranges
			input->uidev.absmax[i] = input->absinfo[i].maximum;
			input->uidev.absmin[i] = input->absinfo[i].minimum;
			// no extra filtering in linux input
			input->uidev.absfuzz[i] = 0;
			input->uidev.absflat[i] = 0;
		}
	}
	
	// Copy mt events
	if(input->mt_channel_cnt)
	{
		i = ABS_MT_SLOT;
		res = ioctl(input->uin, UI_SET_ABSBIT, i);
		if (res < 0) { pr_err(  "ioctl UI_SET_ABSBIT for %d failed: %s\n", i, strerror(errno)); return res;   }

		// abs axis ranges
		input->uidev.absmax[i] = 4;
		input->uidev.absmin[i] = 0;

		for (i = ABS_MT_SLOT + 1; i < ABS_MAX; i++) {
			switch(i)
			{
			case ABS_MT_TRACKING_ID:
			case ABS_MT_POSITION_X:
			case ABS_MT_POSITION_Y:
				break;
			default:
				if( !mtdev_has_mt_event( input->mtdev, i))
					continue;
			}

			res = ioctl(input->uin, UI_SET_ABSBIT, i);
			if (res < 0) { pr_err(  "ioctl UI_SET_ABSBIT for %d failed: %s\n", i, strerror(errno)); return res;   }

			// no extra filtering in linux input
			input->uidev.absfuzz[i] = 0;
			input->uidev.absflat[i] = 0;
			// abs axis ranges
			input->uidev.absmax[i] = mtdev_get_abs_maximum(input->mtdev, i);
			input->uidev.absmin[i] = mtdev_get_abs_minimum(input->mtdev, i);
		}
	}

	// Copy all keys
	if(EvdevBitIsSet(input->bitmask, EV_KEY)){
		res =  ioctl (input->uin, UI_SET_EVBIT, EV_KEY);
		if (res < 0) { pr_err(  "ioctl UI_SET_EVBIT for EV_KEY failed: %s\n", strerror(errno)); return res;   }
		
		for (i = 0; i < KEY_CNT; i++)
			if (EvdevBitIsSet(input->key_bitmask, i)) 
			{
				res = ioctl(input->uin, UI_SET_KEYBIT, i);
				if (res < 0) { pr_err(  "ioctl UI_SET_KEYBIT for %d failed: %s\n", i, strerror(errno)); return res;   }
			}
	}

    if(write(input->uin, &input->uidev, sizeof(input->uidev)) < 0)
	{ pr_err( " Failed write uinput\n"); return -1; }


    if(ioctl(input->uin, UI_DEV_CREATE) < 0)
	{ pr_err( "Failed to create uinput device\n"); return -1; }

	return 0;
}
#endif

static int get_screen_size(int * xres, int * yres)
{
	struct fb_var_screeninfo var;
	char * fbdevice = "/dev/fb0";
	int fb_fd;
	fb_fd = open(fbdevice, O_RDWR);
	if (fb_fd == -1) {
		perror("open fbdevice");
		return -1;
	}

	if (ioctl(fb_fd, FBIOGET_VSCREENINFO, &var) < 0) {
		perror("ioctl FBIOGET_VSCREENINFO");
		close(fb_fd);
		return -1;
	}

	*xres = var.xres;
	*yres = var.yres;
	close(fbdevice);
	return 0;
}

int create_out_device( struct input_device * input, bool multi)
{
	int res, i, j;
	int xres, yres;


	if( get_screen_size(&xres, &yres))
	{
		pr_dbg("%s %d Error reding fb size\n", __func__, __LINE__);
		xres = input->absinfo[ ABS_X].maximum - input->absinfo[ ABS_X].minimum;
		yres = input->absinfo[ ABS_Y].maximum - input->absinfo[ ABS_Y].minimum;
	}

	pr_dbg("%s %d\n", __func__, __LINE__);
    input->uin = open( UINPUT, O_WRONLY | O_NONBLOCK);
    if( input->uin < 0){ pr_err( "Failed to open uinput: %d \n", input->uin ); return input->uin; }

    memset(&input->uidev, 0, sizeof(input->uidev));
    snprintf(input->uidev.name, UINPUT_MAX_NAME_SIZE, "guf touch");

    input->uidev.id.bustype = input->info.bustype;
    input->uidev.id.vendor  = input->info.vendor;
    input->uidev.id.product = input->info.product;
    input->uidev.id.version = input->info.version;

	// Copy all abs
	res =  ioctl (input->uin, UI_SET_EVBIT, EV_ABS);
	if (res < 0) { pr_err(  "ioctl UI_SET_EVBIT for EV_ABS failed: %s\n", strerror(errno)); return res;   }

#if 0
	if(input->mtdev)
	{
		int abs_axes[] = { ABS_MT_SLOT, ABS_MT_TRACKING_ID, ABS_MT_POSITION_X, ABS_MT_POSITION_Y, ABS_MT_PRESSURE};
		for (j = 0; j < ARRAY_SIZE(abs_axes); j++) {
			i = abs_axes[j];
			res = ioctl(input->uin, UI_SET_ABSBIT, i);
			if (res < 0) { pr_err(  "ioctl UI_SET_ABSBIT for %d failed: %s\n", i, strerror(errno)); return res;   }

			// no extra filtering in linux input
			input->uidev.absfuzz[i] = 0;
			input->uidev.absflat[i] = 0;
			// abs axis ranges
			input->uidev.absmax[i] = mtdev_get_abs_maximum(input->mtdev, i);
			input->uidev.absmin[i] = mtdev_get_abs_minimum(input->mtdev, i);
		}
	}
	else
	{
		int abs_axes[] = { ABS_X, ABS_Y, ABS_PRESSURE };
		for (j = 0; j < ARRAY_SIZE(abs_axes); j++) {
			i = abs_axes[j];
			res = ioctl(input->uin, UI_SET_ABSBIT, i);
			if (res < 0) { pr_err(  "ioctl UI_SET_ABSBIT for %d failed: %s\n", i, strerror(errno)); return res;   }

			// no extra filtering in linux input
			input->uidev.absfuzz[i] = 0;
			input->uidev.absflat[i] = 0;

			if (EvdevBitIsSet(input->abs_bitmask, i)) 
			{
				// abs axis ranges
				input->uidev.absmax[i] = input->absinfo[i].maximum;
				input->uidev.absmin[i] = input->absinfo[i].minimum;
			}else{
				// Should not happen
				pr_err(  "Evdev Bit is not set for %d failed: %s\n", i, strerror(errno)); return -1;
			}
		}
	}
#else
	ioctl(input->uin, UI_SET_ABSBIT,  ABS_X);
	input->uidev.absfuzz[ ABS_X] = 0;
	input->uidev.absflat[ ABS_X] = 0;
	input->uidev.absmax[ ABS_X] =  xres;
	input->uidev.absmin[ ABS_X] =  0;
	EvdevSetBit(input->u_abs_bitmask, ABS_X);

	ioctl(input->uin, UI_SET_ABSBIT,  ABS_Y);
	input->uidev.absfuzz[ ABS_Y] = 0;
	input->uidev.absflat[ ABS_Y] = 0;
	input->uidev.absmax[ ABS_Y] =  yres;
	input->uidev.absmin[ ABS_Y] =  0;
	EvdevSetBit(input->u_abs_bitmask, ABS_Y);
	
	if (EvdevBitIsSet(input->abs_bitmask, ABS_PRESSURE)) {
		ioctl(input->uin, UI_SET_ABSBIT,  ABS_PRESSURE);
		input->uidev.absfuzz[ ABS_PRESSURE] = 0;
		input->uidev.absflat[ ABS_PRESSURE] = 0;
		input->uidev.absmax[ ABS_PRESSURE] =  input->absinfo[ ABS_PRESSURE].maximum;
		input->uidev.absmin[ ABS_PRESSURE] =  input->absinfo[ ABS_PRESSURE].minimum;
		EvdevSetBit(input->u_abs_bitmask, ABS_PRESSURE);
	}
	
	if( multi && input->mt_channel_cnt)
	{
		ioctl(input->uin, UI_SET_ABSBIT,  ABS_MT_POSITION_X);
		input->uidev.absfuzz[ ABS_MT_POSITION_X] = 0;
		input->uidev.absflat[ ABS_MT_POSITION_X] = 0;
		input->uidev.absmax[ ABS_MT_POSITION_X] =  xres;
		input->uidev.absmin[ ABS_MT_POSITION_X] =  0;
		EvdevSetBit(input->u_abs_bitmask, ABS_MT_POSITION_X);

		ioctl(input->uin, UI_SET_ABSBIT,  ABS_MT_POSITION_Y);
		input->uidev.absfuzz[ ABS_MT_POSITION_Y] = 0;
		input->uidev.absflat[ ABS_MT_POSITION_Y] = 0;
		input->uidev.absmax[ ABS_MT_POSITION_Y] =  yres;
		input->uidev.absmin[ ABS_MT_POSITION_Y] =  0;
		EvdevSetBit(input->u_abs_bitmask, ABS_MT_POSITION_Y);

		if (EvdevBitIsSet(input->abs_bitmask, ABS_MT_PRESSURE)) {
			ioctl(input->uin, UI_SET_ABSBIT,  ABS_MT_PRESSURE);
			input->uidev.absfuzz[ ABS_MT_PRESSURE] = 0;
			input->uidev.absflat[ ABS_MT_PRESSURE] = 0;
			input->uidev.absmax[ ABS_MT_PRESSURE] =  input->absinfo[ ABS_MT_PRESSURE].maximum;
			input->uidev.absmin[ ABS_MT_PRESSURE] =  input->absinfo[ ABS_MT_PRESSURE].minimum;
			EvdevSetBit(input->u_abs_bitmask, ABS_MT_PRESSURE);
		}

		ioctl(input->uin, UI_SET_ABSBIT,  ABS_MT_SLOT);
		if (EvdevBitIsSet(input->abs_bitmask, ABS_MT_SLOT)) {
			input->uidev.absmax[ ABS_MT_SLOT] =  input->absinfo[ ABS_MT_SLOT].maximum;
			input->uidev.absmin[ ABS_MT_SLOT] =  input->absinfo[ ABS_MT_SLOT].minimum;
		}else{
			input->uidev.absmax[ ABS_MT_SLOT] =  MAX_MT_POINTS;
			input->uidev.absmin[ ABS_MT_SLOT] =  0;
		}
		EvdevSetBit(input->u_abs_bitmask, ABS_MT_SLOT);

		ioctl(input->uin, UI_SET_ABSBIT,  ABS_MT_TRACKING_ID);
		if (EvdevBitIsSet(input->abs_bitmask, ABS_MT_TRACKING_ID)) {
			input->uidev.absmax[ ABS_MT_TRACKING_ID] =  input->absinfo[ ABS_MT_TRACKING_ID].maximum;
			input->uidev.absmin[ ABS_MT_TRACKING_ID] =  input->absinfo[ ABS_MT_TRACKING_ID].minimum;
		}else{
			input->uidev.absmax[ ABS_MT_TRACKING_ID] =  MAX_MT_TRACKING_ID;
			input->uidev.absmin[ ABS_MT_TRACKING_ID] =  0;
		}
		EvdevSetBit(input->u_abs_bitmask, ABS_MT_TRACKING_ID);
	}

#endif

	// Copy all keys
	if(EvdevBitIsSet(input->bitmask, EV_KEY)){
		res =  ioctl (input->uin, UI_SET_EVBIT, EV_KEY);
		if (res < 0) { pr_err(  "ioctl UI_SET_EVBIT for EV_KEY failed: %s\n", strerror(errno)); return res;   }
		
		for (i = 0; i < KEY_CNT; i++)
			if (EvdevBitIsSet(input->key_bitmask, i)) 
			{
				res = ioctl(input->uin, UI_SET_KEYBIT, i);
				if (res < 0) { pr_err(  "ioctl UI_SET_KEYBIT for %d failed: %s\n", i, strerror(errno)); return res;   }
			}
	}

    if(write(input->uin, &input->uidev, sizeof(input->uidev)) < 0)
	{ pr_err( " Failed write uinput\n"); return -1; }


    if(ioctl(input->uin, UI_DEV_CREATE) < 0)
	{ pr_err( "Failed to create uinput device\n"); return -1; }

	return 0;
}
