#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>
#include <fcntl.h>
#include <errno.h>
#include <linux/input.h>
#include <linux/uinput.h>
#include <signal.h>
#include <sys/select.h>

#include <mtdev.h>

#include "touch_point.h"
#include "touch_report.h"
#include "debug.h"
#include "evdev_helper.h"
#include "input_device.h"
#include "lua_interface.h"
#include "tslib_wrapper.h"
#include "config.h"

//====================================================

#define EVENT "/dev/input/event0"
// The events are stored in an array, this number defines the 
//
#define DEFAULT_MAX_EVENTS 5
/* Input event set combines all events 
 * until one SYN ev
 */

//====================================================

static const int max_mt_points = MAX_MT_POINTS;
static bool shutdown = false;
static bool verbose = false;
int single_touch = true;
int multi_touch = true;

static int last_mt_tracking_id = 0;
//====================================================

static void sig_handler(int signo)
{
	// If normal closing down failed
	if( shutdown) exit(-1);

	if (signo == SIGINT)
	{
		printf("touchfiler: received SIGINT\n");
		shutdown = true;
	}
}

static int touch_report_clear( struct input_device * input, struct touch_report * report)
{
	int i;

	report->ev_cnt = 0;
	report->syn_received = false;

	// TODO needs the single touch to be cleared ?
	if(		report->single_touch->updated 
		&& (!report->single_touch->current.down  && !report->single_touch->last_send.down ))
	{
		touch_point_clear( &report->single_touch->current );
		touch_point_clear( &report->single_touch->previous );
		touch_point_clear( &report->single_touch->processed );
		touch_point_clear( &report->single_touch->last_send );
	}

	// Clear unused touchpoints
	if( EvdevBitIsSet( input->abs_bitmask, ABS_MT_SLOT ))
	{
		for( i = 0; i < report->max_tp_cnt; i++)
		{
				//&&  report->mt_touch[i].current.mt_tracking_id == -1 )	// Protocol B is supposed to send tracking ID = -1 on release
			if(		report->mt_touch[i].updated 
				&& (!report->mt_touch[i].current.down  && !report->mt_touch[i].last_send.down ))
			{
				touch_point_clear( &report->mt_touch[i].current );
				touch_point_clear( &report->mt_touch[i].previous );
				touch_point_clear( &report->mt_touch[i].processed );
				touch_point_clear( &report->mt_touch[i].last_send );
			}
		}
		// TODO Check if a driver is missing this -1 tracking ID and detect it differently
	}
	else
	{
		for( i = 0; i < report->max_tp_cnt; i++)
		{
			// Type A protocol just stops sending x and y
			if(    (!report->mt_touch[i].updated ) 
				|| ((!report->mt_touch[i].current.down ) && !report->mt_touch[i].last_send.down )
				)
			{
				touch_point_clear( &report->mt_touch[i].current );
				touch_point_clear( &report->mt_touch[i].processed );
				touch_point_clear( &report->mt_touch[i].last_send );
				touch_point_clear( &report->mt_touch[i].previous );
			}
		}
	}

	// Reset all updated flags
	report->single_touch->updated = false;
	for(i = 0; i < report->max_tp_cnt; i++)
		report->mt_touch[i].updated = false;
}

/* 
 * Extract all xyp data sets and calls the filter on them
 */
static int decode_report( const struct input_device * input, struct touch_report * report)
{
	int i;
	struct input_event *ev;
	int btn = -1;
	struct touch_point * p_mt, *p_single;
	bool single = false;
	struct touch_point _tp_single, *tp_single = &_tp_single;
	struct touch_point tp_mt[MAX_MT_POINTS];

	int next_mt_p = 0;
	struct touch_point * next_mt_channel(void)
	{
		// next mt slot might begin but check if the current was used at all
		// to prevent empty slots when both ABS_MT_SLOT and SYN_MT_REPORT are send 
		// ( which is nothing the protocol allows, but who knows)
		if( touch_point_empty(&tp_mt[next_mt_p]))
		{
			return &tp_mt[next_mt_p];
		}

		next_mt_p ++;
		if( next_mt_p == max_mt_points)
		{
			// TODO ignore the rest of the mt channels or allocate more memory
			pr_err("No more touch points available\n");
		}
		return &tp_mt[next_mt_p];
	}
	/*	Event decoding

		Events ABS_X -- ABS_MT_SLOT - 1 sind single touch events
			Touch down == KEY
			oder Touch down = X/Y event vorhanden, wie touch up detektieren?

		Events >= ABS_MT_SLOT
			
			Alle events bis SYN_MT_REPORT in einen mt punkt
			
			mt->slot wird gesetzt wenn  ein event empfangen wird
			
			Touch Down
				gesetzt fuer ABS_MT_TRACKING_ID != -1, up ABS_MT_TRACKING_ID == -1

				1 fùr event received, 0 fùr kein event received 

		Einsortieren wenn x,y,p oder d gesetzt

	*/


	p_single = tp_single;

	touch_point_clear(tp_single);
	for( i = 0; i < max_mt_points; i++)
		touch_point_clear(&tp_mt[i]);

	p_mt = next_mt_channel();

	for( i = 0; i < report->ev_cnt; i++)
	{
		ev = &report->events[i];
		switch( ev->type )
		{
		case EV_ABS:
			switch( ev->code)
			{
				case ABS_X:
					p_single->x = ev->value;
					break;
				case ABS_Y:
					p_single->y = ev->value;
					break;
				case ABS_PRESSURE:
					p_single->p = ev->value;
					break;
				case ABS_MT_SLOT:
					report->current_mt_slot = ev->value;
					p_mt = next_mt_channel();
					break;
				case ABS_MT_TRACKING_ID:
				case ABS_MT_POSITION_X:
				case ABS_MT_POSITION_Y:
				case ABS_MT_PRESSURE:
					touch_point_add_event( p_mt, ev, report->current_mt_slot);
					break;
			}
			break;
		case EV_KEY:
			//pr_dbg("Key: %d\n", ev->value) ;
			if(ev->code == input->touch_key)
				p_single->down = ev->value;
			break;
		case EV_SYN:
			switch(ev->code)
			{
			case SYN_REPORT: // should be last of the events
				p_mt = next_mt_channel();	// Increment counter
				if( i != report->ev_cnt - 1)
					pr_dbg("Syn Report not last event %d, %d\n", i, report->ev_cnt );
				break;
			case SYN_MT_REPORT:
				p_mt = next_mt_channel();
				break;
			case SYN_DROPPED:
				// TODO handle dropped according to:
				/*
				 * SYN_DROPPED:
				 - Used to indicate buffer overrun in the evdev client's event queue.
				 Client should ignore all events up to and including next SYN_REPORT
				 event and query the device (using EVIOCG* ioctls) to obtain its
				 current state
				   */
				break;
			case SYN_CONFIG:
				break;
			}
			
			break;
		default:
			// Just keep the event, but don't handle it
			break;
		}
	}

	//pr_dbg("%d mt: % 4d % 4d % 4d % 4d % 4d % 4d \n", next_mt_p, p_single->x, p_single->y, tp_mt[0].x, tp_mt[0].y,  tp_mt[1].x, tp_mt[1].y);
	
	if( !touch_point_empty(p_single))
		touch_point_container_add_touch_point( report->single_touch, p_single);

	touch_report_insert_mt_points( report, tp_mt, next_mt_p);
	return 0;
}

static int generate_new_tracking_id(void)
{
	last_mt_tracking_id++;
	if( last_mt_tracking_id > MAX_MT_TRACKING_ID)
		last_mt_tracking_id = 0;
	return last_mt_tracking_id;
}

static int fill_missing_data( const struct input_device * input, struct touch_report * report)
{
	int i, ret = 0;
	int copy_int_if_not_set( int * a, int * b)
	{
		if( *a == NOT_YET_WRITTEN)
		{
			if( *b == NOT_YET_WRITTEN)
			{	*a = NOT_YET_WRITTEN; return -1;	}
			*a = *b;
		}
		return 0;
	}

	int copy_param_if_not_set( struct touch_point * tpc, struct touch_point * tpp, int slot)
	{
		int ret = 0, i;

		// TODO use ioctls to request the parameters that have never been set
		if(copy_int_if_not_set( &tpc->x, &tpp->x)) 	{	pr_dbg("Parameter X was never set\n");	ret = -1; 	}
		if(copy_int_if_not_set( &tpc->y, &tpp->y)) 	{	pr_dbg("Parameter Y was never set\n");	ret = -1; 	}
		if(copy_int_if_not_set( &tpc->mt_tracking_id, &tpp->mt_tracking_id)){
			tpc->mt_tracking_id = generate_new_tracking_id();			//  Create tracking ID
		}
		if( copy_int_if_not_set( &tpc->down, &tpp->down) ) { pr_dbg("Parameter Y was never set\n");	ret = -1; 	}
		if(copy_int_if_not_set( &tpc->mt_slot, &tpp->mt_slot))	{
			tpc->mt_slot = slot;//  Set slot
		}
		copy_int_if_not_set( &tpc->p, &tpp->p);			// Is OK if not set, 

		return ret;
	}

	if( report->single_touch->updated)
	{
		if(copy_param_if_not_set( &report->single_touch->current, &report->single_touch->previous, NOT_YET_WRITTEN ))
			ret = -1;
		report->single_touch->processed = report->single_touch->current; 
	}

	for( i = 0; i < report->max_tp_cnt; i++)
	{
		// Create button flag
		if(! EvdevBitIsSet( input->abs_bitmask, ABS_MT_SLOT ))
		{
			// Type A protocol just stops sending x and y
			if( !report->mt_touch[i].updated )
				if(report->mt_touch[i].previous.down == 1)
				{
					report->mt_touch[i].current.down = 0;
					report->mt_touch[i].updated = 1;
				}
		}

		if( report->mt_touch[i].updated )
		{
			if(copy_param_if_not_set( &report->mt_touch[i].current, &report->mt_touch[i].previous, i ))
				ret = -1;
			report->mt_touch[i].processed = report->mt_touch[i].current; 
		}
	}
	return ret;
}

static int process_report( const struct input_device * input, struct touch_report * report, 
	struct lua_script * lua_pre, struct ts_wrapper * ts, struct lua_script * lua_post )
{
	int ret, i;

	ret = decode_report( input, report);

	ret = fill_missing_data( input, report);

	if( lua_pre)
		ret = lua_filtering( lua_pre, report);

	if( ts)
		ret = tswrap_process( ts, report);

	if( lua_post)
		ret = lua_filtering( lua_post, report);


	return 0;
}

static int send_event( struct input_device * input, struct timeval * time, __u16 type, __u16 code, int value)
{
	static struct timeval last_time = {0,0};
	struct input_event ev = { .type = type, .code = code, .value = value};

	//if(value == NOT_YET_WRITTEN)
	//	return 0;
	if( type == EV_ABS && !	EvdevBitIsSet(input->u_abs_bitmask, code))
		return 0;

	dump_event("send", &ev);
	if( time) last_time = *time;
	ev.time = last_time;

	if(write(input->uin, &ev, sizeof(struct input_event)) < 0)
		if(!shutdown) 
			return -1;
	return 1;
}

static int send_touch_slot( struct input_device *input, struct timeval * time, int slot)
{
	if( input->current_slot == slot)
		return 0;
	input->current_slot = slot;
	return send_event( input, time, EV_ABS, ABS_MT_SLOT, slot ); 
}

static int send_touch_point( struct input_device *input, struct touch_point_container * c)
{
	struct touch_point * c_tp, * p_tp;
	c_tp = &c->processed;
	p_tp = &c->last_send;
	int s = 0;
	bool p = c_tp->p != NOT_YET_WRITTEN;
	
	if(!c->is_mt)
	{	// if changed in up down send complete report, otherwise send only changes
		if( c_tp->down != p_tp->down ) {
			if(c_tp->down)	{	// Touch down
				s |= send_event( input, &c->current.time, EV_ABS, ABS_X, c_tp->x ); 
				s |= send_event( input, &c->current.time, EV_ABS, ABS_Y, c_tp->y ); 
				if(p)
					s |= send_event( input, &c->current.time, EV_ABS, ABS_PRESSURE, c_tp->p ); 
				s |= send_event( input, &c->current.time, EV_KEY, input->touch_key, 1); 
			}else			{	// Release
				s |= send_event( input, &c->current.time, EV_ABS, ABS_PRESSURE, 0 ); 
				s |= send_event( input, &c->current.time, EV_KEY, input->touch_key, 0 ); 
			}
		}else{
			if(c_tp->down)	{	// finger is down report changes
				if(c_tp->x != p_tp->x)
					s |= send_event( input, &c->current.time, EV_ABS, ABS_X, c_tp->x ); 
				if(c_tp->y != p_tp->y)
					s |= send_event( input, &c->current.time, EV_ABS, ABS_Y, c_tp->y ); 
				if(p && c_tp->p != p_tp->p)
					s |= send_event( input, &c->current.time, EV_ABS, ABS_PRESSURE, c_tp->p ); 
			}
		}
	}
	else{
		if( c_tp->down != p_tp->down ) {
			if(c_tp->down)	{	// Touch down
				s |= send_touch_slot( input, &c->current.time, c_tp->mt_slot ); 
				s |= send_event( input, &c->current.time, EV_ABS, ABS_MT_TRACKING_ID, c_tp->mt_tracking_id ); 
				s |= send_event( input, &c->current.time, EV_ABS, ABS_MT_POSITION_X, c_tp->x ); 
				s |= send_event( input, &c->current.time, EV_ABS, ABS_MT_POSITION_Y, c_tp->y ); 
				if(p) s |= send_event( input, &c->current.time, EV_ABS, ABS_MT_PRESSURE, c_tp->p ); 

			}else			{	// Release
				s |= send_touch_slot( input, &c->current.time, c_tp->mt_slot ); 
				s |= send_event( input, &c->current.time, EV_ABS, ABS_MT_TRACKING_ID, -1 ); 
			}
		}else{
			if(c_tp->down)	{	// finger is down report changes
				bool x = (c_tp->x != p_tp->x);
				bool y = (c_tp->y != p_tp->y);
				bool p = p && (c_tp->p != p_tp->p);

				if(x|y|p)	s |= send_touch_slot( input, &c->current.time, c_tp->mt_slot ); 
				if(x)		s |= send_event( input, &c->current.time, EV_ABS, ABS_MT_POSITION_X, c_tp->x ); 
				if(y)		s |= send_event( input, &c->current.time, EV_ABS, ABS_MT_POSITION_Y, c_tp->y ); 
				if(p)		s |= send_event( input, &c->current.time, EV_ABS, ABS_MT_PRESSURE, c_tp->p ); 
			}
		}
	}
	memcpy( p_tp, c_tp, sizeof( struct touch_point));
	return s;
}

static int send_report( struct input_device * input, struct touch_report * report)
{
	int i;
	struct input_event * ev;
	struct input_event new_ev;
	int s = 0;

	// Send all single touch events
	if(single_touch)
		if(report->single_touch->updated)
			s |= send_touch_point( input, report->single_touch);
	
	if(multi_touch)
		for( i = 0; i < report->max_tp_cnt; i++)
			if(report->mt_touch[i].updated)
				s |= send_touch_point( input, &report->mt_touch[i]);

	if(s == 1) // Any events send, we are done, send sync with the lasts event timestamp
		send_event( input, 0, EV_SYN, SYN_REPORT, 0 ); 
	else{
		if(s)	// report error
			return -1;
	}

	return 0;
}

static void usage(char * prog)
{
	printf("Usage: %s [--help] [--version] [--input <device>]\n", prog);
}
static void version(void)
{
	printf("%s Version: %s built: %s %s\n", PACKAGE_NAME, PACKAGE_VERSION, __DATE__, __TIME__);
}

int main(int argc, char * const argv[])
{
	int result = 0;
    struct input_event     _ev, *ev;
    int                    dx, dy;
    int                    i, num_bytes;
	struct input_device		_input, *input;
	struct timeval timeout, wakeupperiod;
	fd_set set;
	struct touch_report * report;
	struct lua_script * lua_pre, * lua_post;
	struct ts_wrapper * ts;

	char * eventname = 0;
	ev = &_ev;
	input = &_input;

	static const struct option options[] = {
		{ "help", no_argument, NULL, 'h' },
		{ "version", no_argument, NULL, 'V' },
		{ "verbose", no_argument, NULL, 'v' },
		{ "input", required_argument, NULL, 'i' },
		{ "singletouch", no_argument, NULL, 's' },
		{}
	};
	while (1) {
		int option;

		option = getopt_long(argc, argv, "i:vhV", options, NULL);
		if (option == -1)
			break;
		switch (option) {
		case 'h':
			version();
			usage(argv[0]);
			exit(0);
		case 'V':
			version();
			exit(0);
		case 'v':
			verbose = true;
			break;
		case 's':
			break;
		case 'i':
			if(optarg)
				eventname = strdup( optarg);
			else
			{
				printf("failed to read parameter -i/--input\n");
				usage(argv[0]);
				exit(0);
			}
		}
	}

	if( !eventname)
		eventname = strdup(EVENT);
	
	pr_dbg( "Starting touchfilter\n");
	wakeupperiod.tv_sec = 5;
	wakeupperiod.tv_usec = 0;

    input->fd = open(eventname, O_RDONLY ); //| O_NONBLOCK
    if(input->fd < 0)  die(exit, "error: open %s", eventname);

	if( get_device_info( input))
		die( exit_close, "error: get_device_info");

	if( create_out_device( input, multi_touch)) die( exit_close, "error: create_out_device");
	
	report = touch_report_malloc_event( DEFAULT_MAX_EVENTS, MAX_MT_POINTS );
	if( !report) die( exit_uinput, "error: failed to allocated memory");

	// Initialize current slot from device ( if available)
	report->current_mt_slot = NOT_YET_WRITTEN;
	if( EvdevBitIsSet( input->abs_bitmask, ABS_MT_SLOT ))
	{
		struct input_absinfo absinfo;
		int ret;
		if( ret = ioctl (input->fd, EVIOCGABS(ABS_MT_SLOT ), &absinfo) < 0 )
			fprintf(stderr, "%s: IOCTL EVIOCGABS(ABS_MT_SLOT ) failed: %d\n", __func__, ret);
		else
			report->current_mt_slot = absinfo.value;
	}

	input->current_slot = NOT_YET_WRITTEN;

	lua_pre = setup_lua_script("./prefilter.lua");
	if( !lua_pre) pr_warn( "Failed to setup lua interface, lua pre filter will not be available.\n");
	lua_post = setup_lua_script("./postfilter.lua");
	if( !lua_post) pr_warn( "Failed to setup lua interface, lua post filter will not be available.\n");

	ts = tswrap_initilize(MAX_MT_POINTS + 1);
	if( !ts) pr_warn( "Failed to setup tslib interface, tslib filtering will not be available.\n");
		
	// Register signal to cleanup on shutdown
	if( signal(SIGINT, sig_handler) == SIG_ERR )
		die(exit_uinput, "error: can't catch SIGINT\n");

	FD_ZERO(&set); /* clear the set */

	while(!shutdown)
	{
		timeout = wakeupperiod;
		FD_SET( input->fd, &set); /* add our file descriptor to the set */
		result = select( input->fd + 1, &set, NULL, NULL, &timeout );
		if( result < 0 ) if( !shutdown)	die( exit_free, "select on event failed"); else continue;

		if( result)
		{
			// get events from device
			num_bytes = read( input->fd, ev, sizeof(struct input_event));

			if( num_bytes <= 0)	if(!shutdown)die(exit_free, "Failed to read from event\n"); else continue;

			result = touch_report_add_event( report, ev);
			if( result < 0) die(exit_free, "error: in touch_report_add_event\n");
			if( !result ) continue;			// Continue to read events
		}else
		{
			// timeout
			//pr_warn("timeout\n", i);

			// repeat events in case of current.down = 0 and last_send.down = 1
			int i = 0;
			bool added = false;
			if( 0 == report->single_touch->current.down && 1 == report->single_touch->last_send.down )
			{
				pr_dbg("Repeat tp single\n", i);
				touch_point_container_add_touch_point( report->single_touch, &report->single_touch->current);
				added = true;
			}
			for( i = 0; i < max_mt_points; i++)
				if( 0 == report->mt_touch[i].current.down && 1 == report->mt_touch[i].last_send.down )
				{
					pr_dbg("Repeat tp %d\n", i);
					touch_point_container_add_touch_point( &report->mt_touch[i], &report->mt_touch[i].current);
					added = true;
				}
			if(!added){
				dump_report( report);
			   	continue;
			}
		}

		// process received report
		if( process_report( input, report, lua_pre, ts, lua_post) < 0)
			die(exit_free, "error: in process report\n");

		dump_report( report);

		// Send events
		if( send_report( input, report))
			die(exit_free, "error: in send_report\n");

		if( touch_report_clear(input, report) < 0)
			die(exit_free, "error: in touch_report_clear\n");

		{
			bool waiting = false;
			if( 0 == report->single_touch->current.down && 1 == report->single_touch->last_send.down )
				waiting = true;
			else
				for( i = 0; i < max_mt_points; i++)
					if( 0 == report->mt_touch[i].current.down && 1 == report->mt_touch[i].last_send.down )
					{
						waiting = true;
						break;
					}	
			if(waiting)
			{
				wakeupperiod.tv_usec = 25000;
				wakeupperiod.tv_sec = 0;
			}
			else
			{
				wakeupperiod.tv_usec = 0;
				wakeupperiod.tv_sec = 5;
			}

		}
	}

	pr_dbg("Shutdown touch filter\n");

exit_free:
	if(ts)
		tswarp_free(ts);
	if(lua_pre)
		free_lua_script(lua_pre);
	if(lua_post)
		free_lua_script(lua_post);
exit_free_touch_report:
	touch_report_free(report);
exit_uinput:
	close(input->uin);
exit_close:
	close(input->fd);
exit:
	free(eventname);
	return result;
}

