
#include "touch_report.h"
#include <stdlib.h>
#include <string.h>
#include "debug.h"

#define SQARE_DISTANCE( a, b)( (((a)->x - (b)->x)*((a)->x - (b)->x)) + (((a)->y - (b)->y)*((a)->y - (b)->y)))
#define SQARE_DISTANCE_X( a, b)( (((a)->x - (b)->x)*((a)->x - (b)->x)) )
#define SQARE_DISTANCE_Y( a, b)( (((a)->y - (b)->y)*((a)->y - (b)->y)) )
static int touch_report_insert_mt_points_by_id( struct touch_report * report, struct touch_point * p_mt, int p_mt_len);
static int touch_report_insert_mt_points_by_slot( struct touch_report * report, struct touch_point * p_mt, int p_mt_len);
static int touch_report_insert_mt_points_by_dist( struct touch_report * report, struct touch_point * p_mt, int p_mt_len);

/*
 *  Allocates a new input event according to the events reported by the device 
 */
struct touch_report * touch_report_malloc_event( int max_ev_cnt, int max_mt_pts)
{
	struct touch_report * report;
	int i;

	report = malloc( sizeof(struct touch_report));
	pr_func_line();
	if( !report){ perror(" Failed to allocate memory\n"); return 0;}
	
	report->max_ev_cnt = max_ev_cnt;
	report->ev_cnt = 0;
	report->syn_received = false;
	report->current_mt_slot = NOT_YET_WRITTEN;

	report->events = malloc( sizeof(struct input_event) * report->max_ev_cnt );
	if( !report->events){ perror(" Failed to allocate memory\n"); return 0;}

	report->single_touch = malloc( sizeof(struct touch_point_container) );
	if( !report->single_touch){ perror(" Failed to allocate memory\n"); return 0;}
	report->single_touch->is_mt = false;

	touch_point_clear( &report->single_touch->current);
	touch_point_clear( &report->single_touch->previous);
	touch_point_clear( &report->single_touch->processed);
	touch_point_clear( &report->single_touch->last_send);

	report->max_tp_cnt = max_mt_pts;
	report->mt_touch = malloc( sizeof(struct touch_point_container) * report->max_tp_cnt );
	if( !report->mt_touch){ perror(" Failed to allocate memory\n"); return 0;}

	for( i = 0; i< report->max_tp_cnt; i++)
	{
		report->mt_touch[i].is_mt = true;
		touch_point_clear( &report->mt_touch[i].current);
		touch_point_clear( &report->mt_touch[i].previous);
		touch_point_clear( &report->mt_touch[i].processed);
		touch_point_clear( &report->mt_touch[i].last_send);
	}

	pr_func_line();
	return report;
}


struct touch_report * touch_report_free( struct touch_report * report)
{
	pr_func_line();
	if( !report) return 0;
	free( report->events);
	free( report->single_touch);
	free( report->mt_touch);
	free( report);
}

int touch_report_resize( struct touch_report * report, int max_ev_cnt)
{
	struct input_event * new_events;

	pr_func_line();

	if( !report) return -1;

	// We don't shrink
	if( report->max_ev_cnt >= max_ev_cnt) return 0;

	pr_func_line();

	new_events = malloc( sizeof(struct input_event) * max_ev_cnt );

	if( !new_events){ 
		pr_err("Failed to allocate memory\n");
		return -1; 
	}
	
	pr_func_line();
	report->max_ev_cnt = max_ev_cnt;

	memset( new_events, 0, sizeof(struct input_event) * max_ev_cnt );
	memcpy( new_events, report->events, sizeof(struct input_event) * report->ev_cnt );

	pr_func_line();
	free(report->events);
	
	report->events = new_events;
	
	return 0;
}

/* add ev to set, returns if 1 if the set is complete, 0 if not, negative values on error */
int touch_report_add_event( struct touch_report * report, const struct input_event * ev)
{
	if( report->ev_cnt == report->max_ev_cnt)
		if( touch_report_resize(report, report->max_ev_cnt + 10)) return -1;
	
	dump_event("rece", ev);

	memcpy( &report->events[report->ev_cnt], ev, sizeof(struct input_event ));
	report->ev_cnt++;
	
	// TODO handle this for devices that do not send SYN_REPORT
	if (ev->type == EV_SYN && ev->code == SYN_REPORT)
		report->syn_received  = true;
	return report->syn_received ? 1 : 0;
}

int touch_report_insert_mt_points( struct touch_report * report, struct touch_point * tp, int count)
{
	// Check if the slot is set to select method of tracking
	// Maybe we could also use the tracking id
	if(!count)	return 0;

	//pr_dbg(" %s %d %d %d %d\n", __func__, count, tp[0].mt_slot, tp[0].mt_tracking_id, NOT_YET_WRITTEN);
	if( tp[0].mt_slot != NOT_YET_WRITTEN)
		touch_report_insert_mt_points_by_slot( report, tp, count);
	else if( tp[0].mt_tracking_id != NOT_YET_WRITTEN)
		touch_report_insert_mt_points_by_id( report, tp, count);
	else
		touch_report_insert_mt_points_by_dist( report, tp, count);
}

static int touch_report_insert_mt_points_by_id( struct touch_report * report, struct touch_point * p_mt, int p_mt_len)
{
	int i, j, ret = 0;

	for(j = 0; j < p_mt_len; j++)
	{
		struct touch_point * p = &p_mt[j];
		bool added = false;
		// now map the decoded mt channels to the already existing ones
		// easy for protocol type B using the slot ID
		// harder for the protocol type A, we use the distance
		if( touch_point_empty(p))
			continue;

		if( p->mt_tracking_id == NOT_YET_WRITTEN)
		{
			pr_err("Received events without mt_tracking_id set. Ignoring\n");
			continue;
		}
		// Use slot
		for( i = 0; i < report->max_tp_cnt; i++)
		{
			struct touch_point_container * tpc = &report->mt_touch[i];
			if( p->mt_tracking_id == tpc->current.mt_tracking_id )
			{
				touch_point_container_add_touch_point(tpc, p);
				added = true;
				break;
			}
		}

		if(!added)
		{
			int i;
			for( i = 0; i < report->max_tp_cnt; i++)
			{
				struct touch_point_container * tpc = &report->mt_touch[i];
				if( touch_point_empty( &tpc->current) )
				{
					touch_point_container_add_touch_point( tpc, p);
					added = true;
					break;
				}
			}
		}

		if(!added)
		{
			pr_dbg("Failed to add touch point %d\n", j);
			ret = -1;
		}
	}
	return ret;
}

static int touch_report_insert_mt_points_by_slot( struct touch_report * report, struct touch_point * p_mt, int p_mt_len)
{
	int i, j, ret = 0;

	for(j = 0; j < p_mt_len; j++)
	{
		struct touch_point * p = &p_mt[j];
		bool added = false;
		// now map the decoded mt channels to the already existing ones
		// easy for protocol type B using the slot ID
		// harder for the protocol type A, we use the distance
		if( touch_point_empty(p))
			continue;

		if( p->mt_slot == NOT_YET_WRITTEN)
		{
			pr_err("Received events without mt_slot set. Ignoring\n");
			continue;
		}
		// Use slot
		for( i = 0; i < report->max_tp_cnt; i++)
		{
			struct touch_point_container * tpc = &report->mt_touch[i];
			if( p->mt_slot == tpc->current.mt_slot )
			{
				touch_point_container_add_touch_point(tpc, p);
				added = true;
				break;
			}
		}

		if(!added)
		{
			int i;
			for( i = 0; i < report->max_tp_cnt; i++)
			{
				struct touch_point_container * tpc = &report->mt_touch[i];
				if( touch_point_empty( &tpc->current) )
				{
					touch_point_container_add_touch_point( tpc, p);
					added = true;
					break;
				}
			}
		}

		if(!added)
		{
			pr_dbg("Failed to add touch point %d\n", j);
			ret = -1;
		}
	}
	return ret;
}

static int touch_report_insert_mt_points_by_dist( struct touch_report * report, struct touch_point * p_mt, int p_mt_len)
{
	int i, j, ret;
	unsigned int dist;

	for(j = 0; j < p_mt_len; j++)
	{
		struct touch_point * p = &p_mt[j];
		bool added = false;
		if( touch_point_empty(p))
		{
			pr_dbg("%s Try to add empty tp\n", __func__);
			continue;
		}

		// TODO this mapping can fail if for example a second point is new and in the list before the first point
		// Use distance and position
		int closest_dist = INT32_MAX, closest_id = -1;
		for( i = 0; i < report->max_tp_cnt; i++)
		{
			struct touch_point_container * tpc = &report->mt_touch[i];
			if( tpc->updated)
				continue;

			if( tpc->current.x != NOT_YET_WRITTEN  )		// we make sure later that x and y are always present
			{	
				if( p->x != NOT_YET_WRITTEN && p->y != NOT_YET_WRITTEN)	{
					dist = SQARE_DISTANCE( &(tpc->current), p);	// no sqrt as we only use it to compare
				}else if( p->x != NOT_YET_WRITTEN){
					dist = SQARE_DISTANCE_X( &(tpc->current), p);	// no sqrt as we only use it to compare
				}else if ( p->y != NOT_YET_WRITTEN){
					dist = SQARE_DISTANCE_Y( &(tpc->current), p);	// no sqrt as we only use it to compare
				}else{
					// no x an no y, how should we match this ?
					pr_err("Missing x and y value");
					continue;
				}

				if( closest_dist > dist )
				{
					closest_dist = dist;
					closest_id = i;
				}
			}
		}
		if( closest_id != -1 )
		{
			//pr_dbg("%s add to %d\n", __func__, closest_id);
			touch_point_container_add_touch_point(&report->mt_touch[closest_id], p);
			added = true;
		}

		if(!added)
		{
			//pr_dbg("%s Not yet added\n", __func__);
			int i;
			for( i = 0; i < report->max_tp_cnt; i++)
			{
				struct touch_point_container * tpc = &report->mt_touch[i];
				if( touch_point_empty( &tpc->current) )
				{
					touch_point_container_add_touch_point( tpc, p);
					added = true;
					break;
				}
			}
		}

		if(!added)
		{
			pr_dbg("Failed to add touch point\n");
			ret =  -1;
		}
	}
	return ret;
}

