/*
 * Copyright (C) 2012 Garz & Fricke GmbH
 * Author Hendrik Woidt <hendrik.woidt@garz-fricke.de>
 * 
 * This file is placed under the LGPL.  Please see the file
 * COPYING for more details.
 *
 * This filter implements an adaptive mean Filter. The principle of operation
 * is to change the number of samples which are averaged based on the movement
 * speed. This leads to high jitter suppression during slow moves while keeping
 * latency at a minimum during fast moves.
 *
 * The filter adds up samples until either the depth is reached or the distance
 * between the last passed sample and the average of the samples since then
 * exceeds the specified "threshold" value. When one of these conditions is fullfilled
 * the average of the samples is passed on to the next filter. If the distance 
 * exceeded the threshold the filterdepth is decreased by the number of samples specified 
 * in "deltadepth", otherwise it is incremented by this amount. The depth is clamped to a value 
 * between "maxdepth" and 1. It is initialized on touchdown with a value of "maxdepth"/2
 * 
 * "threshold"
 * The threshold value should be choosen according to the experienced noise. The noise should
 * mostly stay below this value. A good test for this would be to disable all filters and 
 * press at one point. The spread of the received samples will be a good indicator for an 
 * initial value.
 *
 * "maxdepth"
 * The maximum depth should be choosen so that an acceptable smoothness during slow moves is 
 * reached. The smoothness has to be traded off to the latency. Increasing the value of maxdepth
 * will increase both smoothness and latency. It mainly influences behaviour during slow moves
 * 
 * "deltadepth"
 * The depth-increments influence the adaption-rate of the filter. The higher the value is, the
 * faster is the responsiveness to movement-speed-changes. When choosen too low, serious lag during
 * accellerating and deaccelerating  moves can be a result. When choosen too high, the filter will 
 * be susceptible to noise spikes. This mainly influences behaviour during fast moves.
 *
 * "debug"
 * this option controls debug output. When set to a value !=0 it will display a message on stdout for
 * each sample that is passed on to the next filter, which has the following format:
 * XXXx:YYYy (depth: $samples/$depth) delta:$delta
 * where XXX and YYY are the current coordinates, $samples is the number of samples which are averaged
 * for these x and y values and $depth is the current filter depth. (When $samples<$depth it means that
 * the threshold-condition has stopped the sample collection -> see above for a description of 
 * the condition) finally $delta is the value which was used in this sample to determine whether the
 * depth should be increased or decreased.
 *
 * 
 * possible improvements:
 * - Currently the responsiveness to fast short moves is pretty bad. Sometimes there are only 4 samples/sec
 *   which are passed to the next filter. This could be improved by adding a condition which checks the
 *   timestamps of the samples to guarantee that at least a minimum number of samples are emitted per sec.
 * - There is a problem with really short touches. These are not passed on if the median-filter is placed
 *   after this filter in the chain. The problem is that the median filter "swallows" the first sample so
 *   that the upper layers only receive the touch-up event. Since this filter does a pretty good job on
 *   noise suppression this is a minor problem, if the median-filter is not used. This might as well be 
 *   problematic with the dejitter and variance filter.
 */


#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include "tslib.h"
#include "tslib-filter.h"

struct tslib_adaptmean {
	struct tslib_module_info module;
        struct ts_sample acc;
	unsigned char samples;
	unsigned char depth;
        struct ts_sample last;
	unsigned int flags;
	unsigned char threshold;
	unsigned char depthmax;
	unsigned char depthdelta;
	unsigned char debug;
#define VAR_PENUP		0x00000001
};

static void reset_vars(struct tslib_adaptmean *var) {
	var->acc.x = 0;
	var->acc.y = 0;
	var->acc.pressure = 0;
	var->samples = 0;
}

static int sqdist(struct ts_sample *s1, struct ts_sample *s2) {
	int dx, dy;
	dx = s2->x - s1->x;
	dy = s2->y - s1->y;
	return dx*dx + dy*dy;
}

static int adaptmean_read(struct tslib_module_info *info, struct ts_sample *samp, int nr)
{
	struct tslib_adaptmean *var = (struct tslib_adaptmean *)info;
	struct ts_sample current, avg;
	int count = 0;
	int delta;

	while (count < nr) {
		if (info->next->ops->read(info->next, &current, 1) < 1)
			return count;

		if (current.pressure == 0) {
			var->flags |= VAR_PENUP;
			samp [count++] = current;
			reset_vars(var);
		} else if (var->flags & VAR_PENUP) {
			var->flags &= ~VAR_PENUP;
			var->last = current;
			samp [count++] = current;
		} else {
			var->acc.x += current.x;
			var->acc.y += current.y;
			var->acc.pressure += current.pressure;
			var->samples++;

			avg.x = var->acc.x / var->samples;
			avg.y = var->acc.y / var->samples;
			delta = sqdist(&var->last, &avg);

			if(var->samples == var->depth || delta > var->threshold) {
				avg.pressure = var->acc.pressure / var->samples;
				avg.tv = current.tv;
				var->last = avg;
				samp [count++] = avg;

				if(delta > var->threshold)
					var->depth -= var->depthdelta;
				else
					var->depth += var->depthdelta;

				if(var->depth > var->depthmax) var->depth = var->depthmax;
				if(var->depth < 1) var->depth = 1;

				if(var->debug)
					printf("%dx:%dy (depth: %u/%u) delta:%d\n", avg.x, avg.y, var->samples, var->depth, delta);

				reset_vars(var);
			}
		}
	}
	return count;
}

static int adaptmean_fini(struct tslib_module_info *info)
{
	free(info);
        return 0;
}

static const struct tslib_ops adaptmean_ops =
{
	.read	= adaptmean_read,
	.fini	= adaptmean_fini,
};

static int adaptmean_parse(struct tslib_module_info *inf, char *str,
		void *data)
{
	struct tslib_adaptmean *var = (struct tslib_adaptmean *)inf;
	long result;
	int err = errno;

	result = strtol(str,NULL,0);
	if (errno == ERANGE)
	       return -1;

	errno = err;

	switch ((int)data) {
	case 1:
		var->threshold = result * result;
		break;
	case 2:
		var->depthmax = result;
		break;
	case 3:
		var->depthdelta = result;
		break;
	case 4:
		var->debug = result;
		break;
	default:
		return -1;
	}
	return 0;
}


static const struct tslib_vars adaptmean_vars[] =
{
	{ "debug",	(void*) 4, adaptmean_parse},
	{ "deltadepth", (void*) 3, adaptmean_parse},
	{ "maxdepth",	(void*) 2, adaptmean_parse},
	{ "threshold",  (void*) 1, adaptmean_parse},
};

#define NR_VARS (sizeof(adaptmean_vars) / sizeof(adaptmean_vars[0]))

TSAPI struct tslib_module_info *adaptmean_mod_init(struct tsdev *dev, const char *params)
{
	struct tslib_adaptmean *var;

	var = malloc(sizeof(struct tslib_adaptmean));
	if (var == NULL)
		return NULL;

	memset(var, 0, sizeof *var);
	var->module.ops = &adaptmean_ops;

	var->flags = VAR_PENUP;
	var->threshold = 100;
	var->depthmax = 35;
	var->depthdelta = 1;
	var->debug = 0;
	reset_vars(var);
	
	if (tslib_parse_vars(&var->module, adaptmean_vars, NR_VARS, params)) {
		free(var);
		return NULL;
	}


	return &var->module;
}

#ifndef TSLIB_STATIC_ADAPTMEAN_MODULE
	TSLIB_MODULE_INIT(adaptmean_mod_init);
#endif