/*
 *  tslib/plugins/crop.c
 *
 *  Copyright (C) 2012 Garz & Fricke GmbH.
 *
 * This file is placed under the LGPL. Please see the file
 * COPYING for more details.
 *
 * $Id: tslib_add_crop_filter.patch 2586 2013-08-01 07:39:29Z tim.jaacks $
 *
 * Crop scaled touch coordinates to the screen borders.
 * This module needs the guf_xml driver to work, because it
 * reads the screen resolution from /sys/class/misc/guf_xml/
 * configurationFile/variables/display.
 */

#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>

#include <stdio.h>

#include "tslib.h"
#include "tslib-filter.h"

#define MODE_FILE "/sys/class/graphics/fb0/modes"
#define MODE_STRLEN 20

struct tslib_crop {
	struct tslib_module_info module;
	int xres;
	int yres;
};

static int
crop_read(struct tslib_module_info *info, struct ts_sample *samp, int nr)
{
	struct tslib_crop *crop = (struct tslib_crop *)info;
	int ret;

	ret = info->next->ops->read(info->next, samp, nr);
	if (ret >= 0) {
		int nr;

		for (nr = 0; nr < ret; nr++, samp++) {
			if (samp->x < 0)
				samp->x = 0;
			else if (samp->x >= crop->xres)
				samp->x = crop->xres - 1;

			if (samp->y < 0)
				samp->y = 0;
			else if (samp->y >= crop->yres)
				samp->y = crop->yres - 1;
		}
	}

	return ret;
}

static int crop_fini(struct tslib_module_info *info)
{
	free(info);
	return 0;
}

static const struct tslib_ops crop_ops =
{
	.read	= crop_read,
	.fini	= crop_fini,
};

TSAPI struct tslib_module_info *crop_mod_init(struct tsdev *dev, const char *params)
{
	struct tslib_crop *crop;
	int fdmode = -1;
	char str[MODE_STRLEN];

	crop = malloc(sizeof(struct tslib_crop));
	if (crop == NULL)
		return NULL;

	crop->module.ops = &crop_ops;

	fdmode = open(MODE_FILE, O_RDONLY);
	if (fdmode < 0)
		goto error;

	memset(str, 0, MODE_STRLEN);
	if (read(fdmode, str, MODE_STRLEN - 1) > 0)
	{
		char *pos_of_colon = strchr(str, ':');
		char *pos_of_x = strchr(str, 'x');
		char *pos_of_p = strchr(str, 'p');

		if (pos_of_colon == NULL || pos_of_x == NULL || pos_of_p == NULL)
			goto error;

		*pos_of_x = 0;
		*pos_of_p = 0;

		crop->xres = atoi(pos_of_colon + 1);
		if (crop->xres == 0)
			goto error;
		crop->yres = atoi(pos_of_x + 1);
		if (crop->yres == 0)
			goto error;
	}
	else
		goto error;

#ifdef DEBUG
	printf("crop: Screen resolution %dx%d\n", crop->xres, crop->yres);
#endif
	return &crop->module;

error:
	if (fdmode >= 0)
		close(fdmode);
	printf("crop: Could not determine screen resolution!\n");
	return NULL;
}


#ifndef TSLIB_STATIC_CROP_MODULE
	TSLIB_MODULE_INIT(crop_mod_init);
#endif