/*
 *   This program is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <dirent.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include "crc32.h"
#include "ezxml.h"
#include "file_config.h"

// #define DEBUG

#ifdef DEBUG
#define printf_d(args...) printf(args)
#else
#define	printf_d(args...) /* nothing */
#endif

#define	printf_e(args...) fprintf(stderr, args)


/* 
 * read the configuration file, obtain the current xml configuration
 * and store current content in **fisroot_str
 */
int xml_read_from_file(char **fisroot_str)
{
	int fd;
	off_t len;
	ssize_t count;
	int ret=0;

	printf_d("Reading data from file...\n");
	fd = open(CONFIG_FILE_NAME, O_RDONLY);
	if (fd < 0) {
		ret = fd;
		goto clean;
	}

	// Determine file size
	len = lseek(fd, 0, SEEK_END);
	lseek(fd, 0, SEEK_SET);

	*fisroot_str = malloc(len+1);
	if (!*fisroot_str) {
		printf_e("Out of memory\n");
		ret = -ENOMEM;
		goto clean;
	}

	count = read(fd, *fisroot_str, len);
	if (count != len) {
		printf_e("Could not read from file\n");
		ret = -EIO;
		goto clean;
	}
	(*fisroot_str)[len] = 0;

	printf_d("Data ready\n");

clean:
	if (fd >= 0)
		close(fd);
	return ret;
}

/*
 * write **fisroot_str to configuration file
 * If no assume_yes given, prompt for affirmation
 */
int xml_write_to_file(char *fisroot_str, char *ts_conf_str, char *pointercal_str, int assume_yes)
{
	int ret=0;
	char affirm;
	int fd;
	off_t len;
	ssize_t count;

	if ( (! fisroot_str ) ) {
		printf_e("Error writing to device: Insufficient Data\n");
		return 1;
	}
	if ( ! assume_yes ) {
		printf("Update non-volatile configuration - continue (y/n)? ");
		ret=scanf("%c", &affirm);
		if ((ret != 1) || (affirm != 'y'))
			return 1;
		ret = 0;
	}

	printf_d("Opening XML config file for writing...\n");
	fd = open(CONFIG_FILE_NAME, O_WRONLY | O_TRUNC);
	if (fd < 0) {
		printf_e("Could not open file %s for writing (error %d: %s)\n",
				CONFIG_FILE_NAME, errno, strerror(errno));
		ret = errno;
		goto clean;
	}

	len = strlen(fisroot_str);
	count = write(fd, fisroot_str, len);
	if (count != len) {
		printf_e("Could not write to file %s (error %d: %s)\n",
				CONFIG_FILE_NAME, errno, strerror(errno));
		ret = errno;
		goto clean;
	}
	sync();

	if (ts_conf_str) {
		close(fd);
		printf_d("Opening tslib config file for writing...\n");
#ifdef __ANDROID__
		fd = open(TS_CONF_FILE_NAME, O_CREAT | O_WRONLY | O_TRUNC, 0644);
#else
		fd = open(TS_CONF_FILE_NAME, O_CREAT | O_WRONLY | O_TRUNC);
#endif
		if (fd < 0) {
			printf_e("Could not open file %s for writing (error %d: %s)\n",
					TS_CONF_FILE_NAME, errno, strerror(errno));
			ret = errno;
			goto clean;
		}

		len = strlen(ts_conf_str);
		count = write(fd, ts_conf_str, len);
		if (count != len) {
			printf_e("Could not write to file %s (error %d: %s)\n",
					TS_CONF_FILE_NAME, errno, strerror(errno));
			ret = errno;
			goto clean;
		}
		sync();
	}

	if (pointercal_str) {
		close(fd);
		printf_d("Opening pointercal file for writing...\n");
#ifdef __ANDROID__
		fd = open(POINTERCAL_FILE_NAME, O_CREAT | O_WRONLY | O_TRUNC, 0644);
#else
		fd = open(POINTERCAL_FILE_NAME, O_CREAT | O_WRONLY | O_TRUNC);
#endif
		if (fd < 0) {
			printf_e("Could not open file %s for writing (error %d: %s)\n",
					POINTERCAL_FILE_NAME, errno, strerror(errno));
			ret = errno;
			goto clean;
		}

		len = strlen(pointercal_str);
		count = write(fd, pointercal_str, len);
		if (count != len) {
			printf_e("Could not write to file %s (error %d: %s)\n",
					POINTERCAL_FILE_NAME, errno, strerror(errno));
			ret = errno;
			goto clean;
		}
		sync();
	}
clean:
	if (fd >= 0)
		close(fd);
	return ret;
}
